/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import mpii.biotm.util.DeterminePathBasedOnId;
import mpii.biotm.util.ListAllFilenamesInDirectory;
import mpii.biotm.util.ReadFromFile;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.commons.cli.PosixParser;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

public class CommandLineArguments{
	
	protected static final Logger logger = Logger.getLogger(CommandLineArguments.class);
	
	//--------------------------------------
	// Usage:
	// (1) Call initialize()
	// (2) Call parse()
	// (3) Call getArgumentValue() as needed
	//
	// Command line options only specify which input files / text to
	// work on. To specify which modules to run, their parameters,
	// etc. do so by modifying the settings file. Default settings
	// file is:
	// etc/mpii/biotm/biotm.properties
	
	private static Options allAvailableOptions = null;
	public static final String DEFAULT_SETTINGS_FILENAME = "etc/mpii/biotm/biotm.properties";
	private static boolean initialized = false;
	private static CommandLine specifiedArguments = null;
	
	private static boolean existsFileForId(long id, String dir){
		String inputFilename = DeterminePathBasedOnId.getPathWithFilename(id, dir, "txt");
		File inputFile = new File(inputFilename);
		return inputFile.exists();
	}
	
	public static String getArgumentValue(String argumentName){
		String returnString = "";
		
		//--------------------------------------
		// Return empty string if argument is not given by user
		if (specifiedArguments.hasOption(argumentName)){		
			returnString = specifiedArguments.getOptionValue(argumentName);
		}
		
		return returnString;
	}
	
	public static List<Pair<String, String>> getItemsToWorkOn() throws IOException{
		//-------------------------------------
		// In the return list, there are pairs:
		// - Left value is the input, be it a filename, or actual text
		// - Right value is the output, be it a filename where the
		//   output should be saved to, or to be displayed on the
		//   console
		List<Pair<String, String>> returnList = new ArrayList<Pair<String, String>>();
		
		//-------------------------------------
		// Do different things depending on the input mode
		String inputMode = getArgumentValue("input_mode");
		
		//-------------------------------------
		// Everything under a directory
		if (inputMode.equals("directory")){
			String dirIn = removeFinalSlash(getArgumentValue("directory_in"));
			String dirOut = removeFinalSlash(getArgumentValue("directory_out"));
			
			returnList = ListAllFilenamesInDirectory.list(dirIn, dirOut);
		} // Everything under a directory
		
		//-------------------------------------
		// File with one ID per line
		else if (inputMode.equals("id_list_in_file")){
			String rootDirIn = getArgumentValue("id_root_directory_in");
			String rootDirOut = getArgumentValue("id_root_directory_out");
			
			String idFilename = getArgumentValue("id_file");
			List<Long> ids = ReadFromFile.readFileToListOfLongs(idFilename);
			
			for (long id : ids){
				if (existsFileForId(id, rootDirIn)){
					returnList.add(idToItemPair(id, rootDirIn, rootDirOut));
				}
			} // for each ID in range
		} // File with one ID per line
		
		//-------------------------------------
		// ID range
		else if (inputMode.equals("id_range")){
			String rootDirIn = getArgumentValue("id_root_directory_in");
			String rootDirOut = getArgumentValue("id_root_directory_out");
			
			int startId = Integer.valueOf(getArgumentValue("id_from"));
			int endId = Integer.valueOf(getArgumentValue("id_to")); // Inclusive!
			
			for (int id = startId; id <= endId; ++id){
				if (existsFileForId(id, rootDirIn)){
					returnList.add(idToItemPair(id, rootDirIn, rootDirOut));
				}
			} // for each ID in range
		} // ID range
		
		//-------------------------------------
		// Everything under the input directory, but without a
		// corresponding file under the output directory
		if (inputMode.equals("missing_in_directory")){
			String dirIn = removeFinalSlash(getArgumentValue("directory_in"));
			String dirOut = removeFinalSlash(getArgumentValue("directory_out"));
			
			List<Pair<String, String>> tempList = ListAllFilenamesInDirectory.list(dirIn, dirOut);
			
			//-------------------------------------
			// Go thru the filename pairs to see which of
			// the output files are missing, and add them
			// to the return list
			for (Pair<String, String> tempPair : tempList){
				String outputFilename = tempPair.getRight();
				File outputFile = new File(outputFilename);
				if (!outputFile.exists()){
					returnList.add(tempPair);
				}
			}
		} // Everything under a directory
		
		//-------------------------------------
		// Text string
		else if (inputMode.equals("text")){
			Pair<String, String> tempPair = new ImmutablePair<String, String>(getArgumentValue("text"), "console");
			returnList.add(tempPair);
		}
		
		//-------------------------------------
		// Text file
		else if (inputMode.equals("text_file")){
			Pair<String, String> tempPair = new ImmutablePair<String, String>(
					getArgumentValue("text_file_in"),
					getArgumentValue("text_file_out"));
			returnList.add(tempPair);
		}
		
		return returnList;
	}
	
	public static Properties getSettings(String[] args) throws FileNotFoundException, IOException, ParseException{
		String settingsFilename = DEFAULT_SETTINGS_FILENAME;
		
		//-----------------------------------------
		// First take in what the user specified in command line arguments
		CommandLineArguments.initialize();
		CommandLineArguments.parse(args);
		
		if (CommandLineArguments.hasArgument("settings_file")){
			settingsFilename = CommandLineArguments.getArgumentValue("settings_file");
		}
		
		//-----------------------------------------
		// Get some settings from file
		Properties settings = new Properties();
		settings.load(new FileInputStream(settingsFilename));
		
		return settings;
	}

	public static boolean hasArgument(String argumentName){
		return specifiedArguments.hasOption(argumentName);
	}
	
	private static Pair<String, String> idToItemPair(long id, String dirIn, String dirOut){
		String inputFilename = DeterminePathBasedOnId.getPathWithFilename(id, dirIn, "txt");
		String outputFilename = DeterminePathBasedOnId.getPathWithFilename(id, dirOut, "txt");
		return new ImmutablePair<String, String>(inputFilename, outputFilename);
	}
	
	private static void initialize(){
		if (!initialized){
			//-------------------------------------
			// Set up all the available options
			allAvailableOptions = new Options();
			
			//-------------------------------------
			// Settings file and task
			allAvailableOptions.addOption(new Option(
					"settings_file",
					true,
					"settings file; if not specified, default is etc/mpii/biotm/biotm.properties"));
			allAvailableOptions.addOption(new Option(
					"task",
					true,
					"choices are: dinah_creation, bebe_entity_recognition, bebe_data_precomputation"));
			
			//-------------------------------------
			// Based on the task, different further options are relevant
			
			//-------------------------------------
			// Options for dinah_creation
			// TODO
			
			//-------------------------------------
			// Options for bebe_entity_recognition
			allAvailableOptions.addOption(new Option(
					"input_mode",
					true,
					"how the input will be given; choices are: directory, id_list_in_file, id_range, text, text_file"));
			allAvailableOptions.addOption(new Option(
					"directory_in",
					true,
					"for use with directory input_mode; all files under directory_in will be processed, including sub directories"));
			allAvailableOptions.addOption(new Option(
					"directory_out",
					true,
					"for use with directory input_mode; if same as directory_in, all existing files will be overwritten"));
			allAvailableOptions.addOption(new Option(
					"id_root_directory_in",
					true,
					"for use with id-related input_mode's; specify the root directory of input files"));
			allAvailableOptions.addOption(new Option(
					"id_root_directory_out",
					true,
					"for use with id-related input_mode's; if same as id_root_directory_in, all existing files will be overwritten"));
			allAvailableOptions.addOption(new Option(
					"id_file",
					true,
					"file containing one id per line"));
			allAvailableOptions.addOption(new Option("id_from",
					true,
					"start processing from this id"));
			allAvailableOptions.addOption(new Option(
					"id_to",
					true,
					"process up to and including this ID"));
			allAvailableOptions.addOption(new Option(
					"text",
					true,
					"provide a text string enclosed in double quotes to be processed"));
			allAvailableOptions.addOption(new Option(
					"text_file_in",
					true,
					"for use with text_file input_mode; provide full path to a single text file"));
			allAvailableOptions.addOption(new Option(
					"text_file_out",
					true,
					"for use with text_file input_mode; if same as text_file_in, the existing file will be overwritten"));
			
			//-------------------------------------
			// Options for bebe_data_precomputation
			// TODO
		}
		
		initialized = true;
	}
	
	private static void parse(String[] args) throws ParseException{
		//--------------------------------------
		// Parse the user's command line input to see which options have which values
		CommandLineParser parser = new PosixParser();
		specifiedArguments = parser.parse(allAvailableOptions, args);
	}
	
	private static String removeFinalSlash(String inputString){
		//-----------------------------------------
		// If the input string ends with "/", remove it
		if (inputString.endsWith("/")){
			inputString = inputString.substring(0, inputString.length() - 1);
		}
		
		return inputString;
	}
}