/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.bean;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import mpii.biotm.util.RandomNumber;

import org.apache.log4j.Logger;

public class MinHashTableDefinition{
	// I named this class a "definition" rather than "MinHashTable" because this class does
	// not hold the table of bucket ID --> concept ID pairs. Instead, the actual ID data
	// is stored directly in a Berkeley DB database, accessed via the class MinHashBerkeleyDb.
	// In this class, one can find the A coeffs, B coeffs, and can get a hash code given an
	// input.
	
	protected static final Logger logger = Logger.getLogger(MinHashTableDefinition.class);
	
	private List<Integer> aCoeffs;
	private List<Integer> bCoeffs;
	private int numberOfHashFunctions;
	private int prime;
	
	/* Use this constructor when we need to select the random hash functions, i.e.
	 * select the A and B coeffs from a pool. And so this constructor should be
	 * used when building a new MinHash table. */
	public MinHashTableDefinition(int inputNumberOfHashFunctions, int inputPrime,
			List<Integer> poolOfACoeffs, List<Integer> poolOfBCoeffs){
		// Assumption: RandomNumber has already been initialized
		
		//-----------------------------------------
		// Record params that will be constant for this MinHashTable
		numberOfHashFunctions = inputNumberOfHashFunctions;
		prime = inputPrime;
		
		//-----------------------------------------
		// Out of the pool of a's and b's, select numberOfHashFunctions of them
		// Assumption: RandomNumber is already initialized, i.e. do NOT reset the sequence of random numbers!
		aCoeffs = new ArrayList<Integer>();
		bCoeffs = new ArrayList<Integer>();
		int numberOfHashFunctionsToChooseFrom = poolOfACoeffs.size();

		for (int i=0; i < numberOfHashFunctions; ++i){
			int index = RandomNumber.getRandomNumberInRange(0, numberOfHashFunctionsToChooseFrom - 1);
			aCoeffs.add(poolOfACoeffs.get(index));
			bCoeffs.add(poolOfBCoeffs.get(index));
		}
	}
	
	/* Use this constructor when already know the A and B coeffs. And so this constructor
	 * should be used when we know all the settings of a MinHash table, and just need
	 * getHashCode() for lookup purposes. */
	public MinHashTableDefinition(int inputPrime, List<Integer> inputACoeffs, List<Integer> inputBCoeffs){
		numberOfHashFunctions = inputACoeffs.size();
		prime = inputPrime;
		aCoeffs = inputACoeffs;
		bCoeffs = inputBCoeffs;
	}
	
	public List<Integer> getACoeffs(){
		return aCoeffs;
	}
	
	public List<Integer> getBCoeffs(){
		return bCoeffs;
	}
	
	public int getHashCode(Map<Integer, Integer> trigramIdsToCounts){
		//-----------------------------------------
		// Go thru each hashing function, and find the trigram ID
		// that gives the smallest hash value
		List<Integer> minHashValues = new ArrayList<Integer>();
		for (int i=0; i < numberOfHashFunctions; ++i){
			boolean atTheVeryBeginning = true;
			int minHashValue = -1; // -1 means undefined
			
			//-----------------------------------------
			// Apply hashing function to each trigram ID, and record the
			// smallest hash value encountered so far
			for (int trigramId : trigramIdsToCounts.keySet()){
				
				
				//-----------------------------------------
				// Although A coeff, B coeff, trigram ID, and the prime number are
				// all small enough to be handled as int, the multiplcation causes
				// the intermediate result to be bigger than what int can handle.
				// So make sure we do not have overflow. We want to as much as
				// possible use smaller data types so that the resulting database
				// is as compact as possible.
				long intermediateNumber = (long) aCoeffs.get(i) * trigramId + (long) bCoeffs.get(i);
				int hashValue = (int) (intermediateNumber % prime);
				if (hashValue < 0){ // We have an overflow
					logger.fatal("Hash value has overflown");
					System.exit(1);
				}
				
				if (atTheVeryBeginning){
					minHashValue = hashValue;
					atTheVeryBeginning = false;
				}
				else if (hashValue < minHashValue){
					minHashValue = hashValue;
				}
			}
			
			minHashValues.add(minHashValue);
		}
		
		//-----------------------------------------
		// Now we have one smallest hash value for each hash function.
		// Apply some aggregate function to obtain final hash code
		// out of them.
		return getHashCodeFromSmallestHashValues(minHashValues);
	}
	
	private int getHashCodeFromSmallestHashValues(List<Integer> integers){
		//-----------------------------------------
		// Current implementation: Simply sum up the integers
		int sum = 0;
		
		for (int integer : integers){
			sum += integer;
		}
		
		return sum;
	}
}