/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.bean;

import org.apache.commons.lang3.tuple.Pair;

public class Span implements Comparable<Span>{

	private int start;
	private int end;
	
	public Span(int inputStart, int inputEnd){
		start = inputStart;
		end = inputEnd;
	}
	
	public Span(Pair<Integer, Integer> inputPair){
		start = inputPair.getLeft();
		end = inputPair.getRight();
	}
	
	public Span(String inputString){
		// Format: a-b where a, b are integers
		int hyphenPos = inputString.indexOf("-");
		start = Integer.valueOf(inputString.substring(0, hyphenPos));
		end = Integer.valueOf(inputString.substring(hyphenPos + 1));
	}
	
	public int compareTo(Span anotherSpan){
		// Order first by start ascending, then by end descending.
		// Intuitively, with sub strings sharing the same start, this
		// ordering presents the longer sub strings prior to the shorter
		// ones.
		int returnInt = 0;
		
		int anotherStart = anotherSpan.getStart();
		int anotherEnd = anotherSpan.getEnd();
		
		if (start < anotherStart){
			returnInt = -1;
		}
		else if (start > anotherStart){
			returnInt = 1;
		}
		else{ // start == anotherStart
			if (end < anotherEnd){
				returnInt = 1;
			}
			else if (end > anotherEnd){
				returnInt = -1;
			}
		}
		
		return returnInt;
	}
	
	public boolean equals(Object object){
		boolean returnBool = false;
		
		if (object instanceof Span){
			returnBool = (((Span) object).getStart() == start &&
					((Span) object).getEnd() == end);
		}
		
		return returnBool;
	}
	
	public Span getCopy(){
		return new Span(start, end);
	}
	
	public int getEnd(){
		return end;
	}
	
	public int getLength(){
		return end - start;
	}
	
	public int getStart(){
		return start;
	}
	
	public int hashCode(){
		// Concatenate the start and end numbers, and make
		// and integer out of that
		String tempString = start + "" + end;
		return Integer.valueOf(tempString);
	}
	
	public boolean hasSameStartEnd(Span inputSpan){
		// Return true if this span has the same start and end
		// as the input span
		return equals(inputSpan);
	}
	
	public boolean isWithinAnotherSpan(Span inputSpan){
		// Return true of this span is within the input span
		return (inputSpan.getStart() <= start && end <= inputSpan.getEnd());
	}
	
	public boolean isValid(){
		return (0 <= start && 1 <= end && start < end);
	}
	
	public boolean overlapsWithAnotherSpan(Span inputSpan){
		// Return true of this span has overlapping coverage with the input span
		return (inputSpan.getStart() < end || start < inputSpan.getEnd());
	}
	
	public void setEnd(int inputEnd){
		end = inputEnd;
	}
	
	public void setStart(int inputStart){
		start = inputStart;
	}

	public String toString(){
		return (start + "-" + end);
	}
}