/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.bean;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;

import mpii.biotm.util.EmptyStringUtf8;
import mpii.biotm.util.InsertTabs;

public class TextMention implements Comparable<TextMention>{
	
	protected static final Logger logger = Logger.getLogger(TextMention.class);
	
	private List<Span> charSpans = null; // Relative to sentence, i.e. each sentence starts at char 0
	private List<Entity> entities = null;
	private List<String> selectionCriteria = null; // How this text mention was selected
	private int sentenceNumber = -1; // Absolute, i.e. sentence starts at 0 at beginning of document
	private String text = "";
	
	public TextMention(int inputSentenceNumber, List<Span> inputCharSpans, String inputText){
		this(inputSentenceNumber, inputCharSpans, inputText,
				new ArrayList<String>(), // No selection criteria
				new ArrayList<Entity>()); // No entities
	}
	
	public TextMention(int inputSentenceNumber, List<Span> inputCharSpans, String inputText,
			List<String> inputSelectionCriteria, List<Entity> inputEntities){
		sentenceNumber = inputSentenceNumber;
		charSpans = inputCharSpans;
		text = EmptyStringUtf8.getEmptyString() + inputText; // Force the string to be UTF-8
		selectionCriteria = inputSelectionCriteria;
		entities = inputEntities;
	}
	
	public void addEntity(Entity entity){
		entities.add(entity);
		
		//-----------------------------------------
		// Always keep the entities sorted according to their scores
		Collections.sort(entities);
	}
	
	public void addSelectionCriterion(String criterion){
		selectionCriteria.add(criterion);
		
		//-----------------------------------------
		// Always keep the selection criteria sorted
		Collections.sort(selectionCriteria);
	}
	
	private boolean areIdenticalSpans(List<Span> spans1, List<Span> spans2){
		boolean returnBool = true;
		
		if (spans1.size() == spans2.size()){
			int i = 0;
			while (returnBool && i < spans1.size()){
				if (!spans1.get(i).equals(spans2.get(i))){
					returnBool = false;
				}
				++i;
			}
		}
		else{
			returnBool = false;
		}

		return returnBool;
	}
	
	private boolean areInDesiredSpanOrder(List<Span> orderedEarlierSpans, List<Span> orderedLaterSpans){
		boolean returnBool = true;
		
		//-----------------------------------------
		// In the inputs are two lists of spans, and presumably the first
		// input should be ordered earlier than the second one. This method
		// verifies that that is the case.
		//
		// Assumption: Both input spans have the same size, but some
		// individual spans are different, e.g. the 4th span from both
		// inputs are different.
		//
		// Look at the first spans from both input, then second spans, etc.
		// Look at the first pair that are different.
		// Then regular span comparison is used.
		int i = 0;
		while (returnBool & i < orderedEarlierSpans.size()){
			Span earlierSpan = orderedEarlierSpans.get(i);
			Span laterSpan = orderedLaterSpans.get(i);
			
			if (earlierSpan != laterSpan){
				int comparison = earlierSpan.compareTo(laterSpan);
				if (comparison > 0){
					returnBool = false;
				}
			}
			
			++i;
		}
		
		return returnBool;
	}
	
	private String characterSpansToString(){
		String returnString = "";
		
		for (int i=0; i < charSpans.size(); ++i){
			Span charSpan = charSpans.get(i);
			returnString += charSpan.getStart() + "-" + charSpan.getEnd();
			if (i < charSpans.size() - 1){
				returnString += ",";
			}
		}
		
		return returnString;
	}
	
	public int compareTo(TextMention anotherTextMention){
		// Order first by very-start ascending, then by very-end descending.
		// Intuitively, with sub strings sharing the same start, this
		// ordering presents the longer sub strings prior to the shorter
		// ones.
		int returnInt = 0;
		
		int veryStartChar = getVeryStartChar();
		int veryEndChar = getVeryEndChar();
		
		int anotherVeryStartChar = anotherTextMention.getVeryStartChar();
		int anotherVeryEndChar = anotherTextMention.getVeryEndChar();
		
		if (veryStartChar < anotherVeryStartChar){
			returnInt = -1;
		}
		else if (veryStartChar > anotherVeryStartChar){
			returnInt = 1;
		}
		else{ // veryStartChar == anotherVeryStartChar
			if (veryEndChar < anotherVeryEndChar){
				returnInt = 1;
			}
			else if (veryEndChar > anotherVeryEndChar){
				returnInt = -1;
			}
			else{
				// veryEndChar == anotherVeryEndChar
				// So the char spans may be different.
				if (!areIdenticalSpans(getCharacterSpans(), anotherTextMention.getCharacterSpans())){
					// Look at the first span pair that differs. Whichever has
					// the longer span "wins" and becomes ordered earlier.
					if (areInDesiredSpanOrder(getCharacterSpans(), anotherTextMention.getCharacterSpans())){
						returnInt = -1;
					}
					else{
						returnInt = 1;
					}
				}
			}
		}
		
		return returnInt;
	}
	
	public boolean equals(Object object){
		boolean returnBool = false;
		
		//-----------------------------------------
		// Note to self: The method argument must be an object, not a Text Mention.
		// Otherwise say List.contain() will not use this method.
		//
		// Also: If Collections will operation on this, implement and override
		// hashcode().
		if (object != null && object instanceof TextMention){
			returnBool = areIdenticalSpans(getCharacterSpans(), ((TextMention) object).getCharacterSpans());
		}
		
		return returnBool;
	}

	public List<CanonicalEntity> getCanonicalEntities(){
		List<CanonicalEntity> returnList = new ArrayList<CanonicalEntity>();
		
		for (Entity entity : entities){
			if (entity.getType().equals("CanonicalEntity")){
				returnList.add((CanonicalEntity) entity);
			}
		}
		
		return returnList;
	}
	
	public Span getCharacterSpan(int i){
		return charSpans.get(i);
	}
	
	public List<Span> getCharacterSpans(){
		return charSpans;
	}
	
	public String getCharacterSpansAsString(){
		String returnString = "";
		for (int i = 0; i < charSpans.size(); ++i){
			if (i > 0){
				returnString += ",";
			}
			returnString += charSpans.get(i).toString();
		}
		return returnString;
	}
	
	public List<Entity> getEntities(){
		return entities;
	}
	
	public Entity getEntity(int i){
		return entities.get(i);
	}
	
	/** This method returns a list because there may be multiple canonical
	 * entities scoring the same highest score. */
	public List<CanonicalEntity> getHighestScoredCanonicalEntities(){
		List<CanonicalEntity> canonicalEntities = getCanonicalEntities();
		
		if (canonicalEntities.size() > 1){
			double highestScore = canonicalEntities.get(0).getScore();
			
			int numberOfElementsToKeep = 1;
			while (numberOfElementsToKeep < canonicalEntities.size() &&
					canonicalEntities.get(numberOfElementsToKeep).getScore() >= highestScore){
				++numberOfElementsToKeep;
			}
			if (numberOfElementsToKeep < canonicalEntities.size()){
				// It's okee to mutilate the list of canonical entities, because
				// it is freshly instantiated and does not affect the list of
				// all entities.
				canonicalEntities = canonicalEntities.subList(0, numberOfElementsToKeep);
			}
		}
		
		return canonicalEntities;
	}
	
	public int getNumberOfCharacterSpans(){
		return charSpans.size();
	}
	
	public int getNumberOfCanonicalEntities(){
		List<CanonicalEntity> canonicalEntities = getCanonicalEntities();
		return canonicalEntities.size();
	}
	
	public int getNumberOfEntities(){
		return entities.size();
	}
	
	public List<String> getSelectionCriteria(){
		return selectionCriteria;
	}
	
	public int getSentenceNumber(){
		return sentenceNumber;
	}
	
	public String getText(){
		return text;
	}
	
	public int getVeryEndChar(){
		return charSpans.get(charSpans.size()-1).getEnd();
	}
	
	public int getVeryStartChar(){
		return charSpans.get(0).getStart();
	}
	
	public boolean hasCanonicalEntities(){
		return getNumberOfCanonicalEntities() > 0;
	}
	
	public boolean hasEntities(){
		return getNumberOfEntities() > 0;
	}
	
	public boolean isAlreadyMappedToCanonicalEntity(){
		int i = 0;
		boolean foundCanonicalEntity = false;
		
		while (i < entities.size() && !foundCanonicalEntity){
			if (entities.get(i).getType().equals("CanonicalEntity")){
				foundCanonicalEntity = true;
			}
			++i;
		}
		
		return foundCanonicalEntity;
	}
	
	public void setCharSpans(List<Span> inputCharSpans){
		charSpans = inputCharSpans;
	}
	
	public void setEntities(List<Entity> inputEntities){
		entities = inputEntities;
	}
	
	public void setSentenceNumber(int inputSentenceNumber){
		sentenceNumber = inputSentenceNumber;
	}
	
	public void setText(String inputText){
		text = inputText;
	}
	
	public List<String> toArchiveStrings(){
		List<String> returnList = new ArrayList<String>();
		String DELIMITER = "~@~";
		
		//-----------------------------------------
		// One line for details pertaining to the text mention
		String textMentionLine = sentenceNumber + DELIMITER + characterSpansToString() + DELIMITER + getText();
		returnList.add(textMentionLine);
		
		//-----------------------------------------
		// Subsequent lines for details pertaining to the entities, one line per entity
		//
		// Assumption: The entities are already sorted by score descending
		for (Entity entity : getEntities()){
			returnList.addAll(InsertTabs.insertTabs(entity.toArchiveStrings(), 1));
		}
		
		return returnList;
	}
	
	public String toDebugString(){
		return toString();
	}
	
	public String toString(){
		String returnString = "====== Text mention ======\n";
		returnString += "Char spans:         " + characterSpansToString() + "\n";
		returnString += "Text:               " + getText() + "\n";
		
		returnString += "Selection criteria: ";
		for (int i=0; i < selectionCriteria.size(); ++i){
			returnString += selectionCriteria.get(i);
			if (i != selectionCriteria.size()-1){
				returnString += ", ";
			}
		}
		returnString += "\n";

		returnString += "Number of entities: " + entities.size() + "\n";

		for (int i=0; i < getNumberOfEntities(); ++i){
			Entity entity = getEntity(i);
			String temp = "Entity " + i + ":\n";
			temp += entity.toString();
			temp = InsertTabs.insertTabs(temp, 1);
			returnString += temp;
		}
		
		return returnString;
	}
}