/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.paraphrase.precompute;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import mpii.biotm.CommandLineArguments;
import mpii.biotm.util.ProcessUpperLowerCase;
import mpii.biotm.util.ReadFromFile;
import mpii.biotm.util.SaveToFile;
import mpii.biotm.wrapper.morphadorner.WordPluralizer;
import mpii.biotm.wrapper.umls.LookupSemanticGroup;
import mpii.biotm.wrapper.umls.LookupSemanticType;
import mpii.biotm.wrapper.umls.UnwantedSubStringsInConceptNames;

import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;

public class GenerateEntityNamesForParaphrasing{
	
	protected static final Logger logger = Logger.getLogger(GenerateEntityNamesForParaphrasing.class);
	
	private static Map<String, List<String>> cuisToConceptNames = null;
	private static boolean disoOnly = false;
	private static int maxNumberOfWords = -1;
	
	private static List<String> determineUniqueConceptNames(List<String> conceptNames){
		//-----------------------------------------
		// Rearrange the order of the concept names so that
		// those with comma come later. This prioritizes names
		// that are more "naturally English."
		// E.g. for C0023931, the first 3 names are:
		//   Lobstein's Disease -- (1)
		//   Disease, Lobstein  -- (2)
		//   Lobstein Disease   -- (3)
		// (3) is preferrable to (2), but without rearrangement
		// (2) would win.
		conceptNames = sortByCommaSemicolon(conceptNames);
		
		//-----------------------------------------
		// Go thru each concept name, and remove the
		// unwanted sub strings
		List<String> alteredConceptNames = new ArrayList<String>();
		List<String> unchangedConceptNames = new ArrayList<String>();
		for (int i = 0; i < conceptNames.size(); ++i){
			String conceptName = conceptNames.get(i);
			String possiblyAlteredConceptName = UnwantedSubStringsInConceptNames.removeUnwantedSubString(conceptName);
			if (conceptName.equals(possiblyAlteredConceptName)){
				unchangedConceptNames.add(possiblyAlteredConceptName);
			}
			else{
				alteredConceptNames.add(possiblyAlteredConceptName);
			}
		}
		
		//-----------------------------------------
		// A concept name containing an unwanted sub string is highly
		// unlikely to be matched via stringsim. If all the concept
		// names belonging to one CUI contain unwanted sub string, this
		// CUI is highly unlikely to ever be matched by stringsim.
		// Since, for now, we want those concept name / CUI combos that are
		// picked up by stringsim, when all concept names of a CUI are "bad"
		// in this way, discard this CUI. Later in the pipeline, in
		// EstimateMostCommonSenseCuisConceptNames, we want to only
		// consider those concept names that could've been matched
		// via stringsim.
		//
		// Note to self: The output of standardizeConceptName() is for
		// detecting duplicates. The variable conceptName is the
		// one that gets recorded as return value.
		if (unchangedConceptNames.isEmpty()){
			conceptNames.clear();
		}
		else{
			conceptNames = unchangedConceptNames;
		}
		
		//-----------------------------------------
		// Now go thru each concept name again. Consider them all in
		// lower case, and ignore non-alphanumeric characters. Sort
		// the words in ascending order. We are interested in a word
		// combination never seen before. Pay attention that we only
		// accept a concept name when it has no more than the max
		// number of words.
		Set<String> seenConceptNamesStandardized = new HashSet<String>();
		Set<String> uniqueConceptNamesToReturnNotStandardized = new HashSet<String>();
		for (int i = 0; i < conceptNames.size(); ++i){
			String conceptName = conceptNames.get(i);
			String conceptNameStandardized = standardizeConceptName(conceptName);
			
			// We have seen this entity name before
			if (seenConceptNamesStandardized.contains(conceptNameStandardized)){
				// Do nothing
			}
			
			// Too many words
			else if (hasTooManyWords(conceptNameStandardized)){
				// Do nothing
			}
			
			// In rare cases, the concept name is empty because it only
			// contains unwanted sub strings, e.g. C2897807 NOS
			else if (conceptNameStandardized.isEmpty()){
				// Do nothing
			}
/*			
			// We have seen this entity name before, because
			// (1) We had singular form before, now we see plural, OR
			// (2) We had plural form before, now we see singular
			// Either way, take the singular form and discard plural
			// form.
			//
			// differByPlural() needs to go in front of differByAbbreviations()
			// because there are irregular plural forms that would be considered
			// abbreviations, and then the singular/plural replacement wouldn't
			// take place.
			else if (differByPlural(seenConceptNamesStandardized, conceptNameStandardized,
					uniqueConceptNamesToReturnNotStandardized, conceptName)){
				// No need to make replacement in uniqueConceptNamesToReturnNotStandardized,
				// differByPlural() takes care of it already.
				seenConceptNamesStandardized.add(conceptNameStandardized);
			}
*/			
			// We have seen this entity name before, because some word(s)
			// is now abbreviated
			else if (differByAbbreviations(seenConceptNamesStandardized, conceptNameStandardized)){
				// Do nothing
			}
			
			// We have seen this entity name before, because the word
			// order is different. We take advantage of the already
			// ordered concept names, where entity names with comma
			// and modifiers come later.
			// E.g. "branching enzyme" occurs before "enzyme, branching"
			// and it's exactly the comma version that we do not want.
			else if (differByWordOrder(seenConceptNamesStandardized, conceptNameStandardized)){
				// Do nothing
			}
			
			// We have not seen this entity name before
			else{
				uniqueConceptNamesToReturnNotStandardized.add(conceptName);
				seenConceptNamesStandardized.add(conceptNameStandardized);
			}			
		}
		
		return new ArrayList<String>(uniqueConceptNamesToReturnNotStandardized);
	}

	private static boolean differByAbbreviations(Set<String> previousConceptNames, String conceptName){
		boolean found = false;
		
		//-----------------------------------------
		// Go thru each previous concept name. When we find one
		// previous concept name that is the same with the
		// current concept name, except one word is abbreviated,
		// return true.
		//
		// E.g. Previously we've seen "lower abdomen neoplasm"
		//      Now we see "lower abdomen neopl"
		//      ==> return true
		List<String> previousConceptNamesAsList =  new ArrayList<String>(previousConceptNames);
		
		int i = 0;
		while (!found && i < previousConceptNamesAsList.size()){
			String previousConceptName = previousConceptNamesAsList.get(i);
			List<String> previousWords = Arrays.asList(StringUtils.split(previousConceptName, " "));
			
			List<String> currentWords = Arrays.asList(StringUtils.split(conceptName, " "));
			
			// Both concept names must have the same number of words
			if (previousWords.size() == currentWords.size()){
				Collections.sort(previousWords);
				Collections.sort(currentWords);
				
				//-----------------------------------------
				// Go thru each word pair. They are either identical,
				// or the current word is an abbreviation of the previous
				// word.
				int numberOfIdenticalWords = 0;
				int numberOfAbbreviations = 0;
				int numberOfWords = previousWords.size();
				for (int j = 0; j < numberOfWords; ++j){
					String previousWord = previousWords.get(j);
					String currentWord = currentWords.get(j);
					
					if (previousWord.equals(currentWord)){
						++numberOfIdenticalWords;
					}
					else if (isAbbreviationOf(currentWord, previousWord)){
						++numberOfAbbreviations;
					}
				}
				
				if (numberOfAbbreviations > 0 && numberOfIdenticalWords + numberOfAbbreviations == numberOfWords){
					found = true;
				}
			}
			
			++i;
		}
		
		return found;
	}

	private static boolean differByWordOrder(Set<String> previousConceptNames, String conceptName){
		boolean found = false;
		
		//-----------------------------------------
		// Go thru each previous concept name. When we find one
		// previous concept name that is the same with the
		// current concept name, except word order is different,
		// return true.
		//
		// E.g. Previously we've seen "branching enzyme"
		//      Now we see "enzyme, branching"
		//      ==> return true
		List<String> previousConceptNamesAsList =  new ArrayList<String>(previousConceptNames);
		
		int i = 0;
		while (!found && i < previousConceptNamesAsList.size()){
			String previousConceptName = previousConceptNamesAsList.get(i);
			List<String> previousWords = Arrays.asList(StringUtils.split(previousConceptName, " "));
			
			List<String> currentWords = Arrays.asList(StringUtils.split(conceptName, " "));
			
			// Both concept names must have the same number of words
			if (previousWords.size() == currentWords.size()){
				Collections.sort(previousWords);
				Collections.sort(currentWords);
				
				//-----------------------------------------
				// Go thru each word pair. They are all identical.
				int numberOfIdenticalWords = 0;
				int numberOfWords = previousWords.size();
				for (int j = 0; j < numberOfWords; ++j){
					String previousWord = previousWords.get(j);
					String currentWord = currentWords.get(j);
					
					if (previousWord.equals(currentWord)){
						++numberOfIdenticalWords;
					}
				}
				
				if (numberOfIdenticalWords == numberOfWords){
					found = true;
				}
			}
			
			++i;
		}
		
		return found;
	}
	
	public static void generate(Properties settings) throws Exception{
		logger.info("Generating entity names suitable for paraphrasing...");
		
		//-----------------------------------------
		// This method takes MRCONSO.RRF as the input, and outputs
		// the entity names suitable for paraphrasing. This means
		// we try to determine, for one CUI, which STR values best
		// represent the synonyms for this CUI.
		WordPluralizer.initialize();
		UnwantedSubStringsInConceptNames.initialize();
		
		//-----------------------------------------
		// Find out if we want to limit ourselves to CUIs of the DISO semantic group
		String disoOnlyAsString = settings.getProperty("bebe_paraphrase_limit_to_diso");
		if (disoOnlyAsString.equalsIgnoreCase("yes")){
			disoOnly = true;
		}
		
		//-----------------------------------------
		// Find out how many words we can have in an entity name
		maxNumberOfWords = Integer.valueOf(settings.getProperty("bebe_paraphrase_max_number_of_words_in_text_mention"));
		
		//-----------------------------------------
		// Initialize semantic type lookup, if needed
		if (disoOnly){
			LookupSemanticType.initialize(settings);
		}
		
		//-----------------------------------------
		// Get the list of CUI --> concept names
		getCuisToConceptNames(settings);

		//-----------------------------------------
		// Output the results in CSV file
		saveCuiResultsToOutputFile(settings);
		
		//-----------------------------------------
		// Output the sub strings in brackets and their counts
		//displayBracketSubStringCounts();
		
		logger.info("Done");
	}
	
	private static void getCuisToConceptNames(Properties settings) throws IOException{
		cuisToConceptNames = new HashMap<String, List<String>>();
		
		//-----------------------------------------
		// Read in the MRCONSO.RRF file
		String mrconsoFilename = settings.getProperty("umls_mrconso_filename");
		List<String> inputLines = ReadFromFile.readFileToListOfStrings(mrconsoFilename);
		
		//-----------------------------------------
		// Go thru the lines one by one, and extract the CUI and
		// concept name
		int count = 0;
		String cui = "";
		String previousCui = "";
		List<String> conceptNameCache = new ArrayList<String>();
		for (String inputLine : inputLines){
			//-----------------------------------------
			// Report interim progress
			if (count > 0 && count % 100000 == 0){
				logger.info(count + " rows in MRCONSO.RRF processed...");
			}
			
			//-----------------------------------------
			// Tokenize the string to get the CUI and concept name
			// Note to self: StringUtils.split() treats multiple empty
			// columns as a single column
			List<String> tokens = Arrays.asList(inputLine.split("\\|"));
			cui = tokens.get(0);
			String conceptName = tokens.get(14);
			
			//-----------------------------------------
			// When we are done gathering all the concept names for
			// one CUI, process the gathered cache now
			if (!previousCui.isEmpty() && !cui.equals(previousCui)){
				
				//-----------------------------------------
				// If we are only addressing CUIs in the DISO semantic
				// group, check that now
				if (wantToProcess(previousCui)){
					List<String> uniqueConceptNames = determineUniqueConceptNames(conceptNameCache);
					if (!uniqueConceptNames.isEmpty()){
						cuisToConceptNames.put(previousCui, uniqueConceptNames);
					}
				}
				
				//-----------------------------------------
				// Reset things for the next CUI
				conceptNameCache.clear();
			}
			
			//-----------------------------------------
			// Add the concept name to the cache
			previousCui = cui;
			conceptNameCache.add(conceptName);
			
			++count;
		}
		
		//-----------------------------------------
		// Don't forget to process the final CUI
		List<String> uniqueConceptNames = determineUniqueConceptNames(conceptNameCache);
		if (!uniqueConceptNames.isEmpty()){
			cuisToConceptNames.put(previousCui, uniqueConceptNames);
		}
	}
	
	private static boolean hasTooManyWords(String string){
		int numberOfWords = StringUtils.split(string, " ").length;
		return (numberOfWords > maxNumberOfWords);
	}
	
	private static boolean isAbbreviationOf(String s1, String s2){
		boolean returnBool = false;
		
		//-----------------------------------------
		// Return true if input string 1 is an abbreviation
		// of input string 2.
		
		//-----------------------------------------
		// Continue if input string 1 is a prefix of input
		// string 2
		if (s2.startsWith(s1)){
			//-----------------------------------------
			// Take a look at the suffix. When the suffix is
			// more substantial than just an "s" (e.g. for
			// singular / plural spellings), we have an
			// abbreviation.
			int lengthOfS1 = s1.length();
			String suffix = s2.substring(lengthOfS1);
			
			if (suffix.length() >= 4 ||
					(!suffix.equalsIgnoreCase("s") && !suffix.equalsIgnoreCase("es") && !suffix.equalsIgnoreCase("ing")
							&& !suffix.endsWith("ent") && !suffix.endsWith("ion") && !suffix.endsWith("ness"))){
				returnBool = true;
			}
		}
		
		return returnBool;
	}
	
	private static void saveCuiResultsToOutputFile(Properties settings) throws IOException{
		List<String> outputLines = new ArrayList<String>();
		String outputFilename = settings.getProperty("bebe_paraphrase_CUI_version_entity_name_filename");
		
		List<String> cuis = new ArrayList<String>(cuisToConceptNames.keySet());
		Collections.sort(cuis);
		
		for (String cui : cuis){
			List<String> conceptNames = cuisToConceptNames.get(cui);
			Collections.sort(conceptNames);
			
			for (int i = 0; i < conceptNames.size(); ++i){
				String conceptName = conceptNames.get(i);
				// Format: CUI [tab] version [tab] concept name
				outputLines.add(cui + "\t" + i + "\t" + conceptName);
			}
		}
		
		SaveToFile.saveListOfStringsToFile(outputLines, outputFilename);
	}
	
	private static List<String> sortByCommaSemicolon(List<String> conceptNames){
		//-----------------------------------------
		// Whenever we see a concept name with a comma or semicolon,
		// place it in a "comma" bin. At the end, concatenate
		// the no-comma and comma bins. That is, within each
		// bin, the original order is preserved.
		List<String> noComma = new ArrayList<String>();
		List<String> yesComma = new ArrayList<String>();
		
		for (String conceptName : conceptNames){
			if (conceptName.contains(",") ||
					conceptName.contains(";")){
				yesComma.add(conceptName);
			}
			else{
				noComma.add(conceptName);
			}
		}
		
		List<String> combined = new ArrayList<String>(noComma);
		combined.addAll(yesComma);
				
		return combined;
	}
	
	private static String standardizeConceptName(String conceptName){
		//-----------------------------------------
		// We want to be able to distinguish between possibly acronyms and
		// camel case English words. So don't change everything to lower case.
		// If the concept name is all caps, leave it so. Otherwise, change to
		// lower case.
		String conceptNameAltered = conceptName;
		
		if (!ProcessUpperLowerCase.moreThanCamelCase(conceptName)){
			conceptNameAltered = conceptNameAltered.toLowerCase();
		}
		
		conceptNameAltered = conceptNameAltered.replaceAll("[^a-zA-Z0-9]", " ");
		conceptNameAltered = conceptNameAltered.replaceAll("\\s+", " ");
		conceptNameAltered = conceptNameAltered.trim();
		return conceptNameAltered;
	}

	private static boolean wantToProcess(String cui){
		boolean returnBool = true;

		if (disoOnly){
			returnBool = false;
			List<String> semanticTypes = LookupSemanticType.lookup(cui);
			List<String> semanticGroups = LookupSemanticGroup.lookup(semanticTypes);

			if (semanticGroups.contains("DISO")){
				returnBool = true;
			}
		}

		return returnBool;
	}
	
	public static void main(String[] args) throws Exception{
		Properties settings = CommandLineArguments.getSettings(args);
		generate(settings);
//		UnwantedSubStringsInConceptNames.displayBracketSubStringCounts();
	}
}