/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.stringsim;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import mpii.biotm.CommandLineArguments;
import mpii.biotm.bebe.bean.CanonicalEntity;
import mpii.biotm.bebe.bean.Entity;
import mpii.biotm.bebe.bean.Span;
import mpii.biotm.bebe.bean.TextMention;
import mpii.biotm.bebe.stringsim.SimplifyString;
import mpii.biotm.util.ProcessNonAlphanumericCharacter;
import mpii.biotm.util.ProcessNonStandardEntity;
import mpii.biotm.util.ProcessStopWord;
import mpii.biotm.wrapper.umls.LookupMostCommonSenseCui;

import org.apache.commons.lang3.CharUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;

import org.javatuples.Triplet;

public class LookupEntities{
	// TODO: better handling of 's
	// E.g. "physician's assistants" --> "physician's" is not picked up
	// See pub_med_medline 14868
	
	protected static final Logger logger = Logger.getLogger(LookupEntities.class);
	
	public static final String DICTIONARY = "UMLS";
	public static final String ENTITY_SELECTION_CRITERION = "string similarity";
	private static int maxNumberOfWords = -1;
	public static final String TEXT_MENTION_SELECTION_CRITERION = "consecutive words";
	public static boolean useMostCommonSense = true;
	
	private static boolean atLeastOneFalse(List<Boolean> inputList){
		boolean foundTrue = false;
		
		int i = 0;
		while (!foundTrue && i < inputList.size()){
			if (!inputList.get(i)){
				foundTrue = true;
			}
			++i;
		}
		
		return foundTrue;
	}
	
	private static List<Triplet<Integer, Integer, String>> extractGroupsOfConsecutiveWords(String inputString){
		List<Triplet<Integer, Integer, String>> returnList = new ArrayList<Triplet<Integer, Integer, String>>();

		//-----------------------------------------
		// Get all the words from the input string, as delimited by white space and hyphen.
		// Don't call SimplifyString.simplify() here because that way we won't know which
		// chars in the middle of the string are thrown away.
		List<String> wordsInInputString = Arrays.asList(inputString.split("[ \\-]"));
		int numberOfWordsInInputString = wordsInInputString.size();
		
		//-----------------------------------------
		// Figure out the relative start and end char numbers for each word.
		// End char number is exclusive.
		List<Integer> startChars = new ArrayList<Integer>();
		List<Integer> endChars = new ArrayList<Integer>();
		int tempStartChar = 0;
		for (int i=0; i < numberOfWordsInInputString; ++i){
			startChars.add(tempStartChar);
			endChars.add(tempStartChar + wordsInInputString.get(i).length());
			tempStartChar += wordsInInputString.get(i).length() + 1; // The +1 is the space between words
		}
		
		//-----------------------------------------
		// Go thru the words again. Adjust the start and end char positions
		// to exclude brackets, trailing full stop, and the like.
		for (int wordNumber = 0; wordNumber < wordsInInputString.size(); ++wordNumber){
			String word = wordsInInputString.get(wordNumber);
			
			Triplet<String, Integer, Integer> tempTriplet = SimplifyString.simplify(word);
			int numberOfCharsThrownAwayAtStart = tempTriplet.getValue1();
			int numberOfCharsThrownAwayAtEnd = tempTriplet.getValue2();
			
			if (numberOfCharsThrownAwayAtStart != 0 || numberOfCharsThrownAwayAtEnd != 0){
				
				if (numberOfCharsThrownAwayAtStart != 0){
					int originalStartChar = startChars.get(wordNumber);
					startChars.set(wordNumber, originalStartChar + numberOfCharsThrownAwayAtStart);
				}
				
				if (numberOfCharsThrownAwayAtEnd != 0){
					int originalEndChar = endChars.get(wordNumber);
					endChars.set(wordNumber, originalEndChar - numberOfCharsThrownAwayAtEnd);
				}
				
				wordsInInputString.set(wordNumber, tempTriplet.getValue0());
			}
		}
		
		//-----------------------------------------
		// Go thru the words one more time. Take note of which ones are stop words,
		// and which ones are empty words.
		List<Boolean> isStopWord = new ArrayList<Boolean>();
		List<Boolean> isEmptyWord = new ArrayList<Boolean>();
		for (int wordNumber = 0; wordNumber < wordsInInputString.size(); ++wordNumber){
			String word = wordsInInputString.get(wordNumber);
			
			if (ProcessStopWord.isStopWord(word)){
				isStopWord.add(true);
			}
			else{
				isStopWord.add(false);
			}
			
			if (word.isEmpty()){
				isEmptyWord.add(true);
			}
			else{
				isEmptyWord.add(false);
			}
		}
		
		//-----------------------------------------
		// As we consider groups of consecutive words, we may reduce a longer
		// group to a shorter one already processed before.
		// E.g. We first get "ABC"
		//      Then later "ABC &" is shortened to "ABC"
		// We don't want to look up "ABC" twice. So use a set to prevent
		// duplicates.
		Set<Span> characterSpans = new HashSet<Span>();
		Map<Span, String> spansToStrings = new HashMap<Span, String>();
		
		//-----------------------------------------
		// Go thru each possible number of consecutive words
		for (int numberOfWords = 1; numberOfWords <= maxNumberOfWords; ++numberOfWords){
			
			//-----------------------------------------
			// Go thru each possible starting word position
			for (int startWord = 0; startWord < numberOfWordsInInputString - numberOfWords + 1; ++startWord){

				//-----------------------------------------
				// Focus on the group of consecutive words
				List<String> simplifiedWords = wordsInInputString.subList(startWord, startWord + numberOfWords);
				String simplifiedWordsAsString = StringUtils.join(simplifiedWords, " ");
				simplifiedWordsAsString = simplifiedWordsAsString.replaceAll("\\s+", " "); // Collapse consecutive spaces
				int startCharOfFirstWord = startChars.get(startWord);
				int endCharOfLastWord = endChars.get(startWord + numberOfWords - 1);
				
				String originalSubString = inputString.substring(startCharOfFirstWord, endCharOfLastWord);
				
				//-----------------------------------------
				// Use this group of consecutive words if:
				// - The group does not start with a stop/empty word
				// - The group does not end with a stop/empty word
				// - The group contains at least one non-stop/non-empty word
				// - The combined string has length at least 3 characters
				// - There is no punctuation in the middle of the combined string
				//   E.g. Do not use "something, something else", because a
				//        comma is in the middle
				if (!isStopWord.get(startWord) &&
						!isEmptyWord.get(startWord) &&
						!isStopWord.get(startWord + numberOfWords - 1) &&
						!isEmptyWord.get(startWord + numberOfWords - 1) &&
						atLeastOneFalse(isStopWord.subList(startWord, startWord + numberOfWords)) &&
						atLeastOneFalse(isEmptyWord.subList(startWord, startWord + numberOfWords)) &&
						simplifiedWordsAsString.length() >= 3 &&
						!hasPunctuationInMiddle(originalSubString)){

					//-----------------------------------------
					// Accept the group of words
					Span tempSpan = new Span(startCharOfFirstWord, endCharOfLastWord);
					characterSpans.add(tempSpan);
					spansToStrings.put(tempSpan, originalSubString);

				} // if the group of words are acceptable
			} // for each possible starting word position
		} // for each possible number of consecutive words

		//-----------------------------------------
		// Construct the return value
		List<Span> orderedCharacterSpans = new ArrayList<Span>(characterSpans);
		Collections.sort(orderedCharacterSpans);
		
		for (Span characterSpan : orderedCharacterSpans){
			returnList.add(new Triplet<Integer, Integer, String>(characterSpan.getStart(),
					characterSpan.getEnd(), spansToStrings.get(characterSpan)));
		}
		
		return returnList;
	}
	
	private static boolean hasPunctuationInMiddle(String inputString){
		//-----------------------------------------
		// Return true if there is a punctuation char in the middle
		// of the input string, i.e. not as leading or trailing char.
		// E.g. ", something." --> false
		//      "something, something else" --> true
		//
		// By "punctuation" I mean any ASCII non-alphanumeric char
		// except hyphen, full stop and apostrophe.
		//
		// Note to self: Full stop is acceptable because the input
		// string is assumed to come from within one sentence, i.e.
		// the input string never crosses a sentence boundary.
		boolean returnBool = false;
		
		String copyString = inputString;
		
		//-----------------------------------------
		// Trim leading and trailing non-alphanumeric chars
		copyString = ProcessNonAlphanumericCharacter.trimLeadingTrailing(copyString);
		
		//-----------------------------------------
		// Now see if there are any punctuation chars within
		// start and end
		if (!copyString.isEmpty()){
			int i = 0;
			while (!returnBool && i < copyString.length()){
				char c = copyString.charAt(i);
				if (CharUtils.isAscii(c) &&
						!CharUtils.isAsciiAlphanumeric(c) &&
						!ProcessNonStandardEntity.isSpace(c) &&
						c != '.' &&
						c != '\'' &&
						c != '-'){
					returnBool = true;
				}
				++i;
			}
		}
		
		return returnBool;
	}
	
	public static void initialize(Properties settings) throws IOException{
		maxNumberOfWords = Integer.valueOf(settings.getProperty("bebe_string_sim_max_number_of_consecutive_words"));
		LookupMinHash.initialize(settings);
		ProcessCharacterTrigram.initializeByRestoringData(settings);
	}
	
	private static List<TextMention> lookupGroupsOfConsecutiveWordsToMakeTextMentions(
			List<Triplet<Integer, Integer, String>> groupsOfConsecutiveWords, int sentenceNumber){
		List<TextMention> returnList = new ArrayList<TextMention>();
		
		//-----------------------------------------
		// Go thru each group of consecutive words
		for (int groupNumber = 0; groupNumber < groupsOfConsecutiveWords.size(); ++groupNumber){

			//-----------------------------------------
			// Get the groups start/end char positions, and the corresponding string
			int startChar = groupsOfConsecutiveWords.get(groupNumber).getValue0();
			int endChar = groupsOfConsecutiveWords.get(groupNumber).getValue1();
			String string = groupsOfConsecutiveWords.get(groupNumber).getValue2();

			//-----------------------------------------
			// Use MinHash to look up canonical entities
			List<Triplet<String, String, Double>> matches = LookupMinHash.lookup(string);

			//-----------------------------------------
			// When there is a match, transform it into a text mention,
			// and add to the returnlist
			if (matches.size() > 0){
				List<Span> charSpans = new ArrayList<Span>();
				charSpans.add(new Span(startChar, endChar));
				
				List<String> selectionCriteria = new ArrayList<String>();
				selectionCriteria.add(TEXT_MENTION_SELECTION_CRITERION);
				
				TextMention textMention = new TextMention(sentenceNumber, charSpans, string, selectionCriteria,
						new ArrayList<Entity>());
						
				//-----------------------------------------
				// Put all matched entities into the text mention
				for (Triplet<String, String, Double> match : matches){
					CanonicalEntity canonicalEntity = new CanonicalEntity(DICTIONARY,
							match.getValue0(), match.getValue1(), match.getValue2(), ENTITY_SELECTION_CRITERION);
					textMention.addEntity(canonicalEntity);
				}
				Collections.sort(textMention.getEntities());
				returnList.add(textMention);
			} // if MinHash lookup finds some match(es)
		} // for each group of consecutive words
		
		return returnList;
	}
	
	public static List<TextMention> process(String inputString){		
		//-----------------------------------------
		// When we don't know the sentence number, use the
		// dummy value of -1. The code that uses this method
		// should not use the sentence number in any way. We
		// need it here because the TextMention object needs
		// a denormalized field of sentence number.
		int sentenceNumber = -1;

		//-----------------------------------------
		// Identify all the groups of consecutive words we want to work on
		List<Triplet<Integer, Integer, String>> groupsOfConsecutiveWords = extractGroupsOfConsecutiveWords(inputString);

		//-----------------------------------------
		// Get the text mentions via MinHash lookup
		List<TextMention> textMentions = lookupGroupsOfConsecutiveWordsToMakeTextMentions(
				groupsOfConsecutiveWords, sentenceNumber);

		return textMentions;
	}
	
	/** Quick test */
	public static void main(String[] args) throws Exception{
		//-----------------------------------------
		// Set up test data
		List<String> sentences = new ArrayList<String>();
		sentences.add("My horoscope says I have diabetes, oh no!");
		sentences.add("But I thought you have high blood pressure?");
		sentences.add("Oh well, now I have both have diabetes and high blood pressure.");
		
		//-----------------------------------------
		// Initialization
		Properties settings = CommandLineArguments.getSettings(args);
		initialize(settings);
		
		if (useMostCommonSense){
			LookupMostCommonSenseCui.initialize(settings);
		}
		
		//-----------------------------------------
		// Go thru each sentence
		for (int sentenceNumber = 0; sentenceNumber < sentences.size(); ++sentenceNumber){
			String sentence = sentences.get(sentenceNumber);
			
			//-----------------------------------------
			// Identify all the groups of consecutive words we want to work on
			List<Triplet<Integer, Integer, String>> groupsOfConsecutiveWords = extractGroupsOfConsecutiveWords(sentence);
			
			//-----------------------------------------
			// Get the text mentions via MinHash lookup
			List<TextMention> textMentions = lookupGroupsOfConsecutiveWordsToMakeTextMentions(
					groupsOfConsecutiveWords, sentenceNumber);

			//-----------------------------------------
			// Optional step: Use most common sense to "disambiguate"
			if (useMostCommonSense){
				for (TextMention textMention : textMentions){
					String conceptName = textMention.getText();
					String mostCommonSenseCui = LookupMostCommonSenseCui.lookupReturnCui(conceptName);
					if (!mostCommonSenseCui.isEmpty()){
						List<CanonicalEntity> entities = textMention.getCanonicalEntities();
						int mostCommonSenseEntityIndex = -1;
						int i = 0;
						boolean found = false;
						while (!found && i < entities.size()){
							String cuiOfThisEntity = entities.get(i).getId();
							if (mostCommonSenseCui.equals(cuiOfThisEntity)){
								found = true;
								mostCommonSenseEntityIndex = i;
							}
							++i;
						}
						
						if (mostCommonSenseEntityIndex >= 0){
							//-----------------------------------------
							// Create a list with only one entity, the
							// one found to be of the most common sense
							List<Entity> onlyOneEntity = new ArrayList<Entity>();
							onlyOneEntity.add(entities.get(mostCommonSenseEntityIndex));
							textMention.setEntities(onlyOneEntity);
						}
					}
				} // for each text mention
			}
			
			//-----------------------------------------
			// Print out the results
			logger.debug("--------------- Sentence " + sentenceNumber + " ---------------");
			for (int textMentionNumber = 0; textMentionNumber < textMentions.size(); ++textMentionNumber){
				logger.debug("----- Text Mention " + textMentionNumber + " -----");
				logger.debug(textMentions.get(textMentionNumber));
			}
		}// for each sentence
	}
}