/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.stringsim;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import mpii.biotm.bebe.bean.MinHashTableDefinition;
import mpii.biotm.bebe.stringsim.ProcessCharacterTrigram;
import mpii.biotm.util.ReadFromFile;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.MutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

import org.javatuples.Triplet;

public class LookupMinHash{
	//-----------------------------------------
	// How to use:
	// - call initialize()
	// - call lookup() however many times you like
	// - call close()
	
	protected static final Logger logger = Logger.getLogger(LookupMinHash.class);
	
	private static Map<Integer, String> conceptNames = null;
	private static Map<Integer, String> cuis = null;
	private static double jaccardThreshold = (double) 999;
	private static List<MinHashTableDefinition> minHashTableDefinitions = null;
	private static List<Map<Integer, Set<Integer>>> minHashTables = null;
	private static int numberOfHashFunctionsPerTable = -1;
	private static int numberOfMinHashTables = -1;
	private static Map<String, String> params = null;
	private static int prime = -1;
	
	private static double calculateJaccardSimilarity(Map<Integer, Integer> input1, Map<Integer, Integer> input2){
		//-----------------------------------------
		// Notice that we want to do Jaccard similarity in a multiset
		// setting.
		// E.g. string 1 has "abc" 3 times
		//      string 2 has "abc" 2 times
		//      Their union contains 3 "abc"s
		//      Their intersection contains 2 "abc"s
		int intersection = 0;
		int union = 0;
		
		//-----------------------------------------
		// Go thru each the trigram ID in input1, and try to see if such a
		// trigram ID also appears in input2. Remove processed trigram IDs
		// from input2 as we go.
		for (int trigramId1 : input1.keySet()){
			
			//-----------------------------------------
			// See if this trigram ID also appears in input2.
			//
			// Case 1: trigram ID 1 also appears in input2
			if (input2.containsKey(trigramId1)){
				int count1 = input1.get(trigramId1);
				int count2 = input2.get(trigramId1);
				
				//-----------------------------------------
				// Case 1a: This trigram has the same count in both input1
				// and input2. All the count goes to both intersection and
				// union.
				if (count1 == count2){
					intersection += count1;
					union += count1;
				}
				
				//-----------------------------------------
				// Case 1b: This trigram has different counts in input1 and
				// input2. The smaller count goes to intersection. The
				// larger count goes to union.
				else{
					if (count1 > count2){
						intersection += count2;
						union += count1;
					}
					else{
						intersection += count1;
						union += count2;
					}
				}
				
				input2.remove(trigramId1);
			}
			// Case 2: trigram ID 1 does not appear in input2.
			// All of trigram ID 1's count goes to union.
			else{
				union += input1.get(trigramId1);
			}
		}
		
		//-----------------------------------------
		// Any trigram IDs remaining in input2 have their counts going to union.
		for (int trigramId2 : input2.keySet()){
			union += input2.get(trigramId2);
		}

		return (double) intersection / union;
	}
	
	private static Set<Pair<String, String>> getConceptsWithHashCodeForMinHashTable(int minHashTableNumber, int hashCode){
		Set<Pair<String, String>> returnSet = new HashSet<Pair<String, String>>();
		
		//-----------------------------------------
		// First get all the concept IDs under the same hash code. These concept IDs
		// are not CUIs; they are the primary key for the database with CUIs and
		// concept names.
		Set<Integer> conceptIds = minHashTables.get(minHashTableNumber).get(hashCode);
		
		//-----------------------------------------
		// Now we need to get the CUI and concept name pointed to by each
		// of the concept ID.
		if (conceptIds != null){
			for (int conceptId : conceptIds){
				String cui = cuis.get(conceptId);
				String conceptName = conceptNames.get(conceptId);
				returnSet.add(new MutablePair<String, String>(cui, conceptName));
			}
		}
		
		return returnSet;
	}
	
	public static void initialize(Properties settings) throws IOException{
		//-----------------------------------------
		// Initialize parameters
		String csvDir = settings.getProperty("bebe_string_sim_csv_directory");
		jaccardThreshold = Double.valueOf(settings.getProperty("bebe_string_sim_jaccard_similarity_threshold"));
		
		//-----------------------------------------
		// Restore the saved data from files
		logger.info("Restoring MinHash data from CSV files...");
		restoreDataFromCsvFiles(csvDir);
		
		//-----------------------------------------
		// Initialize params stored in the hash map
		initializeMinHashParams();
		
		//-----------------------------------------
		// Restore the saved trigram data from file
		ProcessCharacterTrigram.initializeByRestoringData(settings);
		
		logger.info("Done");
	}
	
	private static void initializeMinHashParams(){
		numberOfMinHashTables = Integer.valueOf(params.get("number of MinHash tables"));
		numberOfHashFunctionsPerTable = Integer.valueOf(params.get("number of hash functions per MinHash table"));
		prime = Integer.valueOf(params.get("prime"));
		
		//-----------------------------------------
		// Get the A and B coeffs stored params, then create the MinHash
		// table definitions from them
		minHashTableDefinitions = new ArrayList<MinHashTableDefinition>();
		for (int i=0; i < numberOfMinHashTables; ++i){
			List<Integer> aCoeffs = new ArrayList<Integer>();
			List<Integer> bCoeffs = new ArrayList<Integer>();
			
			for (int j=0; j < numberOfHashFunctionsPerTable; ++j){
				aCoeffs.add(Integer.valueOf(params.get("MinHash table " + i + " hash function " + j + " coeff a")));
				bCoeffs.add(Integer.valueOf(params.get("MinHash table " + i + " hash function " + j + " coeff b")));
			}
			
			//-----------------------------------------
			// Construct the MinHash table
			MinHashTableDefinition minHashTableDefinition = new MinHashTableDefinition(prime, aCoeffs, bCoeffs);
			minHashTableDefinitions.add(minHashTableDefinition);
		}
	}
	
	// Given a new string, apply MinHash to it, look up from Berkeley DB,
	// and return a set of CUI / UMLS concept name pairs.
	public static List<Triplet<String, String, Double>> lookup(String inputString){
		Set<Pair<String, String>> matchingConcepts = new HashSet<Pair<String, String>>();
		
		//-----------------------------------------
		// Find the bucket (if any) from each MinHash table, then
		// take the union of buckets
		//
		// Note to self: Each bucket is identified by the input
		// string's hash code aka bucket ID. Each bucket contains
		// a list of matching concept names.
		
		//-----------------------------------------
		// Transform the inputString into trigram info
		Pair<Integer, Map<Integer, Integer>> tempPair = ProcessCharacterTrigram.processInput(inputString);
		Map<Integer, Integer> trigramIdsToCounts = tempPair.getRight();
		
		//-----------------------------------------
		// Get the hash code aka bucket ID for each MinHash table,
		// then look up from Berkeley DB the concepts in that bucket
		// for that MinHash table. Take the union of all the concepts
		// from all the MinHash tables.
		for (int i=0; i < numberOfMinHashTables; ++i){
			int hashCode = minHashTableDefinitions.get(i).getHashCode(trigramIdsToCounts);
			Set<Pair<String, String>> matchingConceptsFromOneMinHashTable = getConceptsWithHashCodeForMinHashTable(i, hashCode);
			
			//-----------------------------------------
			// Store all the matching concepts
			matchingConcepts.addAll(matchingConceptsFromOneMinHashTable);
		}
		
		//-----------------------------------------
		// Filter out false positives based on character-level trigram
		// Jaccard similarity score. When score is below a threshold,
		// throw away that concept as false positive.
		List<Triplet<String, String, Double>> filteredMatchingConcepts =
				throwAwayBelowJaccardThreshold(trigramIdsToCounts, matchingConcepts);
		
		// TODO: Combine multiple matches sharing same CUI
		
		return filteredMatchingConcepts;
	}
	
	private static void restoreDataFromCsvFiles(String csvDir) throws IOException{
		params = new HashMap<String, String>();
		
		List<String> paramsList = ReadFromFile.readFileToListOfStrings(csvDir + "Params.csv");
		for (String line : paramsList){
			int tabPosition = line.indexOf("\t");
			String param = line.substring(0, tabPosition);
			String value = line.substring(tabPosition + 1);
			params.put(param, value);
		}
		
		conceptNames = new HashMap<Integer, String>();
		List<String> conceptNamesList = ReadFromFile.readFileToListOfStrings(csvDir + "ConceptNames.csv");
		for (String line : conceptNamesList){
			int tabPosition = line.indexOf("\t");
			int conceptId = Integer.valueOf(line.substring(0, tabPosition));
			String conceptName = line.substring(tabPosition + 1);
			conceptNames.put(conceptId, conceptName);
		}
		
		cuis = new HashMap<Integer, String>();
		List<String> cuisList = ReadFromFile.readFileToListOfStrings(csvDir + "CUIs.csv");
		for (String line : cuisList){
			int tabPosition = line.indexOf("\t");
			int conceptId = Integer.valueOf(line.substring(0, tabPosition));
			String cui = line.substring(tabPosition + 1);
			cuis.put(conceptId, cui);
		}
		
		minHashTables = new ArrayList<Map<Integer, Set<Integer>>>();
		numberOfMinHashTables = Integer.valueOf(params.get("number of MinHash tables"));
		for (int i=0; i < numberOfMinHashTables; ++i){
			Map<Integer, Set<Integer>> minHashTable = new HashMap<Integer, Set<Integer>>();
			List<String> minHashTableList = ReadFromFile.readFileToListOfStrings(csvDir + "MinHashTable" + i + ".csv");
			for (String line : minHashTableList){
				int tabPosition = line.indexOf("\t");
				int hashCode = Integer.valueOf(line.substring(0, tabPosition));
				String conceptIdsAsString = line.substring(tabPosition + 1);
				Set<Integer> conceptIds = new HashSet<Integer>();
				for (String conceptIdAsString : StringUtils.split(conceptIdsAsString, "\t")){
					conceptIds.add(Integer.valueOf(conceptIdAsString));
				}
				minHashTable.put(hashCode, conceptIds);
			}
			minHashTables.add(minHashTable);
		}
	}
	
	private static List<Triplet<String, String, Double>> throwAwayBelowJaccardThreshold(Map<Integer, Integer> trigramIdsToCounts,
			Set<Pair<String, String>> candidates){
		List<Triplet<String, String, Double>> returnSet = new ArrayList<Triplet<String, String, Double>>();
		
		//-----------------------------------------
		// Go thru each candidate, which is a CUI / concept name pair.
		// Compare the concept name to the string being looked up via
		// character trigram Jaccard similarity score. Throw away
		// the candidate if the score falls below threshold.		
		for (Pair<String, String> candidate : candidates){
			//-----------------------------------------
			// Similarly transform the candidate's concept name as a multiset
			// of trigram IDs.
			String conceptName = candidate.getRight();
			Pair<Integer, Map<Integer, Integer>> tempPair = ProcessCharacterTrigram.processInput(conceptName);
			Map<Integer, Integer> trigramIdsToCountsCandidate = tempPair.getRight();
			
			//-----------------------------------------
			// Now we are ready to compute the Jaccard similarity
			double jaccardSimilarity = calculateJaccardSimilarity(trigramIdsToCounts, trigramIdsToCountsCandidate);
			
			//-----------------------------------------
			// Keep the candidate if Jaccard similarity > threshold
			if (jaccardSimilarity >= jaccardThreshold){
				returnSet.add(new Triplet<String, String, Double>(candidate.getLeft(),
						candidate.getRight(), jaccardSimilarity));
			}
		}
		
		return returnSet;
	}
}