/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.stringsim;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.ConcurrentHashMap;

import mpii.biotm.bebe.bean.CharacterTrigram;
import mpii.biotm.bebe.stringsim.SimplifyString;
import mpii.biotm.util.ReadFromFile;
import mpii.biotm.util.SaveToFile;

import org.apache.commons.lang3.tuple.MutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;
import org.javatuples.Triplet;

public class ProcessCharacterTrigram{
	//-----------------------------------------
	// Usage:
	// 1. call the desired initialize...()
	// 2. call processInput() however many times you need
	// 3. call saveTrigramsToIdsAsCsv() if desired
	
	protected static final Logger logger = Logger.getLogger(ProcessCharacterTrigram.class);
	
	private static boolean allowAddingNewTrigramIds = false;
	private static int nextTrigramId = -1;
	
	// Key = character trigram, value = integer ID
	private static ConcurrentHashMap<String, Integer> trigramsToIds = null;
	
	public static List<CharacterTrigram> getTrigrams(String inputString){
		List<CharacterTrigram> returnList = new ArrayList<CharacterTrigram>();
		
		//-----------------------------------------
		// Don't segment the input string into words. But pad the beginning and end of
		// the input string with 2 spaces. Then simply take all character trigrams.
		
		//-----------------------------------------
		// If the input string has only one or two letters, do not pad it,
		// and simply use the entire input string as the "trigram".
		//
		// Notice that if input string is an empty string, return an empty list.
		int length = inputString.length();
		if (length == 1 || length == 2){
			insertToCharacterTrigramList(inputString, returnList);
		}
		
		//-----------------------------------------
		// Input string has 3 or more letters. Replace hyphen, comma,
		// full stop, apostrophe with space. Do the padding. Then get
		// all the trigrams. Return the results alphabetically.
		else{
			Triplet<String, Integer, Integer> triplet = SimplifyString.simplify(inputString);
			inputString = "  " + triplet.getValue0() + "  ";
			
			length = inputString.length();
			for (int i=0; i < length - 2; ++i){
				String trigram = inputString.substring(i, i + 3);
				insertToCharacterTrigramList(trigram, returnList);
			}
			
			//-----------------------------------------
			// Sort the return list
			Collections.sort(returnList);
		}
		
		return returnList;
	}

	/* Use this method when looking up from a previously built hash saved in a CSV file. */
	public static void initializeByRestoringData(Properties settings) throws IOException{
		allowAddingNewTrigramIds = false;		
		trigramsToIds = new ConcurrentHashMap<String, Integer>();
		
		String filename = settings.getProperty("bebe_string_sim_csv_directory") + "Trigrams.csv";
		List<String> trigramsToIdsList = ReadFromFile.readFileToListOfStrings(filename);
		for (String line : trigramsToIdsList){
			int tabPosition = line.indexOf("\t");
			String trigram = line.substring(0, tabPosition);
			int id = Integer.valueOf(line.substring(tabPosition + 1));
			trigramsToIds.put(trigram, id);
		}
	}
	
	/* Use this method when building a new hash. */
	public static void initializeForCreatingData() throws IOException{
		allowAddingNewTrigramIds = true;		
		trigramsToIds = new ConcurrentHashMap<String, Integer>();
	}
	
	private static void insertToCharacterTrigramList(String trigramLetters, List<CharacterTrigram> characterTrigrams){
		//-----------------------------------------
		// Make a new chracter trigram out of the input letters. The
		// count at this point is one.
		//
		// Notice that when creating an instance of CharacterTrigram,
		// all letters are automatically converted to lower case
		CharacterTrigram tempTrigram = new CharacterTrigram(trigramLetters, 1);
		
		//-----------------------------------------
		// Check the list to see if the input letters already exist.
		// If yes, increment the count.
		// If not, add a new entry with a count of 1.
		// Weight is arbitrarily set to 1 for all entries.
		if (characterTrigrams.contains(tempTrigram)){
			int index = characterTrigrams.indexOf(tempTrigram);
			int tempCount = characterTrigrams.get(index).getCount();
			++tempCount;
			characterTrigrams.get(index).setCount(tempCount);
		}
		else{
			characterTrigrams.add(tempTrigram);
		}
	}
	
	public static Pair<Integer, Map<Integer, Integer>> processInput(String inputString){
		int totalCount = 0;
		Map<Integer, Integer> trigramIdsToCounts = new HashMap<Integer, Integer>();
		
		//-----------------------------------------
		// Split the input string into character trigrams
		List<CharacterTrigram> trigrams = getTrigrams(inputString);
		
		//-----------------------------------------
		// trigram ID --> count lookup, and at the same time
		// find the total count. Accumulate new trigrams if
		// inserts are allowed.
		for (CharacterTrigram trigram : trigrams){
			String trigramChars = trigram.getChars();
			int trigramId = -1;
			
			//-----------------------------------------
			// During the hash building phase, we want to add never-before-seen
			// trigrams into the map. But during the lookup phase, we don't want
			// to alter the set of existing trigrams any more.
			if (trigramsToIds.containsKey(trigramChars)){
				trigramId = trigramsToIds.get(trigramChars);
			}
			else if (allowAddingNewTrigramIds){
				trigramsToIds.putIfAbsent(trigramChars, ++nextTrigramId);
				trigramId = nextTrigramId;
			}

			//-----------------------------------------
			// Register this trigram only if it already exists in the lookup
			if (trigramId != -1){
				int count = trigram.getCount();
				trigramIdsToCounts.put(trigramId, count);
				totalCount += count;
			}
		}
		
		return new MutablePair<Integer, Map<Integer, Integer>>(totalCount, trigramIdsToCounts);
	}
	
	public static void saveTrigramsToIdsAsCsv(String filename) throws IOException{
		List<String> trigramsToIdsList = new ArrayList<String>();
		for (String trigram : trigramsToIds.keySet()){
			trigramsToIdsList.add(trigram + "\t" + trigramsToIds.get(trigram));
		}
		SaveToFile.saveListOfStringsToFile(trigramsToIdsList, filename);
	}
}