/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.stringsim;

import org.apache.commons.lang3.CharUtils;
import org.apache.log4j.Logger;
import org.javatuples.Triplet;

public class SimplifyString{
	
	protected static final Logger logger = Logger.getLogger(SimplifyString.class);
	
	public static Triplet<String, Integer, Integer> simplify(String string){
		//-----------------------------------------
		// Simplify the characters in a string to make it more
		// amenable to a string similarity-based matching scheme.
		//
		// Notice if the input is a single word, by the end of this method,
		// the return word may contain spaces.
		// E.g. "Parkinson's" ==> "Parkinson s"
		//
		// Specifically, this method:
		// (1) Figures out how many leading and trailing chars we don't want
		//     when looking up MinHash. Typical examples: the brackets in
		//     "(ABC)"; the period in "end of sentence.".
		// (2) Replaces certain chars (currently hyphen, comma, apostrophe)
		//     with spaces. Then consecutive spaces are collapsed into a
		//     single space. Note that when this collapse happens, the return
		//     char counts are not affected.
		//     E.g. input string:  "abc-'3-protein)", with a span of 15 chars
		//          output string: "abc 3 protein", with 13 actual chars, but
		//                         the end char is only diminished by one to
		//                         account for the final ")"
		
		//-----------------------------------------
		// Return values:
		// String: The modified word
		// First integer: how many chars thrown away at the start
		// Second integer: how many chars thrown away at the end
		
		//-----------------------------------------
		// Replace hyphen, comma, full stop, apostrophe with space
		string = string.replaceAll("[-,'.]", " ");
		
		//-----------------------------------------
		// Trim leading and trailing non-alphanumeric chars, and keep track of
		// how many chars we have trimmed. Notice that we don't trim non-ASCII
		// chars.
		// TODO: We end up with things like "migratory behavior (animal" and "officers (commissioned"... change?
		// TODO: We convert "? pregnancy" into "pregnancy" and that should not happen
		int length = string.length();
		
		int numberOfCharsThrownAwayAtStart = 0;
		while (length > 0 && CharUtils.isAscii(string.charAt(0)) && !CharUtils.isAsciiAlphanumeric(string.charAt(0))){
			string = string.substring(1);
			--length;
			++numberOfCharsThrownAwayAtStart;
		}
		
		int numberOfCharsThrownAwayAtEnd = 0;
		while (length > 0 && CharUtils.isAscii(string.charAt(length - 1)) && !CharUtils.isAsciiAlphanumeric(string.charAt(length - 1))){
			string = string.substring(0, length - 1);
			--length;
			++numberOfCharsThrownAwayAtEnd;
		}
		
		//-----------------------------------------
		// Collapse consecutive spaces
		string = string.replaceAll("\\s+", " ");
		
		return new Triplet<String, Integer, Integer>(string, numberOfCharsThrownAwayAtStart, numberOfCharsThrownAwayAtEnd);
	}
}