/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.stringsim.precompute;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import mpii.biotm.CommandLineArguments;
import mpii.biotm.bebe.bean.MinHashTableDefinition;
import mpii.biotm.bebe.stringsim.ProcessCharacterTrigram;
import mpii.biotm.bebe.stringsim.SimplifyString;
import mpii.biotm.util.RandomNumber;
import mpii.biotm.util.ReadFromFile;
import mpii.biotm.util.SaveToFile;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;
import org.javatuples.Triplet;

public class BuildCsvFiles{
	
	protected static final Logger logger = Logger.getLogger(BuildCsvFiles.class);

	private static Map<Integer, String> conceptIdsToConceptNames = null;
	private static Map<Integer, String> conceptIdsToCuis = null;
	private static List<MinHashTableDefinition> minHashTableDefinitions = null;
	private static List<Map<Integer, Set<Integer>>> minHashTables = null;
	private static int numberOfMinHashTables = -1;
	private static int numberOfHashFunctionsToGenerate = -1;
	private static int numberOfHashFunctionsToUse = -1;
	private static int numberOfDimensions = -1;
	private static Map<String, String> params = null;
	private static List<Integer> poolOfACoeffs = null;
	private static List<Integer> poolOfBCoeffs = null;
	private static int prime = -1;
	
	private static Map<String, List<String>> extractFromMrconso(String mrconsoFilename,
			int maxNumberOfChars, int maxNumberOfWords) throws IOException{
		Map<String, List<String>> cuisToConceptNames = null;
		
		//-----------------------------------------
		// Read in the MRCONSO.RRF file
		List<String> inputLines = ReadFromFile.readFileToListOfStrings(mrconsoFilename);
		
		//-----------------------------------------
		// For each CUI, get all the unique concept names
		cuisToConceptNames = getUniqueConceptNames(inputLines);
		
		//-----------------------------------------
		// For each CUI, filter out concept names that are not suitable
		// for a string similarity-based method
		cuisToConceptNames = filterOutUnsuitableConceptNames(cuisToConceptNames, maxNumberOfChars, maxNumberOfWords);
				
		return cuisToConceptNames;
	}

	private static Map<String, List<String>> filterOutUnsuitableConceptNames(Map<String, List<String>> cuisToConceptNames,
			int maxNumberOfChars, int maxNumberOfWords){
		//-----------------------------------------
		// Go thru each CUI, and remove concept names not suitable
		// for a string similarity-based method
		for (String cui : cuisToConceptNames.keySet()){
			List<String> conceptNames = cuisToConceptNames.get(cui);
			
			int i = 0;
			while (i < conceptNames.size()){
				String conceptName = conceptNames.get(i);
				int numberOfWords = conceptName.split("\\W").length;
				
				// Conditions for being a suitable concept name:
				// (1) 100 or fewer characters long
				// (2) Not more than max number of words, as
				//     specified in settings file
				if (conceptName.length() <= maxNumberOfChars &&
						numberOfWords <= maxNumberOfWords){
					++i;
				}
				else{
					conceptNames.remove(i);
				}
			}
		}
		
		return cuisToConceptNames;
	}
	
	private static int getPrimeBiggerThan(int i){
		int returnInt = i;
		
		while (!isPrime(returnInt)){
			++returnInt;
		}
		
		return returnInt;
	}
	
	// Lower and upper bounds are inclusive
	// Assumption: RandomNumber is already initialized, i.e. do NOT reset the sequence of random numbers!
	private static List<Integer> getRandomNumbersBetween(int min, int max, int numberOfRandomsRequired){
		List<Integer> returnList = new ArrayList<Integer>();

		for (int i=0; i < numberOfRandomsRequired; ++i){
			returnList.add(RandomNumber.getRandomNumberInRange(min, max));
		}

		return returnList;
	}

	private static Map<String, List<String>> getUniqueConceptNames(List<String> inputLines){
		Map<String, List<String>> cuisToConceptNames = new HashMap<String, List<String>>();
		
		//-----------------------------------------
		// Go thru the lines one by one, and accumulate all the unique
		// concept names for one CUI. Concept names are case insensitive.
		for (String inputLine : inputLines){
			
			//-----------------------------------------
			// Tokenize the string to get the CUI and concept name
			// Note to self: StringUtils.split() treats multiple empty
			// columns as a single column
			List<String> tokens = Arrays.asList(inputLine.split("\\|"));
			String cui = tokens.get(0);
			String conceptName = tokens.get(14).toLowerCase();
			
			//-----------------------------------------
			// Simplify the concept name so we get rid of e.g. commas, apostrophes
			Triplet<String, Integer, Integer> tempTriplet = SimplifyString.simplify(conceptName);
			conceptName = tempTriplet.getValue0();

			//-----------------------------------------
			// If the CUI does not yet have this concept name cached,
			// do it.
			if (!conceptName.isEmpty()){
				if (!cuisToConceptNames.containsKey(cui)){
					cuisToConceptNames.put(cui, new ArrayList<String>());
				}
				if (!cuisToConceptNames.get(cui).contains(conceptName)){
					cuisToConceptNames.get(cui).add(conceptName);
				}
			}
			else{
				logger.warn("Empty concept name from " + cui + ", input line: " + inputLine);
			}
		}
		
		return cuisToConceptNames;
	}
	
	private static void hashConceptNames(Map<String, List<String>> cuisToConceptNames){
		//-----------------------------------------
		// A concept ID is not a CUI.
		// Each unique CUI / concept name pair gets its own concept ID.
		int nextConceptId = -1;
		
		//-----------------------------------------
		// Go thru each CUI
		for (String cui : cuisToConceptNames.keySet()){
			if (cuisToConceptNames.get(cui).isEmpty()){
//				logger.warn(cui + " has no concept names!!"); // TODO remove me
			}
			
			//-----------------------------------------
			// Go thru each concept name
			for (String conceptName : cuisToConceptNames.get(cui)){
				++nextConceptId;
				
				//-----------------------------------------
				// Transform the concept name into trigram info
				Pair<Integer, Map<Integer, Integer>> tempPair = ProcessCharacterTrigram.processInput(conceptName);
				Map<Integer, Integer> trigramIdsToCounts = tempPair.getRight();
				
				//-----------------------------------------
				// Place the CUI / concept name pair in hash map, as
				// a concept under a concept ID, and as MinHash table entries
				for (int i=0; i < numberOfMinHashTables; ++i){
					int hashCode = minHashTableDefinitions.get(i).getHashCode(trigramIdsToCounts);
					
					conceptIdsToCuis.put(nextConceptId, cui);
					conceptIdsToConceptNames.put(nextConceptId, conceptName);
					
					if (!minHashTables.get(i).containsKey(hashCode)){
						minHashTables.get(i).put(hashCode, new HashSet<Integer>());
					}
					minHashTables.get(i).get(hashCode).add(nextConceptId);
				}
				
				//-----------------------------------------
				// Report interim progress
				if (nextConceptId % 500000 == 0 && nextConceptId > 0){
					logger.info(nextConceptId + " concepts hashed...");
				}
			} // for each concept name
		} // for each CUI
	}
	
	private static Map<String, List<String>> incorporateConceptNames(Map<String, List<String>> baseMap,
			Map<String, List<String>> additionsMap){		
		//-----------------------------------------
		// In the additions map, there are CUI --> concept names. Incorporate
		// them into the base map so we end up with only one mapping.
		for (String cui : additionsMap.keySet()){
			List<String> additionalConceptNames = additionsMap.get(cui);
			List<String> baseConceptNames = baseMap.get(cui);
			
			for (String additionalConceptName : additionalConceptNames){
				if (!baseConceptNames.contains(additionalConceptName)){
					baseConceptNames.add(additionalConceptName);
				}
			}
		}
		
		return baseMap;
	}
	
	private static void initializeHashMaps(){
		conceptIdsToConceptNames = new HashMap<Integer, String>();
		conceptIdsToCuis = new HashMap<Integer, String>();
		minHashTableDefinitions = new ArrayList<MinHashTableDefinition>();
		minHashTables = new ArrayList<Map<Integer, Set<Integer>>>(); 
		for (int i=0; i < numberOfMinHashTables; ++i){
			minHashTables.add(new HashMap<Integer, Set<Integer>>());
		}
		params = new HashMap<String, String>();
	}
	
	private static boolean isPrime(long n){
		if (n <= 2){
			return n == 2;
		}
		else if (n % 2 == 0){
			return false;
		}
		
		for (int i = 3, end = (int) Math.sqrt(n); i <= end; i += 2){
			if (n % i == 0){
				return false;
			}
		}
		return true;
	}
	
	private static void saveHashMapsAsCsvFiles(String csvDir) throws IOException{
		List<String> conceptNamesList = new ArrayList<String>();
		
		List<Integer> sortedConceptIds = new ArrayList<Integer>(conceptIdsToConceptNames.keySet());
		Collections.sort(sortedConceptIds);
		
		for (int conceptId : sortedConceptIds){
			conceptNamesList.add(conceptId + "\t" + conceptIdsToConceptNames.get(conceptId));
		}
		SaveToFile.saveListOfStringsToFile(conceptNamesList, csvDir + "ConceptNames.csv");
		
		List<String> cuisList = new ArrayList<String>();
		for (int conceptId : sortedConceptIds){
			cuisList.add(conceptId + "\t" + conceptIdsToCuis.get(conceptId));
		}
		SaveToFile.saveListOfStringsToFile(cuisList, csvDir + "CUIs.csv");
		
		for (int i=0; i < numberOfMinHashTables; ++i){
			List<String> minHashTableList = new ArrayList<String>();
			for (int hashCode :  minHashTables.get(i).keySet()){
				minHashTableList.add(hashCode + "\t" + StringUtils.join(minHashTables.get(i).get(hashCode), "\t"));
			}
			SaveToFile.saveListOfStringsToFile(minHashTableList, csvDir + "MinHashTable" + i + ".csv");
		}
		
		List<String> paramsList = new ArrayList<String>();
		for (String param : params.keySet()){
			paramsList.add(param + "\t" + params.get(param));
		}
		SaveToFile.saveListOfStringsToFile(paramsList, csvDir + "Params.csv");
	}
	
	private static void storeParamsInHashMap(){		
		params.put("number of MinHash tables", "" + numberOfMinHashTables);
		params.put("number of hash functions per MinHash table", "" + numberOfHashFunctionsToUse);
		params.put("prime", "" + prime);
		for (int i=0; i < numberOfMinHashTables; ++i){
			for (int j=0; j < numberOfHashFunctionsToUse; ++j){
				params.put("MinHash table " + i + " hash function " + j + " coeff a",
						"" + minHashTableDefinitions.get(i).getACoeffs().get(j));
				params.put("MinHash table " + i + " hash function " + j + " coeff b",
						"" + minHashTableDefinitions.get(i).getBCoeffs().get(j));
			}
		}
	}

	public static void main(String[] args) throws Exception{		
		//-----------------------------------------
		// Get the settings
		Properties settings = CommandLineArguments.getSettings(args);
		
		//-----------------------------------------
		// Initialize the variables
		numberOfMinHashTables = Integer.valueOf(settings.getProperty("bebe_string_sim_number_of_minhash_tables"));
		numberOfHashFunctionsToGenerate = Integer.valueOf(settings.getProperty("bebe_string_sim_number_of_hash_functions_to_choose_from"));
		numberOfHashFunctionsToUse = Integer.valueOf(settings.getProperty("bebe_string_sim_number_of_hash_functions_per_MinHash_table"));
		numberOfDimensions = Integer.valueOf(settings.getProperty("bebe_string_sim_number_of_dimensions"));
		int seedForRandomNumberGenerator = Integer.valueOf(settings.getProperty("bebe_string_sim_seed_for_random_number_generator"));
		String cvsDir = settings.getProperty("bebe_string_sim_csv_directory");
		
		//-----------------------------------------
		// Initialize the hash maps
		logger.info("Initializing...");
		initializeHashMaps();
		ProcessCharacterTrigram.initializeForCreatingData();
		
		//-----------------------------------------
		// Find the smallest prime number larger than the number of dimensions
		prime = getPrimeBiggerThan(numberOfDimensions);
		
		//-----------------------------------------
		// Initialize the random number generator
		// Be careful not to re-initialize again, i.e. don't disrupt the sequence of random numbers!
		RandomNumber.initialize(seedForRandomNumberGenerator);
		
		//-----------------------------------------
		// Generate the pool of random hash functions
		// Each hash function has the form: (a * x + b) % prime
		// Therefore generate the a's and b's
		poolOfACoeffs = getRandomNumbersBetween(1, prime, numberOfHashFunctionsToGenerate);
		poolOfBCoeffs = getRandomNumbersBetween(0, prime, numberOfHashFunctionsToGenerate);
		
		//-----------------------------------------
		// Construct MinHash table definitions
		for (int i=0; i < numberOfMinHashTables; ++i){
			MinHashTableDefinition minHashTableDefinition =
					new MinHashTableDefinition(numberOfHashFunctionsToUse, prime, poolOfACoeffs, poolOfBCoeffs);
			minHashTableDefinitions.add(minHashTableDefinition);
		}
		
		//-----------------------------------------
		// At this point only the params are not yet stored in hash map, so do that
		storeParamsInHashMap();
		logger.info("Done");
		
		//-----------------------------------------
		// Extract CUIs and concept names suitable for BEBE string
		// similarity-based method from UMLS's MRCONSO.RRF file
		logger.info("Extracting concept names from MRCONSO...");
		String mrconsoFilename = settings.getProperty("umls_mrconso_filename");
		int maxNumberOfChars = Integer.valueOf(settings.getProperty("bebe_string_sim_max_number_of_characters"));
		int maxNumberOfWords = Integer.valueOf(settings.getProperty("bebe_string_sim_max_number_of_consecutive_words"));
		Map<String, List<String>> cuisToConceptNames = extractFromMrconso(mrconsoFilename,
				maxNumberOfChars, maxNumberOfWords);
		logger.info("Done");
		
		//-----------------------------------------
		// Add variations of concept names based on noun plurals
		logger.info("Generating variations based on noun plurals...");
		Map<String, List<String>> cuisToNounPlurals = GenerateVariationsNounPlurals.generate(settings, cuisToConceptNames);
		logger.info("Done");
		
		//-----------------------------------------
		// Add variations of concept names based on verb tenses
		logger.info("Generating variations based on verb tenses...");
		Map<String, List<String>> cuisToVerbTenses = GenerateVariationsVerbTenses.generate(settings, cuisToConceptNames);
		logger.info("Done");
		
		//-----------------------------------------
		// Combine the noun plurals and verb tenses with the "base"
		// set of concept names
		cuisToConceptNames = incorporateConceptNames(cuisToConceptNames, cuisToNounPlurals);
		cuisToConceptNames = incorporateConceptNames(cuisToConceptNames, cuisToVerbTenses);
		
		//-----------------------------------------
		// Hash each CUI / concept name pair. Notice that as a side effect, we will also
		// have cached trigrams and concept names besides the hash results.
		logger.info("Computing hash values...");
		hashConceptNames(cuisToConceptNames);
		logger.info("Done");
		
		//-----------------------------------------
		// Save the hash maps in files
		logger.info("Saving hash maps in CSV files...");
		saveHashMapsAsCsvFiles(cvsDir);
		ProcessCharacterTrigram.saveTrigramsToIdsAsCsv(cvsDir + "Trigrams.csv");
		logger.info("Done");
	}
}