/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.stringsim.precompute;

import edu.stanford.nlp.util.StringUtils;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import mpii.biotm.wrapper.morphadorner.WordPluralizer;
import mpii.biotm.wrapper.wordnet.WordNetDictionary;

import org.apache.commons.lang3.CharUtils;
import org.apache.log4j.Logger;

//-----------------------------------------
// We want to augment the dictionary with pre-computed noun plurals.
// E.g. "M'Naghten's Rule" is CUI C0751248 but "M'Naghten's Rules" does
// not appear under the same CUI. MinHash fails when the plural is not
// available.
public class GenerateVariationsNounPlurals{
	// TODO: plural forms not spelled with an "s"
	// TODO: bug? C1261540 "blood potassium level" does not have "... levels" generated,
	//       but C0729816 also "blood potassium level" does
	
	protected static final Logger logger = Logger.getLogger(GenerateVariationsNounPlurals.class);
	
	private static Set<String> allConceptNamesInLowerCase = null;
	
	private static List<String> detectPlurals(List<String> conceptNames){
		List<String> returnList = new ArrayList<String>();
		
		//-----------------------------------------
		// Sometimes a CUI is rich in lexical variants.
		// E.g. C0004910 has:
		// Bed Bug, Bed Bugs, Bedbug, Bedbugs, Bug Bed, Bugs Bed, Cimex, Cimices
		// In this case, all the plurals are explicitly provided, and we don't
		// need to infer them.
		//
		// However, sometimes the plural is missing.
		// E.g. C0751248 has only: M'Naghten's Rule
		// In this case, we want to infer "M'Naghten's Rules".
		
		//-----------------------------------------
		// Remove concept names that do not end in a letter
		// E.g. "baseball [game]", "XYZ [Disease/Finding]", "So-and-so 1932"
		// will be removed from the list of concept names
		//
		// Remove concept names that contain non-English words
		// E.g. 5 Bromo 2 Deoxycytidine
		//
		// Remove concept names that do not end in an English noun
		int i = 0;
		while (i < conceptNames.size()){
			String entireString = conceptNames.get(i);
			
			// Notice that if the entire string ends in say "]",
			// the last element in words list is the word prior to the "]".
			// That's why depending on what we want to check, we sometimes
			// pass the entire string, and sometimes the divided words.
			List<String> words = splitIntoWords(entireString);
			
			// Sometimes we have a concept name that does not contain
			// any actual word, e.g. C0439164 "%". So wipe out the
			// concept name when this happens.
			if (words.size() == 0){
				conceptNames.remove(i);
			}
			else{
				String lastWord = words.get(words.size() - 1);

				if (!lastCharIsAlpha(entireString) ||
						!StringUtils.isAlpha(lastWord) ||
						!isEnglish(entireString) ||
						!isEnglishNoun(lastWord)){
					conceptNames.remove(i);
				}
				else{
					++i;
				}
			}
		}
		
		//-----------------------------------------
		// Continue only if we still have concept names left
		if (conceptNames.size() > 0){
			
			//-----------------------------------------
			// Remove singular / plural pairs so that we are left with
			// singular concept names that can be made into plural
			conceptNames = removeSingularPluralPairs(conceptNames);
			
			//-----------------------------------------
			// Remove concept names that are already in plural form
			conceptNames = removePlural(conceptNames);
			
			//-----------------------------------------
			// Now all remaining concept names are assumed to be singular.
			// Generate plural forms. If they don't already exist as some
			// concept name for any CUI, return them. We need to do this
			// test because sometimes a singular and a plural entity are
			// already separate CUIs.
			for (String conceptName : conceptNames){
				String pluralConceptName = getPluralForm(conceptName);
				if (!allConceptNamesInLowerCase.contains(pluralConceptName.toLowerCase())){
					returnList.add(pluralConceptName);
				}
			}
		} // if have concept name(s) left
		
		return returnList;
	}
	
	public static Map<String, List<String>> generate(Properties settings, Map<String, List<String>> cuisToConceptNames)
			throws IOException{
		Map<String, List<String>> returnMap = new HashMap<String, List<String>>();
		
		//-----------------------------------------
		// Initialize variables
		String wordNetDir = settings.getProperty("wordnet_directory");
		allConceptNamesInLowerCase = new HashSet<String>();
		
		//-----------------------------------------
		// Initialize WordNet
		WordNetDictionary.initialize(wordNetDir);
		
		//-----------------------------------------
		// Initialize MorphAdorner
		WordPluralizer.initialize();
		
		//-----------------------------------------
		// Make a register of all known concept names
		for (String cui : cuisToConceptNames.keySet()){
			for (String conceptName : cuisToConceptNames.get(cui)){
				allConceptNamesInLowerCase.add(conceptName.toLowerCase());
			}
		}
		
		//-----------------------------------------
		// Go thru each CUI and its corresponding concept names
		for (String cui : cuisToConceptNames.keySet()){
			
			//-----------------------------------------
			// Generate all applicable noun plurals for this CUI
			// Make a copy of the list concept names, because detectPlurals()
			// need to remove some entries
			List<String> copyOfConceptNames = new ArrayList<String>(cuisToConceptNames.get(cui));
			List<String> pluralConceptNames = detectPlurals(copyOfConceptNames);
			
			//-----------------------------------------
			// Put the plurals in the return value
			if (!pluralConceptNames.isEmpty()){
				returnMap.put(cui, pluralConceptNames);
				
				//-----------------------------------------
				// Record the plural concept names in the set of all concept names
				for (String pluralConceptName : pluralConceptNames){
					allConceptNamesInLowerCase.add(pluralConceptName.toLowerCase());
				}
			}
		}

		return returnMap;
	}
	
	private static List<Integer> getIndexesOfSingularPluralPairs(List<String> words){
		
		//-----------------------------------------
		// Keep track of which plurals we need to remove later.
		// Use a set to avoid duplicates.
		Set<Integer> indexes = new HashSet<Integer>();
		
		//-----------------------------------------
		// Find singular/plural pairs of words,
		// e.g. "acid" and "acids"
		int i = 0;
		
		// Start with one word. Determine whether it's in singular form.
		// If it's in singular form, look at all other words to see if
		// the other words are the plural form.
		while (i < words.size()){
			// Keep track of whether i-th word got any matching plural j-words
			boolean iWordHasPluralJWord = false;
			
			// Continue if the i-th word is in singular form
			String iWord = words.get(i);
			if (isSingular(iWord)){

				// See if the j-th word is the i-th word's plural form
				int j = 0;
				while (j < words.size()){
					if (i != j){
						String jWord = words.get(j);
						
						if (isSingularPluralPair(iWord, jWord)){
							indexes.add(j);
							iWordHasPluralJWord = true;
						}
					}
					
					++j;
				} // for each j-th word
			} // if i-th word is in singular form
			
			if (iWordHasPluralJWord){
				indexes.add(i);
			}
			
			++i;
		} // for each i-th word
		
		return new ArrayList<Integer>(indexes);
	}
	
	private static String getPluralForm(String inputString){
		//-----------------------------------------
		// Assumptions:
		// - The input string contains only English
		//   words
		// - The input string has at least one word
		// - The last word is in a noun in singular form
		
		//-----------------------------------------
		// Get the last word
		List<String> words = splitIntoWords(inputString);
		String lastWord = words.get(words.size() - 1);
		
		//-----------------------------------------
		// Get the plural form of the last word
		// Note to self: Changing the last word to lower case here
		// because EnglishPluralizer seems to have a bug -- "STUDY"
		// causes an error but "study" works fine.
		String pluralLastWord = WordPluralizer.pluralize(lastWord.toLowerCase());
		
		//-----------------------------------------
		// Replace last word with new plural form
		words.set(words.size() - 1, pluralLastWord);
		
		return StringUtils.join(words, " ");
	}
	
	private static boolean isEnglish(String inputString){
		//-----------------------------------------
		// Return true if the input string only has alpha,
		// hyphen, apostrophe, and space characters.
		return inputString.matches("^[a-zA-Z-' ]+$");
	}
	
	private static boolean isEnglishNoun(String word){
		boolean returnBool = false;
		word = word.toLowerCase();
		
		if (!word.isEmpty()){			
			//-----------------------------------------
			// See if the word is a WordNet noun as-is, but take care
			// to consider words 2 chars or shorter as non-noun
			if (word.length() >= 3 && WordNetDictionary.isNoun(word)){
				returnBool = true;
			}
			
			//-----------------------------------------
			// WordNet doesn't have the word as a noun as-is.
			// Maybe the word is already in plural. See if we
			// can reduce the word to singular form.
			//
			// Look at words ending in "s". Constrain length to be >= 4
			// because we don't want "no" or "vi" to be considered a noun.
			if (!returnBool){
				if (word.endsWith("s") && word.length() >= 4){
					word = word.substring(0, word.length() - 1);
					if (WordNetDictionary.isNoun(word)){
						returnBool = true;
					}
				}
			}
			
			//-----------------------------------------
			// Look at words ending in "es". Constrain length to be >= 5
			// because we don't want "hi" to be considered a noun.
			if (!returnBool){
				if (word.endsWith("es") && word.length() >= 5){
					word = word.substring(0, word.length() - 2);
					if (WordNetDictionary.isNoun(word)){
						returnBool = true;
					}
				}
			}
		} // Input word is not empty
		
		return returnBool;
	}
	
	private static boolean isSingular(String word){
		boolean returnBool = true;
		
		//-----------------------------------------
		// Assumption: the input word is an English word, listed as
		// a noun in WordNet
		
		//-----------------------------------------
		// A word ending in a letter that is not "s" is singular.
		// So investigate further if the word ends in "s".
		word = word.toLowerCase();
		if (word.endsWith("s")){
			//-----------------------------------------
			// Use a simple rule: If the last two letters are "ss",
			// the word is singular.
			if (!word.endsWith("ss")){
				returnBool = false;
			}
		}

		return returnBool;
	}
	
	private static boolean isSingularPluralPair(String singularWord, String pluralWord){
		boolean returnBool = false;
		
		//-----------------------------------------
		// If the plural word is the singular word plus "s" or "es",
		// return true.
		singularWord = singularWord.toLowerCase();
		pluralWord = pluralWord.toLowerCase();
		
		if (pluralWord.startsWith(singularWord)){
			String suffix = pluralWord.substring(singularWord.length());
			
			if (suffix.equals("s") || suffix.equals("es")){
				returnBool = true;
			}
		}
		
		return returnBool;
	}
	
	private static boolean lastCharIsAlpha(String inputString){
		boolean returnBool = false;
		
		if (!inputString.isEmpty()){
			char lastChar = inputString.charAt(inputString.length() - 1);
			returnBool = CharUtils.isAsciiAlpha(lastChar);
		}
		
		return returnBool;
	}
	
	private static int lengthOfLongestList(List<List<String>> listOfLists){
		int maxLength = 0;
		for (List<String> list : listOfLists){
			if (list.size() > maxLength){
				maxLength = list.size();
			}
		}
		
		return maxLength;
	}
	
	private static List<List<String>> removeElementsOfIndexes(List<List<String>> listsOfWords, List<Integer> indexes){
		//-----------------------------------------
		// Remove all the indicated lists.
		// Do this by sorting the indexes in descending order, and
		// remove in descending order. This way, we don't have to
		// adjust the indexes as the list shrinks.
		Collections.sort(indexes);
		Collections.reverse(indexes);
		for (int index : indexes){
			listsOfWords.remove(index);
		}
		
		return listsOfWords;
	}
	
	private static List<String> removeSingularPluralPairs(List<String> conceptNames){
		List<String> returnList = new ArrayList<String>();
		
		//-----------------------------------------
		// Given a list of concept names, find out which are singular/plural
		// pairs, and remove them.
		// E.g. bee, bees, apiary --> apiary
		//
		// Sometimes the remaining concept name is already in plural, but
		// that is rare. The concept names returned is expected to be made
		// into plural form, and we would get a wrong concept name. Since
		// it's rare, we'll worry about that later, if ever...
		
		//-----------------------------------------
		// If the input list has only one concept name, return
		// it as is.
		if (conceptNames.size() == 1){
			returnList.addAll(conceptNames);
		}
		
		//-----------------------------------------
		// Two or more concept names
		else{
			//-----------------------------------------
			// Transform all the concept names into words
			List<List<String>> listsOfWords = new ArrayList<List<String>>();
			for (String conceptName : conceptNames){
				listsOfWords.add(splitIntoWords(conceptName));
			}
			
			//-----------------------------------------
			// Remove singular/plural pairs based on the first, second,
			// up to the last word.
			//
			// E.g. for first word:
			// Characteristic Acceptor, Characteristics Acceptor
			//
			// E.g. for second word:
			// Accessory Nerve Spinal, Accessory Nerves Spinal
			//
			// E.g. for last word:
			// Acoustic Impedance Test, Acoustic Impedance Tests
			
			//-----------------------------------------
			// The last word scenario is most common, so process that first.
			// Then move on to first word, second word, etc.
			listsOfWords = removeSingularPluralPairsBasedOnLastWords(listsOfWords);
			
			//-----------------------------------------
			// For any remaining concept names, find out the number of
			// words in the concept name with the most words
			int maxNumberOfWords = lengthOfLongestList(listsOfWords);
			
			//-----------------------------------------
			// Go thru the first, second, ... word
			// Note to self: no need to work on the case when i == maxNumberOfWords
			// because the last word has been processed already
			for (int i=0; i < maxNumberOfWords; ++i){
				listsOfWords = removeSingularPluralPairsBasedOnIthWords(listsOfWords, i);
			}
			
			//-----------------------------------------
			// At this point, we finally have all whittled away all the singular/plural
			// pairs. Transform the words back to a single string.
			for (List<String> words : listsOfWords){
				returnList.add(StringUtils.join(words, " "));
			}
		} // Two or more concept names
		
		return returnList; 
	}
	
	private static List<String> removePlural(List<String> conceptNames){
		int i = 0;
		while (i < conceptNames.size()){
			if (isSingular(conceptNames.get(i))){
				++i;
			}
			else{
				conceptNames.remove(i);
			}
		}
		
		return conceptNames; 
	}
	
	private static List<List<String>> removeSingularPluralPairsBasedOnIthWords(List<List<String>> listsOfWords, int i){
		//-----------------------------------------
		// Get the i-th word from each concept name.
		// If the concept name has fewer than i words, use an empty string.
		List<String> words = new ArrayList<String>();
		for (List<String> listOfWords : listsOfWords){
			if (listOfWords.size() > i){
				words.add(listOfWords.get(i));
			}
			else{
				words.add("");
			}
		}
		
		//-----------------------------------------
		// Get the indexes that point to singular/plural word pairs
		List<Integer> indexesToBeRemoved = getIndexesOfSingularPluralPairs(words);
		
		//-----------------------------------------
		// Remove the concept names (list of words) marked
		listsOfWords = removeElementsOfIndexes(listsOfWords, indexesToBeRemoved);
		
		return listsOfWords;
	}
	
	private static List<List<String>> removeSingularPluralPairsBasedOnLastWords(List<List<String>> listsOfWords){
		//-----------------------------------------
		// Get the last word from each concept name.
		List<String> lastWords = new ArrayList<String>();
		for (List<String> listOfWords : listsOfWords){
			lastWords.add(listOfWords.get(listOfWords.size() - 1));
		}
		
		//-----------------------------------------
		// Get the indexes that point to singular/plural word pairs
		List<Integer> indexesToBeRemoved = getIndexesOfSingularPluralPairs(lastWords);
		
		//-----------------------------------------
		// Remove the concept names (list of words) marked
		listsOfWords = removeElementsOfIndexes(listsOfWords, indexesToBeRemoved);
		
		return listsOfWords;
	}
	
	private static List<String> splitIntoWords(String inputString){
		List<String> returnList = new ArrayList<String>();
		
		//-----------------------------------------
		// Split the string into words
		List<String> words = StringUtils.split(inputString, "\\W");
		
		// Note to self: When I did words.remove(), java complains of
		// java.lang.UnsupportedOperationException from AbstractList.java.
		// So just do it the dumb way and copy over every non-empty word.
		
		// Because we split by \W, it is possible to have empty tokens.
		// Remove all empty tokens.
		for (String word : words){
			if (!word.isEmpty()){
				returnList.add(word);
			}
		}

		return returnList;
	}
}