/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.bebe.stringsim.precompute;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import mpii.biotm.util.ReadFromFile;
import mpii.biotm.wrapper.umls.LookupSemanticGroup;
import mpii.biotm.wrapper.umls.LookupSemanticType;

import org.apache.log4j.Logger;

import edu.mit.jwi.Dictionary;
import edu.mit.jwi.IDictionary;
import edu.mit.jwi.item.POS;
import edu.northwestern.at.utils.corpuslinguistics.inflector.Person;
import edu.northwestern.at.utils.corpuslinguistics.inflector.VerbTense;
import edu.northwestern.at.utils.corpuslinguistics.inflector.conjugator.EnglishConjugator;

//-----------------------------------------
// We want to augment the dictionary with pre-computed verb variants.
// E.g. "Widened" is CUI C0332464, but it's variants "widen," "widens,"
// etc. are not explicitly listed in UMLS. As a result, BEBE's string
// similarity MinHash module misses them. This class looks at a list
// of single-word concept names, checks if they are verbs, and if yes,
// generate any missing variants.
public class GenerateVariationsVerbTenses{
	
	protected static final Logger logger = Logger.getLogger(GenerateVariationsVerbTenses.class);
	
	private static Set<String> allConceptNamesInLowerCase = null;
	private static EnglishConjugator conjugator = null;
	private static Map<String, String> cuisToSemanticGroups = null;
	private static IDictionary dictionary = null;
	private static List<POS> poses = new ArrayList<POS>();
	static{
		poses.add(POS.ADJECTIVE);
		poses.add(POS.ADVERB);
		poses.add(POS.NOUN);
		poses.add(POS.VERB);
	}
	
	private static List<String> findMissingVerbTenses(String infinitive){
		List<String> returnList = new ArrayList<String>();
		
		//-----------------------------------------
		// Assumption: input param is a verb in infinitive
		
		//-----------------------------------------
		// From the infinitive (1), conjugate the verb to get present third person (2),
		// gerund (3), past tense (4), past participle (5).
		String presentThirdPerson = conjugator.conjugate(infinitive, VerbTense.PRESENT, Person.THIRD_PERSON_SINGULAR);
		String gerund = conjugator.conjugate(infinitive, VerbTense.PRESENT_PARTICIPLE, Person.FIRST_PERSON_SINGULAR);
		String past = conjugator.conjugate(infinitive, VerbTense.PAST, Person.FIRST_PERSON_SINGULAR);
		String pastParticiple = conjugator.conjugate(infinitive, VerbTense.PAST_PARTICIPLE, Person.FIRST_PERSON_SINGULAR);
		
		//-----------------------------------------
		// From (1) to (5), get a unique list of words. Any word in this unique
		// list that is not already a known concept name will be returned.
		if (!allConceptNamesInLowerCase.contains(infinitive)){
			returnList.add(infinitive);
		}

		if (!allConceptNamesInLowerCase.contains(presentThirdPerson) && !returnList.contains(presentThirdPerson)){
			returnList.add(presentThirdPerson);
		}

		if (!allConceptNamesInLowerCase.contains(gerund) && !returnList.contains(gerund)){
			returnList.add(gerund);
		}

		if (!allConceptNamesInLowerCase.contains(past) && !returnList.contains(past)){
			returnList.add(past);
		}

		if (!allConceptNamesInLowerCase.contains(pastParticiple) && !returnList.contains(pastParticiple)){
			returnList.add(pastParticiple);
		}
		
		return returnList;
	}
	
	private static List<String> findVerbTenses(String word){
		List<String> returnList = new ArrayList<String>();
		
		//-----------------------------------------
		// Assumption: input param is a single word
		String modifiedWord = word.toLowerCase();
		
		//-----------------------------------------
		// We are looking for a verb in any form (past tense, gerund, etc.)
		// So continue only when the word contains a vowel or the letter y
		if (modifiedWord.contains("a") || modifiedWord.contains("e") || modifiedWord.contains("i") ||
				modifiedWord.contains("o") ||	modifiedWord.contains("u") || modifiedWord.contains("y")){
			
			//-----------------------------------------
			// Determine if the word is a verb and if so, get it's infinitive
			modifiedWord = getInfinitive(modifiedWord);
			
			//-----------------------------------------
			// Knowing that we have a verb in infinitive, find all its tenses
			// that are not known concept names
			if (!modifiedWord.isEmpty()){
				returnList = findMissingVerbTenses(modifiedWord);
			}
		}
		
		return returnList;
	}
	
	public static Map<String, List<String>> generate(Properties settings, Map<String, List<String>> cuisToConceptNames)
			throws Exception{
		Map<String, List<String>> returnMap = new HashMap<String, List<String>>();
		
		//-----------------------------------------
		// Initialize WordNet
		String wordNetDir = settings.getProperty("wordnet_directory");
		URL url = new URL("file", null, wordNetDir);
		dictionary = new Dictionary(url);
		dictionary.open();

		//-----------------------------------------
		// Initialize MorphAdorner
		conjugator = new EnglishConjugator();

		//-----------------------------------------
		// Make a cache of all concept names
		allConceptNamesInLowerCase = new HashSet<String>();
		for (String cui : cuisToConceptNames.keySet()){
			for (String conceptName : cuisToConceptNames.get(cui)){
				allConceptNamesInLowerCase.add(conceptName.toLowerCase());
			}
		}
		
		//-----------------------------------------
		// Make a cache of the semantic groups of all CUIs
		populateSemanticGroups(settings);
		
		//-----------------------------------------
		// Go thru each CUI and its corresponding concept names
		for (String cui : cuisToConceptNames.keySet()){
						
			//-----------------------------------------
			// Make sure we don't mistake say a drug name with an English word.
			// Judge via semantic type.
			if (isPossiblyVerb(cui)){
				List<String> tempResults = new ArrayList<String>();
				
				//-----------------------------------------
				// Go thru each concept name
				for (String conceptName : cuisToConceptNames.get(cui)){
					
					//-----------------------------------------
					// Continue only if the concept name is a single word
					if (isSingleWord(conceptName)){
			
						//-----------------------------------------
						// Find any verb tenses that are not known concept names
						List<String> verbTenses = findVerbTenses(conceptName);
						for (String verbTense : verbTenses){

							//-----------------------------------------
							// Only output the same verb tense once
							if (!tempResults.contains(verbTense)){
								tempResults.add(verbTense);
							}
						}
					}
				}
				
				//-----------------------------------------
				// If this CUI has verb tenses to be added, register
				// them in the return value
				if (!tempResults.isEmpty()){
					returnMap.put(cui, tempResults);
				}
			}
		}
		
		//-----------------------------------------
		// Clean up
		dictionary.close();
		
		return returnMap;
	}

	private static String getInfinitive(String word){
		// TODO: irregular verbs
		
		//-----------------------------------------
		// Assumption: Input word is in all lower case
		
		//-----------------------------------------
		// If the input word is a verb, return it's infinitive.
		// E.g. "widened" --> "widen"
		// Otherwise return empty string.
		
		//-----------------------------------------
		// Transform the word to infinitive.
		// See if WordNet contains the infinitive as a verb. If not,
		// wipe out the word so we return an empty string.
		int length = word.length();
		
		//-----------------------------------------
		// Third person present tense
		if (word.endsWith("s")){
			
			// Word ends with "ies", e.g. "applies"
			if (word.endsWith("ies")){
				// Replace "ies" with "y" and see if that gives an infinitive
				word = word.substring(0, length - 3) + "y";
				length -= 3;
				
				if (!wordNetHasVerb(word)){
					word = "";
				}
				// else: word is an infinitive, all good, nothing more to do
			}
			
			// Word ends with "s", e.g. "bakes", "adds"
			else{
				// Remove "s" and see if that gives an infinitive
				word = word.substring(0, length - 1);
				--length;
				
				if (!wordNetHasVerb(word)){
					word = "";
				}
				// else: word is an infinitive, all good, nothing more to do
			}
		}
		
		//-----------------------------------------
		// Gerund
		else if (word.endsWith("ing") && length >= 5){
			// Clip off the "ing" at the end
			word = word.substring(0, length - 3);
			length -= 3;
			
			// If we don't have an infinitive, go for more processing
			if (!wordNetHasVerb(word)){
			
				// See if the final two chars are the same. If yes,
				// clip off one of them.
				// E.g. bidding ==> bidd ==> bid
				String lastChar = word.substring(length - 1, length);
				String secondToLastChar = word.substring(length - 2, length - 1);
				if (lastChar.equals(secondToLastChar)){
					word = word.substring(0, length - 1);
					
					if (!wordNetHasVerb(word)){
						word = "";
					}
					// else: word is an infinitive, all good, nothing more to do
				}
				
				// Final two chars are different, and the word is not in WordNet.
				// Wipe out the word to return an empty string.
				else{
					word = "";
				}
			}
			
			// else: word is an infinitive, all good, nothing more to do
		}
		
		//-----------------------------------------
		// Past / past participle
		else if (word.endsWith("ed") && length > 2){
			// Clip off the "ed" at the end
			word = word.substring(0, length - 2);
			length -= 2;

			// If we don't have an infinitive, go for more processing
			if (!wordNetHasVerb(word)){

				// See if putting the "e" back to the word will give a verb
				// E.g. baked ==> bak ==> bake
				word += "e";
				++length;

				if (!wordNetHasVerb(word)){
					word = "";
				}
				// else: word is an infinitive, all good, nothing more to do
			}

			// else: word is an infinitive, all good, nothing more to do
		}
		
		//-----------------------------------------
		// Everything else: possibly infinitive, check for that
		else if (!wordNetHasVerb(word)){
			word = "";
		}
		
		// else: word is infinitive, all good, nothing more to do
		
		return word;
	}
	
	private static boolean isPossiblyVerb(String cui){
		boolean returnBool = false;
		
		//-----------------------------------------
		// Find the semantic group of the given CUI. If the semantic type
		// indicates that any concept name under this CUI could possibly
		// be a verb, return true.
		String semanticGroup = cuisToSemanticGroups.get(cui);
		
		if (semanticGroup.equalsIgnoreCase("Activities & Behaviors") ||
				semanticGroup.equalsIgnoreCase("Concepts & Ideas") ||
				semanticGroup.equalsIgnoreCase("Phenomena") ||
				semanticGroup.equalsIgnoreCase("Physiology") ||
				semanticGroup.equalsIgnoreCase("Procedures")){
			returnBool = true;
		}
		
		return returnBool;
	}
	
	private static boolean isSingleWord(String inputString){
		//-----------------------------------------
		// Rather than try to split the input string, simply
		// make sure the input string has only English letters.
		// This works for our focus of conjugating English verbs.
		return inputString.matches("^[a-zA-Z]+$");
	}
	
	private static void populateSemanticGroups(Properties settings) throws IOException{
		//-----------------------------------------
		// Assumption: Each CUI belongs only to one semantic group. I last verified
		// this with UMLS2013AA. Even when a CUI belongs to multiple semantic types,
		// all the semantic types belong to the same semantic group.
		//
		// Update: With UMLS2014AB, this is no longer true... TODO
		cuisToSemanticGroups = new HashMap<String, String>();
		
		//-----------------------------------------
		// Read in the MRSTY.RRF file
		String mrstyFilename = settings.getProperty("umls_mrsty_filename");
		List<String> inputLines = ReadFromFile.readFileToListOfStrings(mrstyFilename);
		
		//-----------------------------------------
		// Go thru the lines one by one, and extract the TUI for each CUI.
		// If a CUI has multiple TUIs, just overwrite previous ones with
		// the new one. See assumption stated above.
		for (String inputLine : inputLines){
			
			//-----------------------------------------
			// Tokenize the string to get the CUI and TUI
			// Note to self: StringUtils.split() treats multiple empty
			// columns as a single column
			List<String> tokens = Arrays.asList(inputLine.split("\\|"));
			String cui = tokens.get(0);
			String tui = tokens.get(1);
			
			//-----------------------------------------
			// Find out the TUI's corresponding semantic group
			String semanticType = LookupSemanticType.lookupViaTui(tui);
			String semanticGroup = LookupSemanticGroup.lookup(semanticType);
			
			//-----------------------------------------
			// Register the CUI's semantic group
			cuisToSemanticGroups.put(cui, semanticGroup);
		}
	}
	
	private static boolean wordNetHasVerb(String word){
		//-----------------------------------------
		// Return true if WordNet contains the input word as a verb
		return (dictionary.getIndexWord(word, POS.VERB) != null);
	}
}