/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.util;

/**
 * We want to standardize how the content of a corpus is stored. Assume that
 * a corpus has its articles referenced by a numerical ID. Then determine the
 * million-th and thousand-th numbers, and use them as parent and child
 * folders. This is best explained via an example:
 * 
 * ID = 12,345,678
 * Root path = /home/siu/corpora/medline/
 * Then the content of this ID will be stored as:
 *    /home/siu/corpora/medline/12/345/12345678.txt
 */
public class DeterminePathBasedOnId{
	
	/**
	 * Given an ID, a root path, and a file extension, return the
	 * extended path with the million-th parent and thousand-th child
	 * directories added, and with the filename at the end. 
	 * E.g. ID = 12,345,678
	 *      Root path = /home/siu/corpora/medline/
	 *      Return: /home/siu/corpora/medline/12/345/12345678.txt
	 * @param id
	 * @param inputPath
	 */
	public static String getPathWithFilename(int id, String inputPath, String fileExtension){
		return getPathWithFilename((long) id, inputPath, fileExtension);
	}
	
	public static String getPathWithFilename(long id, String inputPath, String fileExtension){
		return getPathDirectoriesOnly(id, inputPath) + id + "." + fileExtension;
	}
	
	public static String getPathWithFilename(String id, String inputPath, String fileExtension){
		return getPathWithFilename(Long.parseLong(id), inputPath, fileExtension);
	}
	
	public static String getPathDirectoriesOnly(int id, String inputPath){
		return getPathDirectoriesOnly((long) id, inputPath);
	}
	
	/**
	 * Given an ID and a root path, return the extended path with
	 * the million-th parent and thousand-th child directories added.
	 * E.g. ID = 12,345,678
	 *      Root path = /home/siu/corpora/medline/
	 *      Return: /home/siu/corpora/medline/12/345/
	 * @param id
	 * @param inputPath
	 */
	public static String getPathDirectoriesOnly(long id, String inputPath){
		// Make sure the path ends with '/'
		if (!inputPath.endsWith("/")){
			inputPath += "/";
		}
		return inputPath + DetermineIdPrefix.byMillion(id) + "/" + DetermineIdPrefix.byThousand(id) + "/";
	}
	
	public static String getPathDirectoriesOnly(String id, String inputPath){
		return getPathDirectoriesOnly(Integer.parseInt(id), inputPath);
	}
}
