/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.util;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import mpii.biotm.util.ComparatorFilenamesByLongNumberSuffix;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

public class ListAllFilenamesInDirectory{
	
	protected static final Logger logger = Logger.getLogger(ListAllFilenamesInDirectory.class);
	private static ComparatorFilenamesByLongNumberSuffix comparator = null;
	private static boolean initialized = false;
	
	static class ComparatorFilenamePair implements Comparator<Pair<String, String>>{

		public int compare(Pair<String, String> p1, Pair<String, String> p2){
			String from1 = p1.getLeft();
			String from2 = p2.getLeft();
			return comparator.compare(from1, from2);
		}
	}
	
	private static void initialize(){
		if (!initialized){
			comparator = new ComparatorFilenamesByLongNumberSuffix();
			initialized = true;
		}
	}
	
	/** Given a root folder, list all the files in this folder
	 * as well we all sub folders recursively. */
	public static List<String> list(String folder){
		initialize();
		
		//-----------------------------------------
		// Make sure the folder name does not end with '/'
		folder = removeFinalSlash(folder);
		
		//-----------------------------------------
		// Make recursive call
		List<String> returnList = new ArrayList<String>();
		returnList = recurseDirectory(folder, "", 0, returnList);
		
		//-----------------------------------------
		// Sort the list
		Collections.sort(returnList, comparator);
		
		return returnList;
	}

	/** Given a root input folder, list all the files in this folder
	 * as well we all sub folders recursively. For each file found,
	 * also name a corresponding output filename, such that the
	 * output folder structure will mirror the input folder structure. */
	public static List<Pair<String, String>> list(String inFolder, String outFolder){
		initialize();
		
		//-----------------------------------------
		// Make sure the folder names do not end with '/'
		inFolder = removeFinalSlash(inFolder);
		outFolder = removeFinalSlash(outFolder);
		
		//-----------------------------------------
		// Make recursive call
		List<Pair<String, String>> returnList = new ArrayList<Pair<String, String>>();
		returnList = recurseDirectory(inFolder, outFolder, "", 0, returnList);
		
		/*
		// *** TEMP ***
		// Get rid of files we already processed
		try{
			List<Long> docIds = ReadFromFile.readFileToListOfLongs("/mnt/sdb2/siu/eclipse_workspace/amy_public/etc/mpii/biotm/IDs_done.txt");
			List<Pair<String, String>> newReturnList = new ArrayList<Pair<String, String>>();
			
			for (Pair<String, String> oldReturnPair : returnList){
				String oldInputFilename = oldReturnPair.getLeft();
				int slashPos = oldInputFilename.lastIndexOf("/");
				int dotPos = oldInputFilename.lastIndexOf(".");
				long cuiAndVersion = Long.valueOf(oldInputFilename.substring(slashPos + 1, dotPos));
				if (!docIds.contains(cuiAndVersion)){
					newReturnList.add(oldReturnPair);
				}
			}
			
			returnList = newReturnList;
		}
		catch(Exception e){
			logger.debug(e);
		}
		*/
		
		//-----------------------------------------
		// Sort the list
		Collections.sort(returnList, new ComparatorFilenamePair());
		
		return returnList;
	}
	
	private static List<String> recurseDirectory(String dirName, String offset, int level, List<String> accumulatedList){
		//-----------------------------------------
		// Assumption: dirName and offset do
		// not end with a slash "/", but, offset does start
		// with a slash "/".
		
		//-----------------------------------------
		// Make a File instance for the path under consideration
		String pathToLookAt = dirName + offset;
		File fileOrDir = new File(pathToLookAt);
		
		//-----------------------------------------
		// Case 1: Path is a file
		//         Figure out the corresponding output path
		//         and place the results in accumulated list
		if (fileOrDir.isFile()){

			//-----------------------------------------
			// Add to accumulated list
			accumulatedList.add(pathToLookAt);
		} // Case 1
		
		//-----------------------------------------
		// Case 2: Path is a directory
		//         Get its content, and recursively call recurseDirectory()
		//         on each component
		else if (fileOrDir.isDirectory()){
			List<File> files = Arrays.asList(fileOrDir.listFiles());
			for (File file : files){
				String newOffset = removeFinalSlash(offset + "/" + file.getName());
				recurseDirectory(dirName, newOffset, level + 1, accumulatedList);
			}
		}
		
		return accumulatedList;
	}
	
	private static List<Pair<String, String>> recurseDirectory(String dirNameIn, String dirNameOut,
			String offset, int level, List<Pair<String, String>> accumulatedList){
		//-----------------------------------------
		// Assumption: dirNameIn, dirNameOut, and offset do
		// not end with a slash "/", but, offset does start
		// with a slash "/".
		
		//-----------------------------------------
		// Make a File instance for the path under consideration
		String pathToLookAt = dirNameIn + offset;
		File fileOrDir = new File(pathToLookAt);
		
		//-----------------------------------------
		// Case 1: Path is a file
		//         Figure out the corresponding output path
		//         and place the results in accumulated list
		if (fileOrDir.isFile()){

			//-----------------------------------------
			// Construct corresponding output filename
			String outputFilename = dirNameOut + offset;
			
			//-----------------------------------------
			// Add to accumulated list
			Pair<String, String> tempPair = new ImmutablePair<String, String>(pathToLookAt, outputFilename);
			accumulatedList.add(tempPair);
		} // Case 1
		
		//-----------------------------------------
		// Case 2: Path is a directory
		//         Get its content, and recursively call recurseDirectory()
		//         on each component
		else if (fileOrDir.isDirectory()){
			List<File> files = Arrays.asList(fileOrDir.listFiles());
			for (File file : files){
				String newOffset = removeFinalSlash(offset + "/" + file.getName());
				recurseDirectory(dirNameIn, dirNameOut, newOffset, level + 1, accumulatedList);
			}
		}
		
		return accumulatedList;
	}
	
	private static String removeFinalSlash(String folder){
		if (folder.endsWith("/")){
			folder = folder.substring(0, folder.length() - 1);
		}
		return folder;
	}
}
