/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.tuple.MutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

public class ProcessNonStandardEntity{
	// Entity here means HTML entity, e.g. &nbsp;
	protected static final Logger logger = Logger.getLogger(ProcessNonStandardEntity.class);
	
	private static int numberOfRounds = 4;
	
	private static final Map<String,String> nonStandardEntities = new HashMap<String,String>();
	static{
		nonStandardEntities.put("&Agr;",  "&Alpha;");
		nonStandardEntities.put("&Bgr;",  "&Beta;");
		nonStandardEntities.put("&Ggr;",  "&Gamma;");
		nonStandardEntities.put("&Dgr;",  "&Delta;");
		nonStandardEntities.put("&Egr;",  "&Epsilon;");
		nonStandardEntities.put("&Zgr;",  "&Zeta;");
		nonStandardEntities.put("&EEgr;", "&Eta;");
		nonStandardEntities.put("&Igr;",  "&Iota;");
		nonStandardEntities.put("&Kgr;",  "&Kappa;");
		nonStandardEntities.put("&Lgr;",  "&Lambda;");
		nonStandardEntities.put("&Mgr;",  "&Mu;");
		nonStandardEntities.put("&Ngr;",  "&Nu;");
		nonStandardEntities.put("&Xgr;",  "&Xi;");
		nonStandardEntities.put("&Ogr;",  "&Omicron;");
		nonStandardEntities.put("&Pgr;",  "&Pi;");
		nonStandardEntities.put("&Rgr;",  "&Rho;");
		nonStandardEntities.put("&Sgr;",  "&Sigma;");
		nonStandardEntities.put("&Tgr;",  "&Tau;");
		nonStandardEntities.put("&Ugr;",  "&Upsilon;");
		nonStandardEntities.put("&PHgr;", "&Phi;");
		nonStandardEntities.put("&KHgr;", "&Chi;");
		nonStandardEntities.put("&PSgr;", "&Psi;");
		nonStandardEntities.put("&OHgr;", "&Omega;");
		
		nonStandardEntities.put("&Epsi;", "&Epsilon;");
		nonStandardEntities.put("&Upsi;", "&Upsilon;");
		
		nonStandardEntities.put("&agr;",  "&alpha;");
		nonStandardEntities.put("&bgr;",  "&beta;");
		nonStandardEntities.put("&ggr;",  "&gamma;");
		nonStandardEntities.put("&dgr;",  "&delta;");
		nonStandardEntities.put("&egr;",  "&epsilon;");
		nonStandardEntities.put("&zgr;",  "&zeta;");
		nonStandardEntities.put("&eegr;", "&eta;");
		nonStandardEntities.put("&igr;",  "&iota;");
		nonStandardEntities.put("&kgr;",  "&kappa;");
		nonStandardEntities.put("&lgr;",  "&lambda;");
		nonStandardEntities.put("&mgr;",  "&mu;");
		nonStandardEntities.put("&ngr;",  "&nu;");
		nonStandardEntities.put("&xgr;",  "&xi;");
		nonStandardEntities.put("&ogr;",  "&omicron;");
		nonStandardEntities.put("&pgr;",  "&pi;");
		nonStandardEntities.put("&rgr;",  "&rho;");
		nonStandardEntities.put("&sgr;",  "&sigma;");
		nonStandardEntities.put("&tgr;",  "&tau;");
		nonStandardEntities.put("&ugr;",  "&upsilon;");
		nonStandardEntities.put("&phgr;", "&phi;");
		nonStandardEntities.put("&khgr;", "&chi;");
		nonStandardEntities.put("&psgr;", "&psi;");
		nonStandardEntities.put("&ohgr;", "&omega;");

		nonStandardEntities.put("&epsi;", "&epsilon;");
		nonStandardEntities.put("&upsi;", "&upsilon;");
		
		nonStandardEntities.put("&vepsilon;", "&epsilon;");
		nonStandardEntities.put("&epsiv;",    "&epsilon;");
		nonStandardEntities.put("&ohm;",      "&Omega;");
		nonStandardEntities.put("&varpi;",    "&pi;");
		nonStandardEntities.put("&phis;",     "&phi;");
		nonStandardEntities.put("&phiv;",     "&phi;");
		nonStandardEntities.put("&varphi;",   "&phi;");
		nonStandardEntities.put("&varrho;",   "&rho;");
		nonStandardEntities.put("&thetas;",   "&theta;");
		nonStandardEntities.put("&thetav;",   "&theta;");
		nonStandardEntities.put("&vtheta;",   "&theta;");
		nonStandardEntities.put("&sigmav;",   "&sigma;");
		nonStandardEntities.put("&vsigma;",   "&sigma;");
		nonStandardEntities.put("&sfgr;",     "&sigma;");
		
		nonStandardEntities.put("&Ascr;",  "A");
		nonStandardEntities.put("&Bscr;",  "B");
		nonStandardEntities.put("&Cscr;",  "C");
		nonStandardEntities.put("&Dscr;",  "D");
		nonStandardEntities.put("&Escr;",  "E");
		nonStandardEntities.put("&Fscr;",  "F");
		nonStandardEntities.put("&Gscr;",  "G");
		nonStandardEntities.put("&Hscr;",  "H");
		nonStandardEntities.put("&Iscr;",  "I");
		nonStandardEntities.put("&Jscr;",  "J");
		nonStandardEntities.put("&Kscr;",  "K");
		nonStandardEntities.put("&Lscr;",  "L");
		nonStandardEntities.put("&Mscr;",  "M");
		nonStandardEntities.put("&Nscr;",  "N");
		nonStandardEntities.put("&Oscr;",  "O");
		nonStandardEntities.put("&Pscr;",  "P");
		nonStandardEntities.put("&Qscr;",  "Q");
		nonStandardEntities.put("&Rscr;",  "R");
		nonStandardEntities.put("&Sscr;",  "S");
		nonStandardEntities.put("&Tscr;",  "T");
		nonStandardEntities.put("&Uscr;",  "U");
		nonStandardEntities.put("&Vscr;",  "V");
		nonStandardEntities.put("&Wscr;",  "W");
		nonStandardEntities.put("&Xscr;",  "X");
		nonStandardEntities.put("&Yscr;",  "Y");
		nonStandardEntities.put("&Zscr;",  "Z");
		
		nonStandardEntities.put("&ascr;",  "a");
		nonStandardEntities.put("&bscr;",  "b");
		nonStandardEntities.put("&cscr;",  "c");
		nonStandardEntities.put("&dscr;",  "d");
		nonStandardEntities.put("&escr;",  "e");
		nonStandardEntities.put("&fscr;",  "f");
		nonStandardEntities.put("&gscr;",  "g");
		nonStandardEntities.put("&hscr;",  "h");
		nonStandardEntities.put("&iscr;",  "i");
		nonStandardEntities.put("&jscr;",  "j");
		nonStandardEntities.put("&kscr;",  "k");
		nonStandardEntities.put("&lscr;",  "l");
		nonStandardEntities.put("&mscr;",  "m");
		nonStandardEntities.put("&nscr;",  "n");
		nonStandardEntities.put("&oscr;",  "o");
		nonStandardEntities.put("&pscr;",  "p");
		nonStandardEntities.put("&qscr;",  "q");
		nonStandardEntities.put("&rscr;",  "r");
		nonStandardEntities.put("&sscr;",  "s");
		nonStandardEntities.put("&tscr;",  "t");
		nonStandardEntities.put("&uscr;",  "u");
		nonStandardEntities.put("&vscr;",  "v");
		nonStandardEntities.put("&wscr;",  "w");
		nonStandardEntities.put("&xscr;",  "x");
		nonStandardEntities.put("&yscr;",  "y");
		nonStandardEntities.put("&zscr;",  "z");
				
		nonStandardEntities.put("&sp;",     " ");
		nonStandardEntities.put("&apos;",   "'");
		nonStandardEntities.put("&dollar;", "\\$");
		nonStandardEntities.put("&U20AC;",  "&euro;");
		nonStandardEntities.put("&percnt;", "%");
		nonStandardEntities.put("&percent;","%");
		nonStandardEntities.put("&lpar;",   "(");
		nonStandardEntities.put("&rpar;",   ")");
		nonStandardEntities.put("&lsqb;",   "[");
		nonStandardEntities.put("&rsqb;",   "]");
		nonStandardEntities.put("&lcub;",   "{");
		nonStandardEntities.put("&rcub;",   "}");
		nonStandardEntities.put("&plus;",   "+");
		nonStandardEntities.put("&hyphen;", "-");
		nonStandardEntities.put("&hyp;",    "-");
		nonStandardEntities.put("&amp;",    "&");
		nonStandardEntities.put("&ast;",    "*");
		nonStandardEntities.put("&sol;",    "/");
		nonStandardEntities.put("&bsol;",   "\\\\");
		nonStandardEntities.put("&equals;", "=");
		nonStandardEntities.put("&period;", ".");
		nonStandardEntities.put("&comma;",  ",");
		nonStandardEntities.put("&semi;",   ";");
		nonStandardEntities.put("&colon;",  ":");
		nonStandardEntities.put("&ratio;",  ":");
		nonStandardEntities.put("&grave;",  "`");
		nonStandardEntities.put("&quest;",  "?");
		nonStandardEntities.put("&excl;",   "!");
		nonStandardEntities.put("&hairsp;", " ");
		nonStandardEntities.put("&sbsp;",   " ");
		nonStandardEntities.put("&quad;",   " ");
		nonStandardEntities.put("&emsp;",   " ");
		nonStandardEntities.put("&emsp13;", " ");
		nonStandardEntities.put("&emsp14;", " ");
		nonStandardEntities.put("&num;",    "#");
		nonStandardEntities.put("&commat;", "@");
		nonStandardEntities.put("&lowbar;", "_");
		nonStandardEntities.put("&verbar;", "|");
		nonStandardEntities.put("&vbar;",   "|");
		nonStandardEntities.put("&ldet;",   "|");
		nonStandardEntities.put("&rdet;",   "|");
		nonStandardEntities.put("&circ;",   "^");
		nonStandardEntities.put("&greater;",">");
		
		nonStandardEntities.put("&degree;", "&deg;");
		nonStandardEntities.put("&horbar;", "&mdash;");
		nonStandardEntities.put("&hibar;",  "&oline;");
		nonStandardEntities.put("&brkbar;", "&brvbar;");
		nonStandardEntities.put("&macron;", "&macr;");
		nonStandardEntities.put("&die;",    "&uml;");
		nonStandardEntities.put("&endash;", "&ndash;");
		nonStandardEntities.put("&emdash;", "&mdash;");
		nonStandardEntities.put("&lrarr;",  "&harr;");
		nonStandardEntities.put("&lsquor;", "&sbquo;");
		nonStandardEntities.put("&rsquor;", "&rsquo;");
		nonStandardEntities.put("&ldquor;", "&bdquo;");
		nonStandardEntities.put("&rdquor;", "&rdquo;");
		nonStandardEntities.put("&ldots;",  "&hellip;");
		nonStandardEntities.put("&cdots;",  "&hellip;");
		nonStandardEntities.put("&ctdot;",  "&hellip;");
		nonStandardEntities.put("&dellip;", "&hellip;");
		nonStandardEntities.put("&ddots;",  "&hellip;");
		nonStandardEntities.put("&dotfill;","&hellip;");
		nonStandardEntities.put("&bepsi;",  "&ni;");
		nonStandardEntities.put("&inf;",    "&infin;");
		nonStandardEntities.put("&pd;",     "&part;");
		nonStandardEntities.put("&ap;",     "&asymp;");
		nonStandardEntities.put("&ape;",    "&cong;");
		nonStandardEntities.put("&emptyv;", "&empty;");
		nonStandardEntities.put("&mnplus;", "&plusmn;");
		
		nonStandardEntities.put("&aleph;", "ℵ");
		nonStandardEntities.put("&becaus;","∵");
		nonStandardEntities.put("&breve;", "˘");
		nonStandardEntities.put("&caron;", "ˇ");
		nonStandardEntities.put("&ctdot;", "⋯");
		nonStandardEntities.put("&dblac;", "˝");
		nonStandardEntities.put("&dot;",   "˙");
		nonStandardEntities.put("&ell;",   "ℓ");
		nonStandardEntities.put("&iff;",   "⇔");
		nonStandardEntities.put("&lrharr;","⇔");
		nonStandardEntities.put("&nsub;",  "⊄");
		nonStandardEntities.put("&ogon;",  "˛");
		nonStandardEntities.put("&planck;","ℏ");
		nonStandardEntities.put("&ring;",  "˚");
		nonStandardEntities.put("&sung;",  "♪");
		nonStandardEntities.put("&vellip;","⋮");
		nonStandardEntities.put("&vdots;", "⋮");
		
		nonStandardEntities.put("&ecedil;",  "ȩ");
		nonStandardEntities.put("&ngrave;",  "ǹ");
		nonStandardEntities.put("&OV0113;",  "K"); // Octal
		nonStandardEntities.put("&OV0312;",  "Ê"); // Octal
		nonStandardEntities.put("&OV0404;",  "Ą"); // Octal
		nonStandardEntities.put("&OV0556;",  "Ů"); // Octal
		
		nonStandardEntities.put("&half;",   "&frac12;");
		nonStandardEntities.put("&frac13;", "&#8531;");
		nonStandardEntities.put("&frac23;", "&#8532;");
		nonStandardEntities.put("&frac15;", "&#8533;");
		nonStandardEntities.put("&frac25;", "&#8534;");
		nonStandardEntities.put("&frac35;", "&#8535;");
		nonStandardEntities.put("&frac45;", "&#8536;");
		nonStandardEntities.put("&frac16;", "&#8537;");
		nonStandardEntities.put("&frac56;", "&#8538;");
		nonStandardEntities.put("&frac18;", "&#8549;");
		nonStandardEntities.put("&frac38;", "&#8540;");
		nonStandardEntities.put("&frac58;", "&#8541;");
		nonStandardEntities.put("&frac78;", "&#8542;");
		
		// Cannot find actual unicode replacement
		nonStandardEntities.put("&Bmacr;", "B"); 
		nonStandardEntities.put("&cmacr;", "c");
		nonStandardEntities.put("&Cmacr;", "C");
		nonStandardEntities.put("&Dmacr;", "D");
		nonStandardEntities.put("&Fmacr;", "F");
		nonStandardEntities.put("&pmacr;", "p");
		nonStandardEntities.put("&tmacr;", "t");
		nonStandardEntities.put("&xmacr;", "x");
		nonStandardEntities.put("&1macr;", "1");
		nonStandardEntities.put("&Ddot;",  "D");
		nonStandardEntities.put("&Fcirc;", "F");
		nonStandardEntities.put("&Ncirc;", "N");
		nonStandardEntities.put("&tacute;","t");
		nonStandardEntities.put("&Vdot;",  "V");
		
		// Most likely typos by original authors
		nonStandardEntities.put("&#x65308;","<"); // Decimal
		nonStandardEntities.put("&Colon;", ";");
		nonStandardEntities.put("&gama;",  "&gamma;");
		nonStandardEntities.put("&HYPHEN;","-");
		nonStandardEntities.put("&lamdda;","&lambda;");
		nonStandardEntities.put("&lrarr2;","&lrarr;");
		nonStandardEntities.put("&1t;",    "&lt;");
		nonStandardEntities.put("&#micro;","&micro;");
		nonStandardEntities.put("&minsu;","-");
		nonStandardEntities.put("&)minus;","-");
		nonStandardEntities.put("&MINUS;", "-");
		nonStandardEntities.put("&&num;",  "&num;");
		nonStandardEntities.put("&plus);", "+");
		nonStandardEntities.put("&rcolon;",";");
		nonStandardEntities.put("&rlarr;", "&lrarr;");
		nonStandardEntities.put("&RSQUO;", "&rsquo;");
		nonStandardEntities.put("&sub2;",  "<sub>2</sub>");
		nonStandardEntities.put("&sup35;", "<sub>35</sub>");
		nonStandardEntities.put("&z.ccirf;","c");
		nonStandardEntities.put("&z.Gt;",  "&gt;");
		nonStandardEntities.put("&z.Lt;",  "&lt;");
		nonStandardEntities.put("&z.rad;", "&radic;");
		
		// Unknown translations
		// &#x100172; // Even with decimal 100172 it's an undefined unicode character
		// &bond;
		// &bondDot;
		// &crcldt;
		// &dbd;
		// &dbond;
		// &elpar;
		// &f1;
		// &inf1; to &inf6;
		// &infBD;
		// &infBR;
		// &infMH;
		// &lapr;
		// &mean;
		// &parr0;
		// &.rpar;
		// &sbd;
		// &sbond;
		// &sqbl0;
		// &sqbr0;
		// &subd;
		// &supCS;
		// &supQS;
		// &supWestmead;
		// &tbd1;
	}
	
	private static final List<String> spaces = new ArrayList<String>();
	static{
		spaces.add(" "); // regular space
		spaces.add("\t"); // tab
		spaces.add("\u00A0"); // no-break space
		spaces.add("\u2002"); // &ensp;
		spaces.add("\u2003"); // &emsp;
		spaces.add("\u2004"); // three-per-em space
		spaces.add("\u2005"); // four-per-em space
		spaces.add("\u2006"); // six-per-em space
		spaces.add("\u2007"); // figure space
		spaces.add("\u2008"); // punctuation space
		spaces.add("\u2009"); // thin space
		spaces.add("\u200A"); // hair space
		spaces.add("\u200B"); // zero width space
		spaces.add("\u2028"); // line separator
		spaces.add("\u2029"); // paragraph separator
		spaces.add("\u202F"); // narrow no-break space
		spaces.add("\u205F"); // medium mathematical space
		spaces.add("\u3000"); // ideographic space
		spaces.add("\u303F"); // ideographic half fill space
		spaces.add("\uFEFF"); // zero width no-break space
		spaces.add("\uE0020"); // tag space
	}
	
	public static String convertIrregularSpacesToRegularSpaces(String inputString){
		String returnString = "";
		
		for (int i = 0; i < inputString.length(); ++i){
			String oneChar = inputString.substring(i, i + 1);
			
			if (isSpace(oneChar)){
				returnString += " ";
			}
			else{
				returnString += oneChar;
			}
		}
		
		return returnString;
	}
	
	public static String charToUnicode(char oneChar){
		return ("\\u" + Integer.toHexString(oneChar | 0x10000).substring(1));
	}
	
	public static String charToUnicode(String stringWithOneChar){
		return charToUnicode(stringWithOneChar.charAt(0));
	}

	public static Pair<String, Boolean> cleanUpText(String inputString, String debugId){
		String returnString = EmptyStringUtf8.getEmptyString();
		returnString += inputString;
		boolean success = true;
		Pattern pattern = Pattern.compile("&\\S+?;"); // HTML entities

		// We have to do more than one round because some text is "doubly"
		// encoded, e.g. "&amp;Bgr;&lt;up>A&lt;/up>"
		for (int i=0; i < numberOfRounds; ++i){
			//----------------------------------------
			// Remove non-standard HTML entities not handled by existing utility
			Matcher matcher = pattern.matcher(returnString);
			StringBuffer tempBuffer = new StringBuffer();
			boolean changed = false;
			while (matcher.find()){
				String entity = matcher.group();
				if (nonStandardEntities.containsKey(entity)){
					changed = true;
					matcher.appendReplacement(tempBuffer, nonStandardEntities.get(entity));
				}
			}
			if (changed){
				matcher.appendTail(tempBuffer);
				returnString = tempBuffer.toString();
			}

			//----------------------------------------
			// Remove special HTML entities via existing utility
			// Note that only HTML 4.0 entities are supported
			returnString = StringEscapeUtils.unescapeHtml4(returnString);

			//----------------------------------------
			// Convert all types of spaces into a simple space
			for (String space : spaces){
				returnString = returnString.replace(space, " ");
			}

			//----------------------------------------
			// Consolidate multiple spaces to a single space
			returnString = returnString.replaceAll("\\s+", " ");

			//----------------------------------------
			// Remove leading and trailing spaces
			returnString = returnString.trim();
		}
		
		//----------------------------------------
		// See if the string still has untreated HTML entities
		// If yes, print it out so we know about it
		
		// HTML entities but without capital-letters-only and digits-only occurrences
		// E.g. &C; and &PD; and &2; are most likely valid text not meant to be HTML entities
		// Thus use a pattern to avoid false alarms
		Pattern patternCapsOnly = Pattern.compile("^&[A-Z][A-Z]?;$"); // 1 or 2 capital letters
		Pattern patternDigitsOnly = Pattern.compile("^&[0-9]+;$");
		Matcher matcher = pattern.matcher(returnString);
		while (matcher.find()){
			String temp = matcher.group();
			Matcher matcherCapsOnly = patternCapsOnly.matcher(temp);
			Matcher matcherDigitsOnly = patternDigitsOnly.matcher(temp);
			if (!matcherCapsOnly.matches() && !matcherDigitsOnly.matches()){
				success = false;
				logger.warn("Not cleaned for " + debugId + ": " + temp);
			}
		}
		
		//----------------------------------------
		// All done, return fixed string
		return new MutablePair<String, Boolean>(returnString, Boolean.valueOf(success));
	}
	
	public static boolean isSpace(String c){
		boolean returnBool = false;
		if (spaces.contains(c)){
			returnBool = true;
		}
		return returnBool;
	}
	
	public static boolean isSpace(char c){
		return isSpace(Character.toString(c));
	}
	
	/** Expected input format: \u1234 */
	public static String unicodeToChar(String unicode){
		String numericalPart = unicode.substring(2);
		int decimal = Integer.parseInt(numericalPart, 16);
		return ((char)decimal + "");
	}
}
