/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.util;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;

public class ReadFromFile{
	
	protected static final Logger logger = Logger.getLogger(ReadFromFile.class);
	
	/**
	 * Given a filename, open a buffered reader in UTF-8 format.
	 * This method may be useful with the file is big, and readFileToListOfStrings()
	 * may exhaust the memory. After getting a BufferedReaderUtf8 instance, call
	 * BufferedReaderUtf8.readLine() to get each line, e.g.:
	 * String line = "";
	 * while ((line = br.readLine()) != null){
	 *   ... do something ...
	 * }
	 * Don't forget to close the reader.
	 */
	public static BufferedReaderUtf8 openFileAsBufferedReaderUtf8(String filename) throws IOException{
		BufferedReaderUtf8 br = new BufferedReaderUtf8(filename, BufferedReaderUtf8.FILENAME);
		return br;
	}
	
	/**
	 * Given a filename, read all the content into a list of integers, one
	 * line per entry.
	 * @param filename
	 * @return a list of integers, each integer is a line from the input file
	 * @throws IOException 
	 */
	public static List<Integer> readFileToListOfIntegers(String filename) throws IOException{
		List<Integer> returnList = new ArrayList<Integer>();
		BufferedReaderUtf8 br = new BufferedReaderUtf8(filename, BufferedReaderUtf8.FILENAME);
		String line = "";
		while ((line = br.readLine()) != null){
			returnList.add(Integer.valueOf(line));
		}
		br.close();
		return returnList;
	}
	
	/**
	 * Given a filename, read all the content into a list of longs, one
	 * line per entry.
	 * @param filename
	 * @return a list of longs, each long is a line from the input file
	 * @throws IOException 
	 */
	public static List<Long> readFileToListOfLongs(String filename) throws IOException{
		List<Long> returnList = new ArrayList<Long>();
		BufferedReaderUtf8 br = new BufferedReaderUtf8(filename, BufferedReaderUtf8.FILENAME);
		String line = "";
		while ((line = br.readLine()) != null){
			returnList.add(Long.valueOf(line));
		}
		br.close();
		return returnList;
	}
	
	/**
	 * Given a filename, read all the content into a list of strings, one
	 * line per entry.
	 * @param filename
	 * @return a list of strings, each string is a line from the input file
	 * @throws IOException 
	 */
	public static List<String> readFileToListOfStrings(String filename) throws IOException{
		List<String> returnList = new ArrayList<String>();
		BufferedReaderUtf8 br = new BufferedReaderUtf8(filename, BufferedReaderUtf8.FILENAME);
		String line = "";
		while ((line = br.readLine()) != null){
			returnList.add(line);
		}
		br.close();
		return returnList;
	}
	
	/**
	 * Same as readFileToListOfStrings(), except all lines starting with '//'
	 * are treated as comments and omitted.
	 * @param filename
	 * @return
	 * @throws IOException
	 */
	public static List<String> readFileToListOfStringsOmittingComments(String filename) throws IOException{
		List<String> returnList = readFileToListOfStrings(filename);
		int i = 0;
		while (i < returnList.size()){
			if (returnList.get(i).startsWith("//")){
				returnList.remove(i);
			}
			else{
				++i;
			}
		}
		return returnList;
	}
	
	/**
	 * Given a filename, read all the content into a string.
	 * @param filename
	 * @return a string, including all newlines in the input file
	 * @throws IOException 
	 */
	public static String readFileToString(String filename) throws IOException{
		List<String> tempList = readFileToListOfStrings(filename);
		return StringUtils.join(tempList, "\n");
	}
}
