/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.util;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

public class SaveToFile {
	
	protected static final Logger logger = Logger.getLogger(SaveToFile.class);
	
	/**
	 * Append all the strings in the cache list to the given filename, one
	 * string per line.
	 * @param cache
	 * @param filename
	 * @throws IOException
	 */
	public static void appendListOfStringsToFile(List<String> strings, String filename) throws IOException{
		BufferedWriterUtf8 bw = new BufferedWriterUtf8(filename, BufferedWriterUtf8.FILENAME,
				BufferedWriterUtf8.APPEND);
		for (String string : strings){
			bw.writeLine(string);
		}
		bw.close();
	}
	
	/**
	 * Append the given string as-is to a given filename.
	 * 
	 * Notice that there is no extra newline added to the end of the string.
	 * @param string
	 * @param filename
	 * @throws IOException
	 */
	public static void appendStringToFile(String string, String filename) throws IOException{
		BufferedWriterUtf8 bw = new BufferedWriterUtf8(filename, BufferedWriterUtf8.FILENAME,
				BufferedWriterUtf8.APPEND);
		bw.write(string);
		bw.close();
	}
	
	/**
	 * If the filename, as an absolute path, contains any folders that do not
	 * yet exist, create them.
	 * @param filename
	 */
	public static void createMissingFolders(String absolutePath){
		//-----------------------------------------
		// Get the path minus the actual filename
		int index = absolutePath.lastIndexOf("/");
		if (index >= 0){
			String path = absolutePath.substring(0, index);
			
			//-----------------------------------------
			// Create all necessary folders in the path
			File file = new File(path);
			file.mkdirs();
		}
	}
	
	/**
	 * Given a filename, open a buffered writer in UTF-8 format.
	 * Don't forget to close the writer.
	 */
	public static BufferedWriterUtf8 openFileAsBufferedWriterUtf8(String filename) throws IOException{
		BufferedWriterUtf8 bw = new BufferedWriterUtf8(filename, BufferedWriterUtf8.FILENAME);
		return bw;
	}
	
	/**
	 * Save all the strings in the cache list to the given filename, one
	 * string per line.
	 * @param cache
	 * @param filename
	 * @throws IOException
	 */
	public static void saveListOfStringsToFile(List<String> strings, String filename) throws IOException{
		createMissingFolders(filename);
		BufferedWriterUtf8 bw = new BufferedWriterUtf8(filename, BufferedWriterUtf8.FILENAME);
		for (String string : strings){
			bw.writeLine(string);
		}
		bw.close();
	}
	
	/**
	 * Save all the integers in the cache list to the given filename, one
	 * integer per line.
	 * @param cache
	 * @param filename
	 * @throws IOException
	 */
	public static void saveListOfIntegersToFile(List<Integer> integers, String filename) throws IOException{
		createMissingFolders(filename);
		BufferedWriterUtf8 bw = new BufferedWriterUtf8(filename, BufferedWriterUtf8.FILENAME);
		for (int integer : integers){
			bw.writeLine(integer + "");
		}
		bw.close();
	}
	
	/**
	 * Save a map to a file with delimiters separating the keys and values.
	 * @throws IOException 
	 */
	public static void saveMapToDelimitedFile(Map<? extends Object, ? extends Object> map, String delimiter, String filename)
			throws IOException{
		createMissingFolders(filename);
		BufferedWriterUtf8 bw = new BufferedWriterUtf8(filename, BufferedWriterUtf8.FILENAME);
		
		// Simply iterate thru the map, without doing anything about key order.
		for (Map.Entry<? extends Object, ? extends Object> entry : map.entrySet()){
			bw.writeLine(entry.getKey().toString() + delimiter + entry.getValue().toString());
		}
		bw.close();
	}
	
	/**
	 * Save the given string as-is to a given filename.
	 * 
	 * Notice that there is no extra newline added to the end of the string.
	 * @param string
	 * @param filename
	 * @throws IOException
	 */
	public static void saveStringToFile(String string, String filename) throws IOException{
		createMissingFolders(filename);
		BufferedWriterUtf8 bw = new BufferedWriterUtf8(filename, BufferedWriterUtf8.FILENAME);
		bw.write(string);
		bw.close();
	}
}
