/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.wrapper.morphadorner;

import java.util.Arrays;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;

import edu.northwestern.at.utils.corpuslinguistics.inflector.pluralizer.EnglishPluralizer;

public class WordPluralizer{

	protected static final Logger logger = Logger.getLogger(WordPluralizer.class);
	
	private static boolean initialized = false;
	private static EnglishPluralizer pluralizer = null;
	
	public static void initialize(){
		if (!initialized){
			pluralizer = new EnglishPluralizer();
			
			initialized = true;
		}
	}
	
	public static boolean isPlural(String word){
		// Not returning !isSingular() at the moment because that
		// method is very crude.
		return !singularize(word).isEmpty();
	}
	
	public static boolean isSingular(String word){
		boolean returnBool = true;
		
		//-----------------------------------------
		// Assumption: the input word is an English word, listed as
		// a noun in WordNet
		
		//-----------------------------------------
		// A word ending in a letter that is not "s" is singular.
		// So investigate further if the word ends in "s".
		word = word.toLowerCase();
		if (word.endsWith("s")){
			//-----------------------------------------
			// Use a simple rule: If the last two letters are "ss",
			// the word is singular.
			if (!word.endsWith("ss")){
				returnBool = false;
			}
		}

		return returnBool;
	}
	
	public static String pluralize(String word){
		String returnString = "";
		
		if (StringUtils.isAllLowerCase(word)){
			returnString = pluralizer.pluralize(word);
		}
		
		//-----------------------------------------
		// MorphAdorner pluralizer bombs out when the word contains
		// strange characters like semicolons. So don't do anything
		// if the input string is not a single word.
		else if (!StringUtils.isAlphanumericSpace(word)){
			// Do nothing, will return empty string
		}
		
		//-----------------------------------------
		// MorphAdorner pluralizer does not work with upper case letter.
		// So we need to get plural form from the lower case variant
		// and then modify the original word accordingly.
		else{
			//-----------------------------------------
			// Convert the input word to lower case, and get
			// plural format
			String tempWord = word.toLowerCase();
			tempWord = pluralizer.pluralize(tempWord);
			
			//-----------------------------------------
			// See if there are any letters different / added
			// when making the plural format
			int positionOfDifference = -1;
			int i = 0;
			while (i < word.length() && positionOfDifference < 0){
				String originalChar = word.substring(i, i + 1).toLowerCase();
				String pluralChar = tempWord.substring(i, i + 1);
				
				if (!originalChar.equals(pluralChar)){
					positionOfDifference = i;
				}
				
				++i;
			}
			
			//-----------------------------------------
			// Case 1: Pluralizer only added extra chars.
			// So append these extra chars, in upper case if appropriate.
			if (positionOfDifference < 0){
				int numberOfExtraChars = tempWord.length() - word.length();
				String extraChars = tempWord.substring(tempWord.length() - numberOfExtraChars);
				
				String lastCharInOriginalWord = word.substring(word.length() - 1);
				if (StringUtils.isAllUpperCase(lastCharInOriginalWord)){
					extraChars = extraChars.toUpperCase();
				}
				
				returnString = word + extraChars;
			}
			
			//-----------------------------------------
			// Case 1: Pluralizer completely replaced the original word
			// with a new word (e.g. "I" --> "we").
			// So replace the original word with the new word, in upper
			// case if appropriate.
			else if (positionOfDifference == 0){
				// For "I", we cannot tell if the plural should be "we" or "We" or "WE".
				// Most of the time, it would be "we", which is what the pluralizer
				// gives. So return that.
				if (word.equals("I")){
					// Nothing to do
				}
				else{
					String lastCharInOriginalWord = word.substring(word.length() - 1);
					if (StringUtils.isAllUpperCase(lastCharInOriginalWord)){
						tempWord = tempWord.toUpperCase();
					}
				}
				
				returnString = tempWord;
			}
			
			//-----------------------------------------
			// Case 3: Pluralizer modified some existing chars.
			// So from the original word, remove the appropriate chars,
			// and append the suffix, in upper case if appropriate.
			else{
				String unchangedPrefixOfOriginalWord = word.substring(0, positionOfDifference);
				String changedSuffix = tempWord.substring(positionOfDifference);
				
				String lastCharInUnchangedPrefix = unchangedPrefixOfOriginalWord.substring(unchangedPrefixOfOriginalWord.length() - 1);
				if (StringUtils.isAllUpperCase(lastCharInUnchangedPrefix)){
					changedSuffix = changedSuffix.toUpperCase();
				}
				
				returnString = unchangedPrefixOfOriginalWord + changedSuffix;
			}
		}
		
		return returnString;
	}
	
	public static String pluralizeLastWord(String wordsAsString){
		String returnString = "";
		
		List<String> words = Arrays.asList(StringUtils.split(wordsAsString, " "));
		int numberOfWords = words.size();
		String lastWord = words.get(numberOfWords - 1);
		String lastWordInPlural = pluralize(lastWord);
		if (!lastWordInPlural.isEmpty() && !lastWordInPlural.equalsIgnoreCase(lastWord)){
			words.set(numberOfWords - 1, lastWordInPlural);
			returnString = StringUtils.join(words, " ");
		}
		
		return returnString;
	}
	
	public static String singularize(String word){
		String returnString = "";
		
		//-----------------------------------------
		// A crude way to change a word to singular format.
		// First guess if the input word is in plural format.
		// If yes, guess the singular format, and return it.
		// If not, return empty string.
		word = word.toLowerCase();
			
		//-----------------------------------------
		// Check a few cases
		// Should match with the cases in isPlural()
		// Also see isPlural() for sample matches
		int length = word.length();
		if (word.endsWith("ies") && length > 3){ // E.g. allergies
			returnString = word.substring(0, length - 3) + "y";
		}
		else if (word.endsWith("sses")){ // E.g. processes
			returnString = word.substring(0, length - 2);
		}
		else if (word.matches("^.+[^aeiou]es$")){ // E.g. cakes
			returnString = word.substring(0, length - 1);
		}
		else if (word.matches("^.+[^aeious]s$")){ // E.g. prisms
			returnString = word.substring(0, length - 1);
		}
		else if (word.matches("^.+[aeo]s$")){ // E.g. juntas but not virus, synthesis
			returnString = word.substring(0, length - 1);
		}
		
		return returnString;
	}
}
