/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.wrapper.umls;

import org.apache.commons.lang3.tuple.MutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

public class CuiAndVersion{
	
	protected static final Logger logger = Logger.getLogger(CuiAndVersion.class);
	
	//-----------------------------------------
	// In summary:
	//
	// As an int: Chop off leading "C" in a CUI, and convert
	//            the rest as an integer. No version information.
	//            E.g. C0012345 --> 12345
	//
	// As a long: Incorcoprate version number. Multiple the CUI's
	//            int value by 1000, then append version number
	//            as a 3-digit string
	//            E.g. C0012345, version 67 --> 12345067
	
	public static boolean isCui(String inputString){
		return inputString.matches("^C[0-9]{7,7}$");
	}
	
	public static String toCui(int cuiAsInt){
		String cui = cuiAsInt + "";
		while (cui.length() < 7){
			cui = "0" + cui;
		}
		cui = "C" + cui;
		return cui;
	}
	
	public static String toCui(long cuiAndVersion){
		return toCui(toInt(cuiAndVersion));
	}
	
	public static int toInt(String cui){
		return Integer.valueOf(cui.substring(1));
	}
	
	public static int toInt(long cuiAndVersion){
		return (int) (cuiAndVersion / 1000);
	}
	
	public static long toLong(String cui){
		// Default version number is 0
		return toLong(cui, 0);
	}
	
	public static long toLong(String cui, int version){
		//-----------------------------------------
		// Combine the CUI's numerical portion, and the
		// version as a three-digit number, to get a
		// single integer.
		// E.g. C0000123 version 4 --> 123004
		// E.g. C1234567 version 135 --> 1234567135
		
		//-----------------------------------------
		// The the numerical part of CUI
		int cuiAsInt = toInt(cui);
		
		return (long) cuiAsInt * 1000 + version;
	}
	
	public static Pair<String, Integer> toPair(long cuiAndVersion){
		String cui = toCui(cuiAndVersion);
		int version = toVersion(cuiAndVersion);

		return new MutablePair<String, Integer>(cui, version);
	}
	
	public static int toVersion(long cuiAndVersion){
		//-----------------------------------------
		// Extract the version part
		// Note to self: Be sure to NOT do this!:
		// int version = (int) cuiAndVersion % 1000;
		// Because that might alter the value of
		// cuiAndVersion and, subsequently, the
		// remainder.
		return (int) (cuiAndVersion % 1000);
	}
}