/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.wrapper.umls;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import mpii.biotm.CommandLineArguments;
import mpii.biotm.util.BufferedReaderUtf8;
import mpii.biotm.util.BufferedWriterUtf8;
import mpii.biotm.util.InsertTabs;
import mpii.biotm.util.ProcessUpperLowerCase;
import mpii.biotm.util.SaveToFile;
import mpii.biotm.wrapper.morphadorner.WordPluralizer;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.MutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.log4j.Logger;

public class EstimateMostCommonSenseCuisConceptNames{
	//-----------------------------------------
	// I noticed that many text mentions were recognized as certain
	// CUIs because the text mention is ambiguous in the following
	// (example) way:
	// C0202348 is "caffeine level test", "caffeine measurement", "caffeine"
	// C0006644 is "caffeine"
	// So for "caffeine" to map to the test is usually bad. Which brings
	// us to the next observation: very often, a word's most common sense
	// applies to most occurrences. "Caffeine" on its own rarely means the
	// test. In other words, the prior propability of the most common sense
	// usually dominates the other senses by far.
	//
	// Therefore, this class aims to do the following:
	// (1) For a given word / string, estimate which CUI represents the
	//     most common sense. Do this by examining the distribution of
	//     concept names within a CUI, and in between CUIs. The main
	//     heuristic is that the most common sense CUI has many occurrences
	//     of "caffeine", while the caffeine test CUI has "caffeine" as an
	//     abbreviation of "caffeine level test" AND only one occurrence
	//     of "caffeine" within this test-meaning CUI.
	// (2) Output two lists, the CUI / concept name pairs that we deem to
	//     have most common senses, and those from which we cannot draw any
	//     conclusion.
	// (3) Output an evaluation file so we can easily examine what got put
	//     into which bucket.
	//
	// Another thing that stringsim missed: In C0425965 "possible pregnancy",
	// the concept name "? pregnancy" should not have had the leading
	// question mark removed. (It turns out there are, as of UMLS2013AB, only
	// 3 CUIs starting with a question mark. Unfortunately one of them is a
	// popular entity "pregnancy".)
	//
	// As input, take the CUIs + concept names previously generated for
	// paraphrasing. These are the reasonably unique concept names under
	// a given CUI, the concept names are not too long, etc. See
	// mpii.biotm.bebe.paraphrase.precompute.GenerateEntityNamesForParaphrasing
	// for details.
	
	protected static final Logger logger = Logger.getLogger(EstimateMostCommonSenseCuisConceptNames.class);
	
	// The input file should be the output of
	// mpii.biotm.bebe.paraphrase.precompute.GenerateEntityNamesForParaphrasing
	// where duplicate / obviously unsuitable concept names are already removed
	private static String inputFilename = "";

	// Keep track of data we derive from MRCONSO
	private static Map<Integer, List<Pair<String, String>>> cuisAsIntsToConceptNamesAndSabsFromMrconso = null;
	private static Map<Integer, List<String>> cuisAsIntsToMostPopularConceptNamesFromMrconso = null;
	private static Map<Integer, Integer> cuisAsIntsToMostPopularConceptNameCountsFromMrconso = null;

	// Keep track of preprocessed data we read in
	private static Map<String, Set<Integer>> conceptNamesToCuisAsIntsFromPreprocessedData = null;
	private static Map<Integer, Set<String>> cuisAsIntsToConceptNamesFromPreprocessedData = null;
	
	// Keep track of our target data -- the most common sense CUI / concept name combinations
	private static Map<String, Integer> conceptNamesToMostCommonSenseCuisAsInts = null;
	private static Map<Integer, String> mostCommonSenseCuisAsIntsToConceptNames = null;
	
	// Other variables
	private static int maxNumberOfChars = -1;
	private static int maxNumberOfWords = -1;
	private static String mrconsoFilename = "";
	private static String outputFilenameMostPopularConceptNames = "";
	private static String outputFilenameEvaluationMostCommonSensesFound = "";
	private static String outputFilenameEvaluationNoConclusion = "";
	private static String outputFilenameMostCommonSenses = "";
	private static String outputFilenameNoConclusion = "";
	
	private static void determineMostCommonSense(){
		logger.info("Determining most common sense for concept names...");

		//-----------------------------------------
		// Go to each concept name
		int count = 0;
		for (String conceptName : conceptNamesToCuisAsIntsFromPreprocessedData.keySet()){
			List<Integer> cuisAsInts = new ArrayList<Integer>(conceptNamesToCuisAsIntsFromPreprocessedData.get(conceptName));
			int numberOfCuis = cuisAsInts.size();
			
			//-----------------------------------------
			// If the concept name is only associated with one CUI,
			// then that CUI is the most common sense.
			if (numberOfCuis == 1){
				int cuiAsInt = cuisAsInts.get(0);
				conceptNamesToMostCommonSenseCuisAsInts.put(conceptName, cuiAsInt);
				mostCommonSenseCuisAsIntsToConceptNames.put(cuiAsInt, conceptName);
			}
			
			//-----------------------------------------
			// The concept name is associated with two or more CUIs
			else{
				//-----------------------------------------
				// Determine which CUI (in terms of the index in cuisAsInts list)
				// is the winner
				int winningIndex = determineWinningIndex(conceptName, cuisAsInts);
				
				//-----------------------------------------
				// Found the winner
				if (winningIndex >= 0){
					int winningCuiAsInt = cuisAsInts.get(winningIndex);
					conceptNamesToMostCommonSenseCuisAsInts.put(conceptName, winningCuiAsInt);
					mostCommonSenseCuisAsIntsToConceptNames.put(winningCuiAsInt, conceptName);
				}
				// else - No winner found, nothing to do
			}
			
			//-----------------------------------------
			// Report interim progress
			++count;
			if (count > 0 && count % 500000 == 0){
				logger.info(count + " concept names processed...");
			}
		} // for each concept name
	}
	
	private static int determineWinningIndex(String conceptName, List<Integer> cuisAsInts){
		int winningIndex = -1;
		
		//-----------------------------------------
		// Base on a few rules, decide if we have a most
		// common sense CUI. Take care to make sure if all
		// concept names are considered not most common sense,
		// we don't declare a winner (otherwise, when using the
		// results of this class, we may end up with a text
		// mention whose CUIs are all considered bad and thus
		// discarded).
		//
		// Return the index of the most common sense CUI found.
		// Return -1 if no winner is found.
		
		//-----------------------------------------
		// Set up a few variables that we'll need later
		int numberOfCuis = cuisAsInts.size();
		boolean conceptNameIsMoreThanCamelCase = false;
		List<Boolean> isAbbreviationFlags = new ArrayList<Boolean>(); // Think sub string
		List<Boolean> isAcronymFlags = new ArrayList<Boolean>();
		List<Boolean> isMostPopularFlags = new ArrayList<Boolean>();
		List<Integer> numbersOfSameConceptNames = new ArrayList<Integer>();
		
		//-----------------------------------------
		// If the concept name is more than camel case (i.e. likely to be an acronym),
		// and there is only one acronym, that's the winner
		if (winningIndex == -1){
			conceptNameIsMoreThanCamelCase = ProcessUpperLowerCase.moreThanCamelCase(conceptName);
			
			if (conceptNameIsMoreThanCamelCase){
				for (int i = 0; i < numberOfCuis; ++i){
					int cuiAsInt = cuisAsInts.get(i);
					isAcronymFlags.add(isAcronym(cuiAsInt, conceptName));
				}

				int numberOfTrues = getNumberOfTrues(isAcronymFlags);
				if (conceptNameIsMoreThanCamelCase && numberOfTrues == 1){
					int i = 0;
					while (i < numberOfCuis && winningIndex == -1){
						if (isAcronymFlags.get(i)){
							winningIndex = i;
						}
						++i;
					}
				}
			}
		}
		
		//-----------------------------------------
		// If there is only one most popular item, that's the winner
		if (winningIndex == -1){
			for (int i = 0; i < numberOfCuis; ++i){
				int cuiAsInt = cuisAsInts.get(i);
				isMostPopularFlags.add(isMostPopular(cuiAsInt, conceptName));
			}
			
			int numberOfTrues = getNumberOfTrues(isMostPopularFlags);
			if (numberOfTrues == 1){
				int i = 0;
				while (i < numberOfCuis && winningIndex == -1){
					if (isMostPopularFlags.get(i)){
						winningIndex = i;
					}
					++i;
				}
			}
		}
		
		//-----------------------------------------
		// If there is only one item that is NOT an abbreviation,
		// that's the winner
		if (winningIndex == -1){
			for (int i = 0; i < numberOfCuis; ++i){
				int cuiAsInt = cuisAsInts.get(i);
				isAbbreviationFlags.add(isAbbreviation(cuiAsInt, conceptName));
			}
			
			int numberOfTrues = getNumberOfTrues(isAbbreviationFlags);
			if (numberOfTrues == numberOfCuis - 1){
				int i = 0;
				while (i < numberOfCuis && winningIndex == -1){
					if (!isAbbreviationFlags.get(i)){
						winningIndex = i;
					}
					++i;
				}
			}
		}
		
		//-----------------------------------------
		// If we reach here, there was no single winner based on
		// most popular flags alone. The true answer could be the
		// second or third (etc.) most popular concept name. So look
		// as each of the CUIs. Count, for each CUI, the number of
		// occurrences (regardless of repeating with same SAB, and
		// treating singular/plural as the same) of the concept name.
		// If there is a single CUI with the highest count, use that
		// CUI as winner.
		if (winningIndex == -1){
			for (int i = 0; i < numberOfCuis; ++i){
				int cuiAsInt = cuisAsInts.get(i);
				numbersOfSameConceptNames.add(getNumberOfSameConceptName(cuiAsInt, conceptName));
			}
			
			int max = 0;
			int tempWinningIndex = -1;
			int numberOfWinners = 0;
			for (int i = 0; i < numbersOfSameConceptNames.size(); ++i){
				int numberOfSameConceptName =  numbersOfSameConceptNames.get(i);
				if (numberOfSameConceptName > max){
					max = numberOfSameConceptName;
					tempWinningIndex = i;
					numberOfWinners = 1;
				}
				else if (numberOfSameConceptName == max){
					++numberOfWinners;
				}
			}
			
			if (numberOfWinners == 1){
				winningIndex = tempWinningIndex;
			}
		}
		
		return winningIndex;
	}

	private static boolean differBySingularPlural(String conceptName1, String conceptName2){
		return (differBySingularPluralOneDirection(conceptName1, conceptName2) ||
				differBySingularPluralOneDirection(conceptName2, conceptName1));
	}
	
	private static boolean differBySingularPluralOneDirection(String conceptName1, String conceptName2){
		boolean returnBool = false;
		
		//-----------------------------------------
		// Look at the last word in concept name 1.
		// If it's singular, make it plural.
		// Then compare to concept name 2.
		// Return true if now both concepts are the same.
		
		//-----------------------------------------
		// Divide concept name 1 into prefix and last word
		String prefix = "";
		String lastWord = "";
		
		int lastSpacePosition = conceptName1.lastIndexOf(" ");
		if (lastSpacePosition > 0){
			prefix = conceptName1.substring(0, lastSpacePosition + 1);
			lastWord = conceptName1.substring(lastSpacePosition + 1);
		}
		else{
			lastWord = conceptName1;
		}
		
		//-----------------------------------------
		// Change the singular/plural form of last word
		if (WordPluralizer.isSingular(lastWord)){
			lastWord = WordPluralizer.pluralize(lastWord);
			
			conceptName1 = prefix + lastWord;
			if (conceptName1.equals(conceptName2)){
				returnBool = true;
			}
		}
		
		return returnBool;
	}
	
	private static String getInfoAbout(int cuiAsInt){
		//-----------------------------------------
		// Generate a string that summarizes the given CUI:
		// - CUI
		// - semantic type(s)
		// - concept names
		StringBuilder stringBuilder = new StringBuilder();
		
		//-----------------------------------------
		// CUI
		stringBuilder.append(CuiAndVersion.toCui(cuiAsInt) + " ");
		
		//-----------------------------------------
		// Semantic type(s)
		List<String> semanticTypes = LookupSemanticType.lookup(cuiAsInt);
		Collections.sort(semanticTypes);
		boolean first = true;
		for (String semanticType : semanticTypes){
			String fullName = LookupSemanticType.getFullName(semanticType);
			if (first){
				first = false;
			}
			else{
				stringBuilder.append(", ");
			}
			stringBuilder.append(semanticType + " (" + fullName + ")");
		}
		stringBuilder.append(" :: ");
		
		//-----------------------------------------
		// Concept names
		List<String> conceptNames = new ArrayList<String>(cuisAsIntsToConceptNamesFromPreprocessedData.get(cuiAsInt));
		Collections.sort(conceptNames);
		first = true;
		for (String conceptName : conceptNames){
			if (first){
				first = false;
			}
			else{
				stringBuilder.append(" :: ");
			}
			stringBuilder.append(conceptName);
		}
		
		return stringBuilder.toString();
	}
	
	private static void getMostPopularConceptNames() throws IOException{
		logger.info("Finding most popular concept names...");
		//-----------------------------------------
		// We are working on the "raw" CUI / concept name / SAB combinations
		// straight from MRCONSO without modification (with too-long concept
		// names already thrown away). The goal of this method
		// is to find out, for each CUI, if there is a most popular concept
		// name. Take into account:
		// - Turn all concept names into lower cases
		// - Unwanted sub strings like "[NOS]"
		
		//-----------------------------------------
		// Go thru each CUI
		int count = 0;
		for (int cuiAsInt : cuisAsIntsToConceptNamesAndSabsFromMrconso.keySet()){
			//-----------------------------------------
			// Count how many times a concept name occurs.
			// Consider concept names in lower cases only.
			// Each concept name can only count once per SAB.
			Map<String, Integer> conceptNamesToCounts = new HashMap<String, Integer>();
			List<Pair<String, String>> conceptNamesAndSabsAlreadySeen = new ArrayList<Pair<String, String>>();
			
			for (Pair<String, String> conceptNameAndSab : cuisAsIntsToConceptNamesAndSabsFromMrconso.get(cuiAsInt)){
				String conceptName = conceptNameAndSab.getLeft();
				conceptName = UnwantedSubStringsInConceptNames.removeUnwantedSubString(conceptName);
				conceptName = conceptName.toLowerCase();
				String sab = conceptNameAndSab.getRight();
				
				//-----------------------------------------
				// In rare cases, the concept name is empty because it only
				// contains unwanted sub strings, e.g. C2897807 NOS
				// So continue only if concept name is not empty
				if (!conceptName.isEmpty()){
					//-----------------------------------------
					// See if we've seen this concept name + SAB combination before
					boolean comboAlreadySeen = false;
					int numberOfCombosToCheck = conceptNamesAndSabsAlreadySeen.size();
					int i = 0;
					while (i < numberOfCombosToCheck && !comboAlreadySeen){
						Pair<String, String> combo = conceptNamesAndSabsAlreadySeen.get(i);
						String conceptNameAlreadySeen = combo.getLeft();
						String sabAlreadySeen = combo.getRight();

						if (sab.equals(sabAlreadySeen) &&
								(conceptName.equalsIgnoreCase(conceptNameAlreadySeen) ||
										differBySingularPlural(conceptName, conceptNameAlreadySeen))){
							comboAlreadySeen = true;
						}

						++i;
					}

					//-----------------------------------------
					// Continue if this concept name and SAB combination
					// has not yet been seen.
					if (!comboAlreadySeen){
						//-----------------------------------------
						// Find the correct concept name to increment count for
						String incrementForConceptName = "";
						List<String> conceptNamesToCompareTo = new ArrayList<String>(conceptNamesToCounts.keySet());
						i = 0;
						while (i < conceptNamesToCounts.size() && incrementForConceptName.isEmpty()){
							String conceptNameToCompareTo = conceptNamesToCompareTo.get(i);

							if (conceptName.equals(conceptNameToCompareTo) ||
									differBySingularPlural(conceptName, conceptNameToCompareTo)){
								incrementForConceptName = conceptNameToCompareTo;
							}

							++i;
						}

						if (incrementForConceptName.isEmpty()){
							conceptNamesToCounts.put(conceptName, 1);
						}
						else{
							int temp = conceptNamesToCounts.get(incrementForConceptName);
							++temp;
							conceptNamesToCounts.put(incrementForConceptName, temp);
						}
						
						//-----------------------------------------
						// Record that we've seen this combo
						conceptNamesAndSabsAlreadySeen.add(new MutablePair<String, String>(conceptName, sab));
					} // if concept name + SAB combo not seen before
				} // if concept name not empty
			} // for each concept name + SAB combo

			//-----------------------------------------
			// Find the concept name(s) that occur max number of times.
			// - If there is 1 concept name with the max number, it's the winner.
			// - If there are 2 or more concept names with the same max number,
			//   they are all winners.
			int max = 0;
			List<String> winningConceptNames = new ArrayList<String>();
			for (String conceptName : conceptNamesToCounts.keySet()){
				int tempCount = conceptNamesToCounts.get(conceptName);
				
				if (tempCount > max){
					max = tempCount;
					winningConceptNames.clear();
					winningConceptNames.add(conceptName);
				}
				else if (tempCount == max){
					winningConceptNames.add(conceptName);
				}
			}
			
			//-----------------------------------------
			// Record the winners
			cuisAsIntsToMostPopularConceptNameCountsFromMrconso.put(cuiAsInt, max);
			Collections.sort(winningConceptNames);
			cuisAsIntsToMostPopularConceptNamesFromMrconso.put(cuiAsInt, winningConceptNames);
			
			//-----------------------------------------
			// Report interim progress
			++count;
			if (count % 10000 == 0){
				logger.info(count + " CUIs processed...");
			}
		} // for each CUI
		
		//-----------------------------------------
		// Save results to file
		List<String> outputLines = new ArrayList<String>();
		List<Integer> cuisAsInts = new ArrayList<Integer>(cuisAsIntsToMostPopularConceptNamesFromMrconso.keySet());
		Collections.sort(cuisAsInts);
		for (int cuiAsInt : cuisAsInts){
			outputLines.add(CuiAndVersion.toCui(cuiAsInt) + "\t"
					+ cuisAsIntsToMostPopularConceptNameCountsFromMrconso.get(cuiAsInt) + "\t"
					+ StringUtils.join(cuisAsIntsToMostPopularConceptNamesFromMrconso.get(cuiAsInt), "~@~"));
		}
		SaveToFile.saveListOfStringsToFile(outputLines, outputFilenameMostPopularConceptNames);
		
		logger.info("Done");
	}
	
	private static int getNumberOfSameConceptName(int cuiAsInt, String conceptNameInFocus){
		int count = 0;
		
		//-----------------------------------------
		// Look at all the "raw" MRCONSO concept names of the given CUI
		// (with too-long concept names already thrown away).
		// Count how many are the same as the concept-name-in-focus.
		// (No need to worry about counting only those not the same as
		// the concept-name-in-focus). Treat upper/lower case as the same.
		// Treat singular/plural as the same.
		if (cuisAsIntsToConceptNamesAndSabsFromMrconso.containsKey(cuiAsInt)){		
			String conceptNameInFocusLowerCase = conceptNameInFocus.toLowerCase();
			List<Pair<String, String>> conceptNamesAndSabs = cuisAsIntsToConceptNamesAndSabsFromMrconso.get(cuiAsInt);
			for (Pair<String, String> conceptNameAndSab : conceptNamesAndSabs){
				String conceptNameLowerCase = conceptNameAndSab.getLeft().toLowerCase();

				if (conceptNameLowerCase.equals(conceptNameInFocusLowerCase) ||
						differBySingularPlural(conceptNameLowerCase, conceptNameInFocusLowerCase)){
					++count;
				}
			}
		}
		
		return count;
	}
	
	private static int getNumberOfTrues(List<Boolean> booleans){
		int sum = 0;
		for (boolean bool : booleans){
			if (bool){
				++sum;
			}
		}
		return sum;
	}

	private static void initialize(Properties settings) throws Exception{
		cuisAsIntsToConceptNamesAndSabsFromMrconso = new HashMap<Integer, List<Pair<String, String>>>();
		cuisAsIntsToMostPopularConceptNamesFromMrconso = new HashMap<Integer, List<String>>();
		cuisAsIntsToMostPopularConceptNameCountsFromMrconso = new HashMap<Integer, Integer>();
		conceptNamesToCuisAsIntsFromPreprocessedData = new HashMap<String, Set<Integer>>();
		cuisAsIntsToConceptNamesFromPreprocessedData = new HashMap<Integer, Set<String>>();
		conceptNamesToMostCommonSenseCuisAsInts = new HashMap<String, Integer>();
		mostCommonSenseCuisAsIntsToConceptNames = new HashMap<Integer, String>();
		LookupSemanticType.initialize(settings);
		UnwantedSubStringsInConceptNames.initialize();
		WordPluralizer.initialize();

		//-----------------------------------------
		// Get filenames
		mrconsoFilename = settings.getProperty("umls_mrconso_filename");
		inputFilename = settings.getProperty("bebe_paraphrase_CUI_version_entity_name_filename");
		outputFilenameMostPopularConceptNames = settings.getProperty("umls_most_common_sense_most_popular_concept_names_filename");
		outputFilenameMostCommonSenses = settings.getProperty("umls_most_common_sense_cuis_concept_names_filename");
		outputFilenameNoConclusion = settings.getProperty("umls_most_common_sense_no_conclusion_filename");
		outputFilenameEvaluationMostCommonSensesFound = settings.getProperty("umls_most_common_sense_evaluation_most_common_senses_found_filename");
		outputFilenameEvaluationNoConclusion = settings.getProperty("umls_most_common_sense_evaluation_no_conclusion_filename");
		
		//-----------------------------------------
		// Get other params
		maxNumberOfChars = Integer.valueOf(settings.getProperty("bebe_string_sim_max_number_of_characters"));
		maxNumberOfWords = Integer.valueOf(settings.getProperty("bebe_string_sim_max_number_of_consecutive_words"));
	}
	
	private static boolean isAbbreviation(int cuiAsInt, String conceptNameInFocus){
		boolean conceptNameInFocusIsAbbreviation = false;
		
		//-----------------------------------------
		// The given CUI has one or more concept names. The given concept
		// name is one of them; call this the concept-name-in-focus. If
		// the concept-name-in-focus is an abbreviation of any other
		// concept names within the CUI, return true.
		//
		// By abbreviation, I mean sub string.
		// TODO better short form recognition than String.contains() ?
		List<String> conceptNames = new ArrayList<String>(cuisAsIntsToConceptNamesFromPreprocessedData.get(cuiAsInt));
		int numberOfConceptNames = conceptNames.size();
		
		//-----------------------------------------
		// If the concept-name-in-focus is a short form of another
		// concept name, that counts as an abbreviation
		int i = 0;
		while (i < numberOfConceptNames && !conceptNameInFocusIsAbbreviation){
			String conceptName = conceptNames.get(i);
			if (!conceptName.equalsIgnoreCase(conceptNameInFocus)){
				if (conceptName.contains(conceptNameInFocus)){
					conceptNameInFocusIsAbbreviation = true;
				}
			}
			++i;
		}
		
		return conceptNameInFocusIsAbbreviation;
	}

	private static boolean isAcceptableConceptName(String conceptName){
		boolean returnBool = false;
		int numberOfWords = conceptName.split("\\W+").length;
		
		//-----------------------------------------
		// Only use those concept names that MinHash pre-computation
		// would also use
		if (conceptName.length() <= maxNumberOfChars &&
				numberOfWords <= maxNumberOfWords &&
				!conceptName.startsWith("?")){
			returnBool = true;
		}
		
		return returnBool;
	}
	
	private static boolean isAcronym(int cuiAsInt, String conceptNameInFocus){
		boolean conceptNameInFocusIsAcronym = false;
		
		//-----------------------------------------
		// The given CUI has one or more concept names. The given concept
		// name is one of them; call this the concept-name-in-focus. If
		// the concept-name-in-focus is an acronym of any other
		// concept names within the CUI, return true.
		List<String> conceptNames = new ArrayList<String>(cuisAsIntsToConceptNamesFromPreprocessedData.get(cuiAsInt));
		int numberOfConceptNames = conceptNames.size();
		
		int i = 0;
		while (i < numberOfConceptNames && !conceptNameInFocusIsAcronym){
			String conceptName = conceptNames.get(i);
			if (!conceptName.equalsIgnoreCase(conceptNameInFocus)){
				if (isAcronymOf(conceptNameInFocus, conceptName)){
					conceptNameInFocusIsAcronym = true;
				}
			}
			++i;
		}
		
		return conceptNameInFocusIsAcronym;
	}
	
	private static boolean isAcronymOf(String conceptName1, String conceptName2){
		//-----------------------------------------
		// Divide concept name 2 into words
		List<String> tokens = Arrays.asList(conceptName2.split("\\W+"));
		
		//-----------------------------------------
		// Make an acronym out of the first letters in
		// each word in concept name 2
		String acronymOfConceptName2 = "";
		for (String token : tokens){
			if (!token.isEmpty()){ // Shouldn't need this line but let's be defensive
				acronymOfConceptName2 += token.substring(0, 1);
			}
		}
		
		//-----------------------------------------
		// Return true if concept name 1 is an acronym of
		// concept name 2
		return (conceptName1.equalsIgnoreCase(acronymOfConceptName2));
	}
	
	private static boolean isMostPopular(int cuiAsInt, String conceptNameInFocus){
		boolean conceptNameInFocusIsMostPopular = false;
		
		//-----------------------------------------
		// If the given concept name is the most popular within the
		// given CUI, return true. Take into account singular/plural
		// variations; e.g. if "apple" is most popular, so is "apples".
		if (cuisAsIntsToMostPopularConceptNamesFromMrconso.containsKey(cuiAsInt)){
			List<String> mostPopularConceptNames = cuisAsIntsToMostPopularConceptNamesFromMrconso.get(cuiAsInt);
			int i = 0;
			while (i < mostPopularConceptNames.size() && !conceptNameInFocusIsMostPopular){
				String mostPopularConceptName = mostPopularConceptNames.get(i);
				if (conceptNameInFocus.equalsIgnoreCase(mostPopularConceptName) ||
						differBySingularPlural(conceptNameInFocus, mostPopularConceptName)){
					conceptNameInFocusIsMostPopular = true;
				}
				++i;
			}
		}
		
		return conceptNameInFocusIsMostPopular;
	}

	private static void readInPreprocessedCuiConceptNames() throws IOException{
		logger.info("Reading in pre-processed CUIs / concept names file...");
		
		//-----------------------------------------
		// Set up file handle
		BufferedReaderUtf8 input = new BufferedReaderUtf8(inputFilename, BufferedReaderUtf8.FILENAME);

		//-----------------------------------------
		// Go thru each input line
		String line = "";
		int count = 0;
		while ((line = input.readLine()) != null){
			//-----------------------------------------
			// Tokenize the string to get the CUI and concept name
			// Note to self: StringUtils.split() treats multiple empty
			// columns as a single column
			List<String> tokens = Arrays.asList(line.split("\\t"));
			String cui = tokens.get(0);
			
			// If the letters in the concept name are more than camel case,
			// leave the concept name unchanged. Otherwise, change the
			// concept name to all lower case.
			String conceptName = toLowerCaseIfNotMoreThanCamelCase(tokens.get(2));
			
			//-----------------------------------------
			// Only use those concept names that MinHash pre-computation
			// would also use
			if (isAcceptableConceptName(conceptName)){
				int cuiAsInt = CuiAndVersion.toInt(cui);
				
				if (!conceptNamesToCuisAsIntsFromPreprocessedData.containsKey(conceptName)){
					conceptNamesToCuisAsIntsFromPreprocessedData.put(conceptName, new HashSet<Integer>());
				}
				conceptNamesToCuisAsIntsFromPreprocessedData.get(conceptName).add(cuiAsInt);
				
				if (!cuisAsIntsToConceptNamesFromPreprocessedData.containsKey(cuiAsInt)){
					cuisAsIntsToConceptNamesFromPreprocessedData.put(cuiAsInt, new HashSet<String>());
				}
				cuisAsIntsToConceptNamesFromPreprocessedData.get(cuiAsInt).add(conceptName);
			}
			
			//-----------------------------------------
			// Report interim progress
			++count;
			if (count % 1000000 == 0){
				logger.info(count + " lines processed...");
			}
		}
		
		//-----------------------------------------
		// Clean up
		input.close();
		logger.info("Done");
	}

	private static void readInMrconsoFile() throws IOException{
		logger.info("Reading in MRCONSO file...");
		//-----------------------------------------
		// Perform no filtering. Simply record each CUI / concept name / SAB
		// combination. Do no alter the concept name in any way, including
		// upper / lower cases.
		
		//-----------------------------------------
		// Set up file handle
		BufferedReaderUtf8 input = new BufferedReaderUtf8(mrconsoFilename, BufferedReaderUtf8.FILENAME);
		
		//-----------------------------------------
		// Go thru each line
		String line = "";
		int count = 0;
		while ((line = input.readLine()) != null){
			
			//-----------------------------------------
			// Tokenize the string to get the CUI and concept name
			// Note to self: StringUtils.split() treats multiple empty
			// columns as a single column
			List<String> tokens = Arrays.asList(line.split("\\|"));
			String conceptName = tokens.get(14);
			
			//-----------------------------------------
			// Skip concept names that are too
			// long, with too many words, etc.
			if (isAcceptableConceptName(conceptName)){
				String cui = tokens.get(0);
				int cuiAsInt = CuiAndVersion.toInt(cui);
				String sab = tokens.get(11);

				//-----------------------------------------
				// If this CUI does not have an entry yet, add one
				if (!cuisAsIntsToConceptNamesAndSabsFromMrconso.containsKey(cuiAsInt)){
					cuisAsIntsToConceptNamesAndSabsFromMrconso.put(cuiAsInt, new ArrayList<Pair<String, String>>());
				}
				
				//-----------------------------------------
				// Simply store the concept name and SAB
				List<Pair<String, String>> conceptNamesAndSabs = cuisAsIntsToConceptNamesAndSabsFromMrconso.get(cuiAsInt);
				Pair<String, String> conceptNameAndSab = new MutablePair<String, String>(conceptName, sab);
				conceptNamesAndSabs.add(conceptNameAndSab);
			}
			
			//-----------------------------------------
			// Report interim progress
			++count;
			if (count > 0 && count % 1000000 == 0){
				logger.info(count + " lines processed...");
			}
		}
		
		//-----------------------------------------
		// Clean up
		input.close();
		logger.info("Done");
	}
	
	private static void saveResults() throws IOException{
		logger.info("Saving results to files...");
		
		//-----------------------------------------
		// Set up file handles
		BufferedWriterUtf8 mostCommonSenseOutput = new BufferedWriterUtf8(outputFilenameMostCommonSenses, BufferedWriterUtf8.FILENAME);
		BufferedWriterUtf8 noConclusionOutput = new BufferedWriterUtf8(outputFilenameNoConclusion, BufferedWriterUtf8.FILENAME);
		
		//-----------------------------------------
		// Get a sorted list of all concept names
		List<String> allConceptNames = new ArrayList<String>(conceptNamesToCuisAsIntsFromPreprocessedData.keySet());
		Collections.sort(allConceptNames);

		//-----------------------------------------
		// Go thru each concept name
		// Depending on whether a most common sense was found,
		// save results in different caches
		Set<String> conceptNamesWithMostCommonSenses = conceptNamesToMostCommonSenseCuisAsInts.keySet();
		for (String conceptName : allConceptNames){
			//-----------------------------------------
			// Most common sense found
			if (conceptNamesWithMostCommonSenses.contains(conceptName)){
				String cui = CuiAndVersion.toCui(conceptNamesToMostCommonSenseCuisAsInts.get(conceptName));
				mostCommonSenseOutput.writeLine(conceptName + "\t" + cui);
			}
			
			//-----------------------------------------
			// No conclusion
			else{
				String cuis = "";
				Set<Integer> cuisAsInts = conceptNamesToCuisAsIntsFromPreprocessedData.get(conceptName);
				for (int cuiAsInt : cuisAsInts){
					if (!cuis.isEmpty()){
						cuis += ", ";
					}
					cuis += CuiAndVersion.toCui(cuiAsInt);
				}
				noConclusionOutput.writeLine(conceptName + "\t" + cuis);
			}
		}
		
		//-----------------------------------------
		// Clean up file handles
		mostCommonSenseOutput.close();
		noConclusionOutput.close();
		
		//-----------------------------------------
		// Dump some info about:
		// (1) The winners and losers, for those
		//     concept names who have most common senses
		// (2) the no-conclusion concept names
		
		//-----------------------------------------
		// Set up file handles
		BufferedWriterUtf8 mostCommonSenseEval = new BufferedWriterUtf8(outputFilenameEvaluationMostCommonSensesFound, BufferedWriterUtf8.FILENAME);
		BufferedWriterUtf8 noConclusionEval = new BufferedWriterUtf8(outputFilenameEvaluationNoConclusion, BufferedWriterUtf8.FILENAME);
		
		//-----------------------------------------
		// Go thru each concept name
		// Depending on whether a most common sense was found,
		// do different things
		for (String conceptName : allConceptNames){
			//-----------------------------------------
			// Most common sense found
			if (conceptNamesWithMostCommonSenses.contains(conceptName)){
				mostCommonSenseEval.writeLine("=== " + conceptName + " ===");
				
				List<String> tempLines = new ArrayList<String>();
				int winningCuiAsInt = conceptNamesToMostCommonSenseCuisAsInts.get(conceptName);
				List<Integer> cuisAsInts = new ArrayList<Integer>(conceptNamesToCuisAsIntsFromPreprocessedData.get(conceptName));
				Collections.sort(cuisAsInts);
				
				tempLines.add("Winner: " + getInfoAbout(winningCuiAsInt));
				for (int cuiAsInt : cuisAsInts){
					if (cuiAsInt != winningCuiAsInt){
						tempLines.add("Loser:  " + getInfoAbout(cuiAsInt));
					}
				}
				
				mostCommonSenseEval.writeLines(InsertTabs.insertTabs(tempLines, 1));
			}
			
			//-----------------------------------------
			// No conclusion
			else{
				noConclusionEval.writeLine("=== " + conceptName + " ===");
				
				List<String> tempLines = new ArrayList<String>();
				List<Integer> cuisAsInts = new ArrayList<Integer>(conceptNamesToCuisAsIntsFromPreprocessedData.get(conceptName));
				Collections.sort(cuisAsInts);
				
				for (int cuiAsInt : cuisAsInts){
					tempLines.add("Loser:  " + getInfoAbout(cuiAsInt));
				}
				
				noConclusionEval.writeLines(InsertTabs.insertTabs(tempLines, 1));
			}
		}		
		
		//-----------------------------------------
		// Clean up file handles
		mostCommonSenseEval.close();
		noConclusionEval.close();
		
		logger.info("Done");
	}
	
	private static String toLowerCaseIfNotMoreThanCamelCase(String string){
		// Disregard non-letter characters. If all the letters are
		// upper case, return the input string unchanged. Otherwise,
		// return the input string all in lower case.
		String returnStr = string;

		if (!ProcessUpperLowerCase.moreThanCamelCase(string)){
			returnStr = string.toLowerCase();
		}
		
		return returnStr;
	}

	public static void main(String[] args) throws Exception{
		//-----------------------------------------
		// Initialize
		Properties settings = CommandLineArguments.getSettings(args);
		initialize(settings);
		
		//-----------------------------------------
		// Read in MRCONSO file
		readInMrconsoFile();

		//-----------------------------------------
		// Get most popular concept name for each CUI
		// Because it takes a while to compute most popular concept names,
		// have two versions of the method:
		// (1) Compute then save results to file
		// (2) Read results from file
		// Either way, the relevant lookup map is populated
		getMostPopularConceptNames();
		
		//-----------------------------------------
		// Read in pre-processed CUIs + concept names
		readInPreprocessedCuiConceptNames();

		//-----------------------------------------
		// Determine for each concept name whether it has a most
		// common sense CUI
		determineMostCommonSense();
	
		//-----------------------------------------
		// Output to files
		saveResults();
	}
}