/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.wrapper.umls;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;

public class LookupSemanticGroup{
	
	protected static final Logger logger = Logger.getLogger(LookupSemanticGroup.class);
	
	// Key = abbreviation, value = full name
	private static final Map<String,String> semanticGroupAbbreviationsToSemanticGroupFullNames = new HashMap<String,String>();
	static{
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("ACTI", "Activities & Behaviors");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("ANAT", "Anatomy");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("CHEM", "Chemicals & Drugs");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("CONC", "Concepts & Ideas");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("DEVI", "Devices");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("DISO", "Disorders");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("GENE", "Genes & Molecular Sequences");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("GEOG", "Geographic Areas");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("LIVB", "Living Beings");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("OBJC", "Objects");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("OCCU", "Occupations");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("ORGA", "Organizations");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("PHEN", "Phenomena");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("PHYS", "Physiology");
		semanticGroupAbbreviationsToSemanticGroupFullNames.put("PROC", "Procedures");
	}
	
	// Key = semantic group, value = semantic types
	private static final Map<String,List<String>> semanticGroupAbbreviationsToSemanticTypeAbbreviations = new HashMap<String,List<String>>();
	static{
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("ACTI",
				new ArrayList<String>(Arrays.asList("acty", "bhvr", "dora", "evnt", "gora", "inbe", "mcha", "ocac",
						"socb")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("ANAT",
				new ArrayList<String>(Arrays.asList("anst", "bdsu", "bdsy", "blor", "bpoc", "bsoj", "celc", "cell",
						"emst", "ffas", "tisu")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("CHEM",
				new ArrayList<String>(Arrays.asList("aapp", "antb", "bacs", "bodm", "carb", "chem", "chvf", "chvs",
						"clnd", "eico", "elii", "enzy", "hops", "horm", "imft", "inch", "irda", "lipd", "nnon",
						"nsba", "opco", "orch", "phsu", "rcpt", "strd", "vita")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("CONC",
				new ArrayList<String>(Arrays.asList("clas", "cnce", "ftcn", "grpa", "idcn", "inpr", "lang", "qlco",
						"qnco", "rnlw", "spco", "tmco")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("DEVI",
				new ArrayList<String>(Arrays.asList("drdd", "medd", "resd")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("DISO",
				new ArrayList<String>(Arrays.asList("acab", "anab", "cgab", "comd", "dsyn", "emod", "fndg", "inpo",
						"mobd", "neop", "patf", "sosy")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("GENE",
				new ArrayList<String>(Arrays.asList("amas", "crbs", "gngm", "mosq", "nusq")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("GEOG",
				new ArrayList<String>(Arrays.asList("geoa")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("LIVB",
				new ArrayList<String>(Arrays.asList("aggp", "amph", "anim", "arch", "bact", "bird", "euka", "famg",
						"fish", "fngs", "grup", "humn", "mamm", "orgm", "plnt", "podg", "popg", "prog", "rept",
						"virs", "vtbt")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("OBJC",
				new ArrayList<String>(Arrays.asList("enty", "food", "mnob", "phob", "sbst")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("OCCU",
				new ArrayList<String>(Arrays.asList("bmod", "ocdi")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("ORGA",
				new ArrayList<String>(Arrays.asList("hcro", "orgt", "pros", "shro")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("PHEN",
				new ArrayList<String>(Arrays.asList("biof", "eehu", "hcpp", "lbtr", "npop", "phpr")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("PHYS",
				new ArrayList<String>(Arrays.asList("celf", "clna", "genf", "menp", "moft", "orga", "orgf", "ortf",
						"phsf")));
		semanticGroupAbbreviationsToSemanticTypeAbbreviations.put("PROC",
				new ArrayList<String>(Arrays.asList("diap", "edac", "hlca", "lbpr", "mbrt", "resa", "topp")));
	}

	// Key = abbreviation of semantic type, value = abbreviation of semantic group
	private static final Map<String,String> semanticTypeAbbreviationsToSemanticGroupAbbreviations = new HashMap<String,String>();
	static{
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("aapp", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("acab", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("acty", "ACTI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("aggp", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("amas", "GENE");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("amph", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("anab", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("anim", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("anst", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("antb", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("arch", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("bacs", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("bact", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("bdsu", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("bdsy", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("bhvr", "ACTI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("biof", "PHEN");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("bird", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("blor", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("bmod", "OCCU");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("bodm", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("bpoc", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("bsoj", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("carb", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("celc", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("celf", "PHYS");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("cell", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("cgab", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("chem", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("chvf", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("chvs", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("clas", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("clna", "PHYS");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("clnd", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("cnce", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("comd", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("crbs", "GENE");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("diap", "PROC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("dora", "ACTI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("drdd", "DEVI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("dsyn", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("edac", "PROC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("eehu", "PHEN");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("eico", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("elii", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("emod", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("emst", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("enty", "OBJC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("enzy", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("euka", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("evnt", "ACTI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("famg", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("ffas", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("fish", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("fndg", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("fngs", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("food", "OBJC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("ftcn", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("genf", "PHYS");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("geoa", "GEOG");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("gngm", "GENE");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("gora", "ACTI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("grpa", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("grup", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("hcpp", "PHEN");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("hcro", "ORGA");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("hlca", "PROC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("hops", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("horm", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("humn", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("idcn", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("imft", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("inbe", "ACTI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("inch", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("inpo", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("inpr", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("irda", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("lang", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("lbpr", "PROC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("lbtr", "PHEN");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("lipd", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("mamm", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("mbrt", "PROC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("mcha", "ACTI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("medd", "DEVI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("menp", "PHYS");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("mnob", "OBJC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("mobd", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("moft", "PHYS");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("mosq", "GENE");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("neop", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("nnon", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("npop", "PHEN");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("nsba", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("nusq", "GENE");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("ocac", "ACTI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("ocdi", "OCCU");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("opco", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("orch", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("orga", "PHYS");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("orgf", "PHYS");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("orgm", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("orgt", "ORGA");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("ortf", "PHYS");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("patf", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("phob", "OBJC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("phpr", "PHEN");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("phsf", "PHYS");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("phsu", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("plnt", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("podg", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("popg", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("prog", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("pros", "ORGA");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("qlco", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("qnco", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("rcpt", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("rept", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("resa", "PROC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("resd", "DEVI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("rnlw", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("sbst", "OBJC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("shro", "ORGA");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("socb", "ACTI");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("sosy", "DISO");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("spco", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("strd", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("tisu", "ANAT");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("tmco", "CONC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("topp", "PROC");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("virs", "LIVB");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("vita", "CHEM");
		semanticTypeAbbreviationsToSemanticGroupAbbreviations.put("vtbt", "LIVB");
	}
	
	public static List<String> getAllSemanticGroups(){
		List<String> semanticGroups = new ArrayList<String>(semanticGroupAbbreviationsToSemanticGroupFullNames.keySet());
		Collections.sort(semanticGroups);
		return semanticGroups;
	}
	
	public static List<String> getSemanticTypes(String semanticGroup){
		return semanticGroupAbbreviationsToSemanticTypeAbbreviations.get(semanticGroup);
	}
	
	public static String getFullName(String semanticGroup){
		return semanticGroupAbbreviationsToSemanticGroupFullNames.get(semanticGroup);
	}
	
	public static String lookup(String semanticType){
		return semanticTypeAbbreviationsToSemanticGroupAbbreviations.get(semanticType);
	}
	
	public static List<String> lookup(List<String> semanticTypes){
		Set<String> semanticGroups = new HashSet<String>();
		
		for (String semanticType : semanticTypes){
			semanticGroups.add(semanticTypeAbbreviationsToSemanticGroupAbbreviations.get(semanticType));
		}
		
		return new ArrayList<String>(semanticGroups);
	}
	
	public static boolean isSemanticGroup(String inputString){
		inputString = inputString.toUpperCase();
		return semanticGroupAbbreviationsToSemanticGroupFullNames.containsKey(inputString);
	}
}