/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.wrapper.umls;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Map.Entry;

import mpii.biotm.util.ReadFromFile;

import org.apache.log4j.Logger;

public class LookupSemanticType{
	
	protected static final Logger logger = Logger.getLogger(LookupSemanticType.class);
	
	// This map contains all CUIs from MRSTY.RRF
	// CUI is an integer because we chop off the leading "C"
	// and convert the rest to a number
	//
	// If a CUI has one semantic type, there is an entry in the
	// first map of the list.
	// If a CUI has three semantic types, there are entries in first
	// three maps of the list, one entry in each map.
	//
	// In each map, key = CUI in integer format, value = semantic type ID
	private static List<Map<Integer, Integer>> cuisToSemanticTypeIds = null;
	
	private static boolean initialized = false;
	
	private static Map<String, Integer> semanticTypeAbbreviationsToSemanticTypeIds = null;
	private static Map<Integer, String> semanticTypeIdsToSemanticTypeAbbreviations = null;
	
	// Key = abbreviation, value = full name
	private static final Map<String,String> semanticTypeAbbreviationsToSemanticTypeFullNames = new HashMap<String,String>();
	static{
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("aapp", "Amino Acid, Peptide, or Protein");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("acab", "Acquired Abnormality");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("acty", "Activity");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("aggp", "Age Group");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("amas", "Amino Acid Sequence");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("amph", "Amphibian");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("anab", "Anatomical Abnormality");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("anim", "Animal");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("anst", "Anatomical Structure");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("antb", "Antibiotic");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("arch", "Archaeon");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("bacs", "Biologically Active Substance");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("bact", "Bacterium");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("bdsu", "Body Substance");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("bdsy", "Body System");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("bhvr", "Behavior");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("biof", "Biologic Function");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("bird", "Bird");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("blor", "Body Location or Region");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("bmod", "Biomedical Occupation or Discipline");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("bodm", "Biomedical or Dental Material");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("bpoc", "Body Part, Organ, or Organ Component");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("bsoj", "Body Space or Junction");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("carb", "Carbohydrate");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("celc", "Cell Component");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("celf", "Cell Function");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("cell", "Cell");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("cgab", "Congenital Abnormality");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("chem", "Chemical");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("chvf", "Chemical Viewed Functionally");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("chvs", "Chemical Viewed Structurally");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("clas", "Classification");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("clna", "Clinical Attribute");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("clnd", "Clinical Drug");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("cnce", "Conceptual Entity");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("comd", "Cell or Molecular Dysfunction");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("crbs", "Carbohydrate Sequence");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("diap", "Diagnostic Procedure");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("dora", "Daily or Recreational Activity");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("drdd", "Drug Delivery Device");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("dsyn", "Disease or Syndrome");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("edac", "Educational Activity");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("eehu", "Environmental Effect of Humans");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("eico", "Eicosanoid");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("elii", "Element, Ion, or Isotope");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("emod", "Experimental Model of Disease");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("emst", "Embryonic Structure");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("enty", "Entity");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("enzy", "Enzyme");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("euka", "Eukaryote");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("evnt", "Event");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("famg", "Family Group");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("ffas", "Fully Formed Anatomical Structure");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("fish", "Fish");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("fndg", "Finding");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("fngs", "Fungus");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("food", "Food");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("ftcn", "Functional Concept");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("genf", "Genetic Function");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("geoa", "Geographic Area");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("gngm", "Gene or Genome");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("gora", "Governmental or Regulatory Activity");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("grpa", "Group Attribute");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("grup", "Group");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("hcpp", "Human-caused Phenomenon or Process");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("hcro", "Health Care Related Organization");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("hlca", "Health Care Activity");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("hops", "Hazardous or Poisonous Substance");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("horm", "Hormone");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("humn", "Human");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("idcn", "Idea or Concept");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("imft", "Immunologic Factor");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("inbe", "Individual Behavior");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("inch", "Inorganic Chemical");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("inpo", "Injury or Poisoning");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("inpr", "Intellectual Product");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("irda", "Indicator, Reagent, or Diagnostic Aid");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("lang", "Language");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("lbpr", "Laboratory Procedure");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("lbtr", "Laboratory or Test Result");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("lipd", "Lipid");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("mamm", "Mammal");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("mbrt", "Molecular Biology Research Technique");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("mcha", "Machine Activity");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("medd", "Medical Device");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("menp", "Mental Process");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("mnob", "Manufactured Object");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("mobd", "Mental or Behavioral Dysfunction");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("moft", "Molecular Function");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("mosq", "Molecular Sequence");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("neop", "Neoplastic Process");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("nnon", "Nucleic Acid, Nucleoside, or Nucleotide");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("npop", "Natural Phenomenon or Process");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("nsba", "Neuroreactive Substance or Biogenic Amine");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("nusq", "Nucleotide Sequence");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("ocac", "Occupational Activity");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("ocdi", "Occupation or Discipline");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("opco", "Organophosphorus Compound");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("orch", "Organic Chemical");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("orga", "Organism Attribute");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("orgf", "Organism Function");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("orgm", "Organism");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("orgt", "Organization");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("ortf", "Organ or Tissue Function");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("patf", "Pathologic Function");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("phob", "Physical Object");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("phpr", "Phenomenon or Process");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("phsf", "Physiologic Function");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("phsu", "Pharmacologic Substance");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("plnt", "Plant");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("podg", "Patient or Disabled Group");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("popg", "Population Group");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("prog", "Professional or Occupational Group");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("pros", "Professional Society");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("qlco", "Qualitative Concept");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("qnco", "Quantitative Concept");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("rcpt", "Receptor");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("rept", "Reptile");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("resa", "Research Activity");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("resd", "Research Device");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("rnlw", "Regulation or Law");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("sbst", "Substance");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("shro", "Self-help or Relief Organization");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("socb", "Social Behavior");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("sosy", "Sign or Symptom");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("spco", "Spatial Concept");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("strd", "Steroid");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("tisu", "Tissue");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("tmco", "Temporal Concept");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("topp", "Therapeutic or Preventive Procedure");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("virs", "Virus");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("vita", "Vitamin");
		semanticTypeAbbreviationsToSemanticTypeFullNames.put("vtbt", "Vertebrate");
	}
	
	// Key = TUI of semantic type, value = abbreviation of semantic type
	private static final Map<String,String> tuisToSemanticTypeAbbreviations = new HashMap<String,String>();
	static{
		tuisToSemanticTypeAbbreviations.put("T001", "orgm");
		tuisToSemanticTypeAbbreviations.put("T002", "plnt");
		tuisToSemanticTypeAbbreviations.put("T004", "fngs");
		tuisToSemanticTypeAbbreviations.put("T005", "virs");
		tuisToSemanticTypeAbbreviations.put("T007", "bact");
		tuisToSemanticTypeAbbreviations.put("T008", "anim");
		tuisToSemanticTypeAbbreviations.put("T010", "vtbt");
		tuisToSemanticTypeAbbreviations.put("T011", "amph");
		tuisToSemanticTypeAbbreviations.put("T012", "bird");
		tuisToSemanticTypeAbbreviations.put("T013", "fish");
		tuisToSemanticTypeAbbreviations.put("T014", "rept");
		tuisToSemanticTypeAbbreviations.put("T015", "mamm");
		tuisToSemanticTypeAbbreviations.put("T016", "humn");
		tuisToSemanticTypeAbbreviations.put("T017", "anst");
		tuisToSemanticTypeAbbreviations.put("T018", "emst");
		tuisToSemanticTypeAbbreviations.put("T019", "cgab");
		tuisToSemanticTypeAbbreviations.put("T020", "acab");
		tuisToSemanticTypeAbbreviations.put("T021", "ffas");
		tuisToSemanticTypeAbbreviations.put("T022", "bdsy");
		tuisToSemanticTypeAbbreviations.put("T023", "bpoc");
		tuisToSemanticTypeAbbreviations.put("T024", "tisu");
		tuisToSemanticTypeAbbreviations.put("T025", "cell");
		tuisToSemanticTypeAbbreviations.put("T026", "celc");
		tuisToSemanticTypeAbbreviations.put("T028", "gngm");
		tuisToSemanticTypeAbbreviations.put("T029", "blor");
		tuisToSemanticTypeAbbreviations.put("T030", "bsoj");
		tuisToSemanticTypeAbbreviations.put("T031", "bdsu");
		tuisToSemanticTypeAbbreviations.put("T032", "orga");
		tuisToSemanticTypeAbbreviations.put("T033", "fndg");
		tuisToSemanticTypeAbbreviations.put("T034", "lbtr");
		tuisToSemanticTypeAbbreviations.put("T037", "inpo");
		tuisToSemanticTypeAbbreviations.put("T038", "biof");
		tuisToSemanticTypeAbbreviations.put("T039", "phsf");
		tuisToSemanticTypeAbbreviations.put("T040", "orgf");
		tuisToSemanticTypeAbbreviations.put("T041", "menp");
		tuisToSemanticTypeAbbreviations.put("T042", "ortf");
		tuisToSemanticTypeAbbreviations.put("T043", "celf");
		tuisToSemanticTypeAbbreviations.put("T044", "moft");
		tuisToSemanticTypeAbbreviations.put("T045", "genf");
		tuisToSemanticTypeAbbreviations.put("T046", "patf");
		tuisToSemanticTypeAbbreviations.put("T047", "dsyn");
		tuisToSemanticTypeAbbreviations.put("T048", "mobd");
		tuisToSemanticTypeAbbreviations.put("T049", "comd");
		tuisToSemanticTypeAbbreviations.put("T050", "emod");
		tuisToSemanticTypeAbbreviations.put("T051", "evnt");
		tuisToSemanticTypeAbbreviations.put("T052", "acty");
		tuisToSemanticTypeAbbreviations.put("T053", "bhvr");
		tuisToSemanticTypeAbbreviations.put("T054", "socb");
		tuisToSemanticTypeAbbreviations.put("T055", "inbe");
		tuisToSemanticTypeAbbreviations.put("T056", "dora");
		tuisToSemanticTypeAbbreviations.put("T057", "ocac");
		tuisToSemanticTypeAbbreviations.put("T058", "hlca");
		tuisToSemanticTypeAbbreviations.put("T059", "lbpr");
		tuisToSemanticTypeAbbreviations.put("T060", "diap");
		tuisToSemanticTypeAbbreviations.put("T061", "topp");
		tuisToSemanticTypeAbbreviations.put("T062", "resa");
		tuisToSemanticTypeAbbreviations.put("T063", "mbrt");
		tuisToSemanticTypeAbbreviations.put("T064", "gora");
		tuisToSemanticTypeAbbreviations.put("T065", "edac");
		tuisToSemanticTypeAbbreviations.put("T066", "mcha");
		tuisToSemanticTypeAbbreviations.put("T067", "phpr");
		tuisToSemanticTypeAbbreviations.put("T068", "hcpp");
		tuisToSemanticTypeAbbreviations.put("T069", "eehu");
		tuisToSemanticTypeAbbreviations.put("T070", "npop");
		tuisToSemanticTypeAbbreviations.put("T071", "enty");
		tuisToSemanticTypeAbbreviations.put("T072", "phob");
		tuisToSemanticTypeAbbreviations.put("T073", "mnob");
		tuisToSemanticTypeAbbreviations.put("T074", "medd");
		tuisToSemanticTypeAbbreviations.put("T075", "resd");
		tuisToSemanticTypeAbbreviations.put("T077", "cnce");
		tuisToSemanticTypeAbbreviations.put("T078", "idcn");
		tuisToSemanticTypeAbbreviations.put("T079", "tmco");
		tuisToSemanticTypeAbbreviations.put("T080", "qlco");
		tuisToSemanticTypeAbbreviations.put("T081", "qnco");
		tuisToSemanticTypeAbbreviations.put("T082", "spco");
		tuisToSemanticTypeAbbreviations.put("T083", "geoa");
		tuisToSemanticTypeAbbreviations.put("T085", "mosq");
		tuisToSemanticTypeAbbreviations.put("T086", "nusq");
		tuisToSemanticTypeAbbreviations.put("T087", "amas");
		tuisToSemanticTypeAbbreviations.put("T088", "crbs");
		tuisToSemanticTypeAbbreviations.put("T089", "rnlw");
		tuisToSemanticTypeAbbreviations.put("T090", "ocdi");
		tuisToSemanticTypeAbbreviations.put("T091", "bmod");
		tuisToSemanticTypeAbbreviations.put("T092", "orgt");
		tuisToSemanticTypeAbbreviations.put("T093", "hcro");
		tuisToSemanticTypeAbbreviations.put("T094", "pros");
		tuisToSemanticTypeAbbreviations.put("T095", "shro");
		tuisToSemanticTypeAbbreviations.put("T096", "grup");
		tuisToSemanticTypeAbbreviations.put("T097", "prog");
		tuisToSemanticTypeAbbreviations.put("T098", "popg");
		tuisToSemanticTypeAbbreviations.put("T099", "famg");
		tuisToSemanticTypeAbbreviations.put("T100", "aggp");
		tuisToSemanticTypeAbbreviations.put("T101", "podg");
		tuisToSemanticTypeAbbreviations.put("T102", "grpa");
		tuisToSemanticTypeAbbreviations.put("T103", "chem");
		tuisToSemanticTypeAbbreviations.put("T104", "chvs");
		tuisToSemanticTypeAbbreviations.put("T109", "orch");
		tuisToSemanticTypeAbbreviations.put("T110", "strd");
		tuisToSemanticTypeAbbreviations.put("T111", "eico");
		tuisToSemanticTypeAbbreviations.put("T114", "nnon");
		tuisToSemanticTypeAbbreviations.put("T115", "opco");
		tuisToSemanticTypeAbbreviations.put("T116", "aapp");
		tuisToSemanticTypeAbbreviations.put("T118", "carb");
		tuisToSemanticTypeAbbreviations.put("T119", "lipd");
		tuisToSemanticTypeAbbreviations.put("T120", "chvf");
		tuisToSemanticTypeAbbreviations.put("T121", "phsu");
		tuisToSemanticTypeAbbreviations.put("T122", "bodm");
		tuisToSemanticTypeAbbreviations.put("T123", "bacs");
		tuisToSemanticTypeAbbreviations.put("T124", "nsba");
		tuisToSemanticTypeAbbreviations.put("T125", "horm");
		tuisToSemanticTypeAbbreviations.put("T126", "enzy");
		tuisToSemanticTypeAbbreviations.put("T127", "vita");
		tuisToSemanticTypeAbbreviations.put("T129", "imft");
		tuisToSemanticTypeAbbreviations.put("T130", "irda");
		tuisToSemanticTypeAbbreviations.put("T131", "hops");
		tuisToSemanticTypeAbbreviations.put("T167", "sbst");
		tuisToSemanticTypeAbbreviations.put("T168", "food");
		tuisToSemanticTypeAbbreviations.put("T169", "ftcn");
		tuisToSemanticTypeAbbreviations.put("T170", "inpr");
		tuisToSemanticTypeAbbreviations.put("T171", "lang");
		tuisToSemanticTypeAbbreviations.put("T184", "sosy");
		tuisToSemanticTypeAbbreviations.put("T185", "clas");
		tuisToSemanticTypeAbbreviations.put("T190", "anab");
		tuisToSemanticTypeAbbreviations.put("T191", "neop");
		tuisToSemanticTypeAbbreviations.put("T192", "rcpt");
		tuisToSemanticTypeAbbreviations.put("T194", "arch");
		tuisToSemanticTypeAbbreviations.put("T195", "antb");
		tuisToSemanticTypeAbbreviations.put("T196", "elii");
		tuisToSemanticTypeAbbreviations.put("T197", "inch");
		tuisToSemanticTypeAbbreviations.put("T200", "clnd");
		tuisToSemanticTypeAbbreviations.put("T201", "clna");
		tuisToSemanticTypeAbbreviations.put("T203", "drdd");
		tuisToSemanticTypeAbbreviations.put("T204", "euka");
	}
	
	public static List<String> getAllSemanticTypes(){
		List<String> semanticTypes = new ArrayList<String>(semanticTypeAbbreviationsToSemanticTypeFullNames.keySet());
		Collections.sort(semanticTypes);
		return semanticTypes;
	}
	
	public static void initialize(List<String> lines){
		//-----------------------------------------
		// Currently this is only used by hadoop code.
		// Preprocess MRSTY.RRF file so that each line
		// only contains the CUI and TUI. Though eventually
		// the method initializeSemanticTypeLookupTable(List<String>)
		// works for both abridged and unabridged MRSTY.RRF files
		if (!initialized){
			logger.info("Generating semantic type lookup table...");
			
			//-----------------------------------------
			// Set up the look up maps
			initializeLookupMaps();

			//-----------------------------------------
			// Fill in the lemmas
			initializeSemanticTypeLookupTable(lines);

			initialized = true;
			logger.info("Done");
		}
	}
	
	public static void initialize(Properties settings) throws Exception{
		if (!initialized){
			logger.info("Generating semantic type lookup table...");
			
			//-----------------------------------------
			// Set up the look up maps
			initializeLookupMaps();

			//-----------------------------------------
			// Fill in the lemmas
			initializeSemanticTypeLookupTable(settings);

			initialized = true;
			logger.info("Done");
		}
	}
	
	private static void initializeLookupMaps(){
		//-----------------------------------------
		// Initialize lookup maps
		cuisToSemanticTypeIds = new ArrayList<Map<Integer, Integer>>();
		semanticTypeAbbreviationsToSemanticTypeIds = new HashMap<String, Integer>();
		semanticTypeIdsToSemanticTypeAbbreviations = new HashMap<Integer, String>();
	}
	
	private static void initializeSemanticTypeLookupTable(List<String> inputLines){		
		//-----------------------------------------
		// Go thru the lines one by one, and extract the TUI for each CUI.
		// If a CUI has multiple TUIs, keep them all.
		int previousCuiAsInt = -1;
		int index = 0; // For which map in the main CUI --> semantic type ID data structure
		for (String inputLine : inputLines){
			//-----------------------------------------
			// Tokenize the string to get the CUI and TUI
			// Note to self: StringUtils.split() treats multiple empty
			// columns as a single column
			List<String> tokens = Arrays.asList(inputLine.split("\\|"));
			String cui = tokens.get(0);
			String tui = tokens.get(1);
			
			int cuiAsInt = CuiAndVersion.toInt(cui);
			
			//-----------------------------------------
			// If the CUI has changed, update the count
			if (cuiAsInt != previousCuiAsInt){
				index = 0;
			}
			
			//-----------------------------------------
			// Find out the TUI's corresponding semantic type
			String semanticType = lookupViaTui(tui);
			
			//-----------------------------------------
			// Find out the semantic type ID
			int semanticTypeId = -1;
			if (semanticTypeAbbreviationsToSemanticTypeIds.containsKey(semanticType)){
				semanticTypeId = semanticTypeAbbreviationsToSemanticTypeIds.get(semanticType);
			}
			else{
				semanticTypeId = semanticTypeAbbreviationsToSemanticTypeIds.size();
				semanticTypeIdsToSemanticTypeAbbreviations.put(semanticTypeId, semanticType);
				semanticTypeAbbreviationsToSemanticTypeIds.put(semanticType, semanticTypeId);
			}
			
			//-----------------------------------------
			// Create a new CUI --> semantic type ID map if needed
			if (index == cuisToSemanticTypeIds.size()){
				cuisToSemanticTypeIds.add(new HashMap<Integer, Integer>());
			}
			
			//-----------------------------------------
			// Register the CUI's semantic type
			cuisToSemanticTypeIds.get(index).put(cuiAsInt, semanticTypeId);
			
			//-----------------------------------------
			// Get ready for next line
			++index;
			previousCuiAsInt = cuiAsInt;
		}
	}
	
	private static void initializeSemanticTypeLookupTable(Properties settings) throws IOException{		
		//-----------------------------------------
		// Read in the MRSTY.RRF file
		String mrstyFilename = settings.getProperty("umls_mrsty_filename");
		List<String> inputLines = ReadFromFile.readFileToListOfStrings(mrstyFilename);
		
		//-----------------------------------------
		// Now we can use the other method
		initializeSemanticTypeLookupTable(inputLines);
	}
	
	public static String getAbbreviation(String fullName){
		// Return empty string if the input string does not
		// correspond to a semantic type
		String returnAbbreviation = "";
		
		//-----------------------------------------
		// There is no smart way of doing this. Look at
		// the abbreviation --> full name entries one by
		// one, until we find it or we exhaust the list.
		Iterator<Entry<String, String>> iterator = semanticTypeAbbreviationsToSemanticTypeFullNames.entrySet().iterator();
		while (returnAbbreviation.isEmpty() && iterator.hasNext()){
			Entry<String, String> entry = iterator.next();
			String abbreviation = entry.getKey();
			String currentFullName = entry.getValue();
			
			if (currentFullName.equalsIgnoreCase(fullName)){
				returnAbbreviation = abbreviation;
			}
		}
		
		return returnAbbreviation;
	}
	
	public static String getFullName(String semanticType){
		return semanticTypeAbbreviationsToSemanticTypeFullNames.get(semanticType);
	}
	
	public static boolean isSemanticType(String inputString){
		inputString = inputString.toLowerCase();
		return semanticTypeAbbreviationsToSemanticTypeFullNames.containsKey(inputString);
	}
	
	public static List<String> lookup(String cui){
		return lookup(CuiAndVersion.toInt(cui));
	}
	
	public static List<String> lookup(int cuiAsInt){
		List<String> returnList = new ArrayList<String>();
		
		int i = 0;
		while (i < cuisToSemanticTypeIds.size() && cuisToSemanticTypeIds.get(i).containsKey(cuiAsInt)){
			int semanticTypeId = cuisToSemanticTypeIds.get(i).get(cuiAsInt);
			String semanticType = semanticTypeIdsToSemanticTypeAbbreviations.get(semanticTypeId);
			returnList.add(semanticType);
			
			++i;
		}
		
		Collections.sort(returnList);
		return returnList;
	}
	
	public static List<String> lookup(long cuiAndVersion){
		return lookup(CuiAndVersion.toInt(cuiAndVersion));
	}
	
	public static String lookupViaTui(String tui){
		return tuisToSemanticTypeAbbreviations.get(tui);
	}
}