/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.wrapper.umls;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import mpii.biotm.util.CompareMapByValueIntegerDescending;

import org.apache.log4j.Logger;

public class UnwantedSubStringsInConceptNames{
	
	protected static final Logger logger = Logger.getLogger(UnwantedSubStringsInConceptNames.class);
	
	//-----------------------------------------
	// Track the sub strings in brackets, and how many
	// of each there are. The goal is to look for the common
	// ones and remove them.
	private static Map<String, Integer> bracketsToCounts = null;
	
	private static boolean initialized = false;
	private static Pattern patternBracketSubString = Pattern.compile("[\\[\\(].+?[\\]\\)]");
	private static List<Pattern> patternsUnwantedSubStringsBefore = null;
	private static List<Pattern> patternsUnwantedSubStringsAfter = null;
	
	//-----------------------------------------
	// The known sub strings in concept names that we don't want
	private static List<String> unwantedSubStrings = new ArrayList<String>();
	static{
		unwantedSubStrings.add("(+)");
		unwantedSubStrings.add("(-)");
		unwantedSubStrings.add("(+-)");
		unwantedSubStrings.add("(+/-)");
		unwantedSubStrings.add("(#___)");
		unwantedSubStrings.add("[#/volume]");
		unwantedSubStrings.add("(___cm)");
		unwantedSubStrings.add("(___ cm)");
		unwantedSubStrings.add("(___ degree.)");
		unwantedSubStrings.add("(___ degrees)");
		unwantedSubStrings.add("(___ min)");
		unwantedSubStrings.add("(___ mm)");
		unwantedSubStrings.add("(& level)");
		unwantedSubStrings.add("(1)");
		unwantedSubStrings.add("(1:1)");
		unwantedSubStrings.add("(1-4)");
		unwantedSubStrings.add("(2)");
		unwantedSubStrings.add("(3)");
		unwantedSubStrings.add("(4)");
		unwantedSubStrings.add("(5)");
		unwantedSubStrings.add("(6)");
		unwantedSubStrings.add("(7)");
		unwantedSubStrings.add("(8)");
		unwantedSubStrings.add("(9)");
		unwantedSubStrings.add("(10)");
		unwantedSubStrings.add("(14)");
		unwantedSubStrings.add("(2010)");
		unwantedSubStrings.add("(2011)");
		unwantedSubStrings.add("(2012)");
		unwantedSubStrings.add("(a)");
		unwantedSubStrings.add("(activity)");
		unwantedSubStrings.add("(allelic variant)");
		unwantedSubStrings.add("[Ambiguous]");
		unwantedSubStrings.add("(Approved Lists 1980)");
		unwantedSubStrings.add("(assessment scale)");
		unwantedSubStrings.add("(attribute)");
		unwantedSubStrings.add("(b)");
		unwantedSubStrings.add("(Base Equivalent)");
		unwantedSubStrings.add("[BAU]");
		unwantedSubStrings.add("(biological function)");
		unwantedSubStrings.add("(body structure)");
		unwantedSubStrings.add("(Bulk)");
		unwantedSubStrings.add("(cell)");
		unwantedSubStrings.add("(cell structure)");
		unwantedSubStrings.add("[Chemical/Ingredient]");
		unwantedSubStrings.add("(combined site)");
		unwantedSubStrings.add("(context-dependent category)");
		unwantedSubStrings.add("(count/vol)");
		unwantedSubStrings.add("(CT Scan)");
		unwantedSubStrings.add("(D)");
		unwantedSubStrings.add("[D]");
		unwantedSubStrings.add("(diagnosis)");
		unwantedSubStrings.add("(diagnostic)");
		unwantedSubStrings.add("[Disease/Finding]");
		unwantedSubStrings.add("(discontinued)");
		unwantedSubStrings.add("(disorder)");
		unwantedSubStrings.add("(documented clinically or microbiologically)");
		unwantedSubStrings.add("(Drosophila)");
		unwantedSubStrings.add("[dup]");
		unwantedSubStrings.add("(E)");
		unwantedSubStrings.add("(EA)");
		unwantedSubStrings.add("(ENTERIC COATED)");
		unwantedSubStrings.add("(environment)");
		unwantedSubStrings.add("[EPC]");
		unwantedSubStrings.add("(etiology)");
		unwantedSubStrings.add("(event)");
		unwantedSubStrings.add("[FACIT]");
		unwantedSubStrings.add("(FC)");
		unwantedSubStrings.add("(FINAL DOSE FORM)");
		unwantedSubStrings.add("(finding)");
		unwantedSubStrings.add("(function)");
		unwantedSubStrings.add("(geographic location)");
		unwantedSubStrings.add("(GM)");
		unwantedSubStrings.add("(GRAM)");
		unwantedSubStrings.add("(GRAMS)");
		unwantedSubStrings.add("(H1N1)");
		unwantedSubStrings.add("(H3N2)");
		unwantedSubStrings.add("(HARD, SOFT, ETC.)");
		unwantedSubStrings.add("(HDR)");
		unwantedSubStrings.add("(history)");
		unwantedSubStrings.add("[hp_C]");
		unwantedSubStrings.add("[hp_X]");
		unwantedSubStrings.add("[Identifier]");
		unwantedSubStrings.add("(I)");
		unwantedSubStrings.add("(II)");
		unwantedSubStrings.add("(III)");
		unwantedSubStrings.add("(INHALATION)");
		unwantedSubStrings.add("(IR)");
		unwantedSubStrings.add("[iU]");
		unwantedSubStrings.add("(L)");
		unwantedSubStrings.add("(L.)");
		unwantedSubStrings.add("(lab test)");
		unwantedSubStrings.add("(LDR)");
		unwantedSubStrings.add("(Linnaeus, 1758)");
		unwantedSubStrings.add("(M)");
		unwantedSubStrings.add("[M]");
		unwantedSubStrings.add("[Mass]");
		unwantedSubStrings.add("[Mass ratio]");
		unwantedSubStrings.add("[Mass/mass]");
		unwantedSubStrings.add("[Mass/time]");
		unwantedSubStrings.add("[Mass/volume]");
		unwantedSubStrings.add("(mammal)");
		unwantedSubStrings.add("(manifestation)");
		unwantedSubStrings.add("(mechanical)");
		unwantedSubStrings.add("(medication)");
		unwantedSubStrings.add("(MeSH Category)");
		unwantedSubStrings.add("(MILLILITERS)");
		unwantedSubStrings.add("(ML)");
		unwantedSubStrings.add("[MoA]");
		unwantedSubStrings.add("[Molar ratio]");
		unwantedSubStrings.add("[Moles/volume]");
		unwantedSubStrings.add("(morphologic abnormality)");
		unwantedSubStrings.add("(MRI)");
		unwantedSubStrings.add("(N)");
		unwantedSubStrings.add("(nail)");
		unwantedSubStrings.add("(navigational concept)");
		unwantedSubStrings.add("(non-specific)");
		unwantedSubStrings.add("nos");
		unwantedSubStrings.add("NOS");
		unwantedSubStrings.add("[nos]");
		unwantedSubStrings.add("[NOS]");
		unwantedSubStrings.add("(nos)");
		unwantedSubStrings.add("(NOS)");
		unwantedSubStrings.add("not otherwise specified");
		unwantedSubStrings.add("(observable entity)");
		unwantedSubStrings.add("(Obsolete)");
		unwantedSubStrings.add("(obsolete)");
		unwantedSubStrings.add("(occupation)");
		unwantedSubStrings.add("of unspecified site");
		unwantedSubStrings.add("(or disorder)");
		unwantedSubStrings.add("(organism)");
		unwantedSubStrings.add("[PE]");
		unwantedSubStrings.add("(person)");
		unwantedSubStrings.add("(pdr for recon)");
		unwantedSubStrings.add("[PhenX]");
		unwantedSubStrings.add("(physical finding)");
		unwantedSubStrings.add("(physical force)");
		unwantedSubStrings.add("(physical object)");
		unwantedSubStrings.add("[PNU]");
		unwantedSubStrings.add("[Presence]");
		unwantedSubStrings.add("(procedure)");
		unwantedSubStrings.add("(product)");
		unwantedSubStrings.add("(qualifier value)");
		unwantedSubStrings.add("(R)");
		unwantedSubStrings.add("[Ratio]");
		unwantedSubStrings.add("(regime/therapy)");
		unwantedSubStrings.add("-RETIRED-");
		unwantedSubStrings.add("(S)");
		unwantedSubStrings.add("(s)");
		unwantedSubStrings.add("(S/I)");
		unwantedSubStrings.add("(S. cerevisiae)");
		unwantedSubStrings.add("(SDV,MDV OR ADDITIVE)");
		unwantedSubStrings.add("(situation)");
		unwantedSubStrings.add("(specimen)");
		unwantedSubStrings.add("(structure)");
		unwantedSubStrings.add("(substance)");
		unwantedSubStrings.add("[Susceptibility]");
		unwantedSubStrings.add("(symptom)");
		unwantedSubStrings.add("(systemic)");
		unwantedSubStrings.add("(T)");
		unwantedSubStrings.add("(Titer)");
		unwantedSubStrings.add("[Titer]");
		unwantedSubStrings.add("[TREATMENT]");
		unwantedSubStrings.add("(treatment)");
		unwantedSubStrings.add("(tumor staging)");
		unwantedSubStrings.add("(unintentional)");
		unwantedSubStrings.add("[Units/volume]");
		unwantedSubStrings.add("unspecified");
		unwantedSubStrings.add("unspecified site");
		unwantedSubStrings.add("(USP)");
		unwantedSubStrings.add("[USP'U]");
		unwantedSubStrings.add("[V]");
		unwantedSubStrings.add("[VA Drug Interaction]");
		unwantedSubStrings.add("[VA Product]");
		unwantedSubStrings.add("[X]");
		unwantedSubStrings.add("(Z)");
	}
	
	private static void constructUnwantedSubStringPatterns(){
		patternsUnwantedSubStringsBefore = new ArrayList<Pattern>();
		patternsUnwantedSubStringsAfter = new ArrayList<Pattern>();
		
		int count = 0;
		String tempBefore = "";
		String tempAfter = "";
		for (String unwantedSubString : unwantedSubStrings){
			unwantedSubString = unwantedSubString.replaceAll("\\[", "\\\\[");
			unwantedSubString = unwantedSubString.replaceAll("\\]", "\\\\]");
			unwantedSubString = unwantedSubString.replaceAll("\\(", "\\\\(");
			unwantedSubString = unwantedSubString.replaceAll("\\)", "\\\\)");
			
			String spacer = "";
			if (unwantedSubString.matches("^[\\s\\w]+$")){
				spacer = " ";
			}
			
			tempBefore += unwantedSubString + spacer + "|";
			tempAfter += spacer + unwantedSubString + "|";
			
			++count;
			if (count == 30){
				tempBefore = tempBefore.substring(0, tempBefore.length() - 1);
				tempAfter = tempAfter.substring(0, tempAfter.length() - 1);
				patternsUnwantedSubStringsBefore.add(Pattern.compile("^(" + tempBefore + ")(.*)$"));
				patternsUnwantedSubStringsAfter.add(Pattern.compile("^(.*)(" + tempAfter + ")$"));
				count = 0;
				tempBefore = "";
				tempAfter = "";
			}
		}
		if (!tempBefore.isEmpty()){
			tempBefore = tempBefore.substring(0, tempBefore.length() - 1);
			tempAfter = tempAfter.substring(0, tempAfter.length() - 1);
			patternsUnwantedSubStringsBefore.add(Pattern.compile("^(" + tempBefore + ")(.*)$"));
			patternsUnwantedSubStringsAfter.add(Pattern.compile("^(.*)(" + tempAfter + ")$"));
		}
	}
	
	// For debugging purposes
	public static void displayBracketSubStringCounts(){
		CompareMapByValueIntegerDescending comparator = new CompareMapByValueIntegerDescending(bracketsToCounts);
		TreeMap<String, Integer> sorted = new TreeMap<String, Integer>(comparator);
		sorted.putAll(bracketsToCounts);
		
		for (String bracketSubString : sorted.keySet()){
			logger.debug(sorted.get(bracketSubString) + " " + bracketSubString);
		}
	}
	
	public static void initialize(){
		if (!initialized){
			constructUnwantedSubStringPatterns();
			bracketsToCounts = new HashMap<String, Integer>();
			initialized = true;
		}
	}
	
	public static String removeUnwantedSubString(String conceptName){
		String returnStr = conceptName;
		
		//-----------------------------------------
		// For debugging purposes, look for sub strings in brackets,
		// to see which sub strings we can remove in the future.
		Matcher matcher = patternBracketSubString.matcher(returnStr);
		while (matcher.find()){
			String bracketSubString = matcher.group();
			
			if (!bracketsToCounts.containsKey(bracketSubString)){
				bracketsToCounts.put(bracketSubString, 0);
			}
			
			int count = bracketsToCounts.get(bracketSubString);
			++count;
			bracketsToCounts.put(bracketSubString, count);
		}
		
		//-----------------------------------------
		// Now we do the real work - remove an unwanted sub string
		// if it's at the beginning or the end of a concept name.
		// But do it twice, since some concept names have
		// two unwanted sub strings.
		for (int i = 0; i < 2; ++i){
			returnStr = removeUnwantedSubStringForOneRound(returnStr);
		}
		
		return returnStr;
	}

	private static String removeUnwantedSubStringForOneRound(String string){
		Matcher matcher = null;
		
		for (Pattern patternUnwantedSubStringBefore : patternsUnwantedSubStringsBefore){
			matcher = patternUnwantedSubStringBefore.matcher(string);
			if (matcher.find()){
				string = matcher.group(2);
			}
		}
		for (Pattern patternUnwantedSubStringAfter : patternsUnwantedSubStringsAfter){
			matcher = patternUnwantedSubStringAfter.matcher(string);
			if (matcher.find()){
				string = matcher.group(1);
			}
		}
		
		//-----------------------------------------
		// Clean up the final string
		string = string.replaceAll(",\\s$", ""); // Remove ", " suffix
		string = string.replaceAll("\\s+", " "); // Compact multiple spaces into single space
		string = string.trim();
		
		return string;
	}
}