/*
 * Copyright 2015 Amy Siu (siu@mpi-inf.mpg.de)
 *
 * This file is part of the BEBE experimental software package.
 *
 * BEBE is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BEBE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BEBE.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package mpii.biotm.wrapper.wordnet;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import org.apache.log4j.Logger;

import edu.mit.jwi.Dictionary;
import edu.mit.jwi.IDictionary;
import edu.mit.jwi.item.IIndexWord;
import edu.mit.jwi.item.ISynset;
import edu.mit.jwi.item.ISynsetID;
import edu.mit.jwi.item.IWord;
import edu.mit.jwi.item.IWordID;
import edu.mit.jwi.item.POS;
import edu.mit.jwi.item.Pointer;

public class WordNetDictionary{
	
	protected static final Logger logger = Logger.getLogger(WordNetDictionary.class);
	private static boolean initialized = false;
	private static IDictionary dictionary = null;
	
	private static List<POS> partsOfSpeech = new ArrayList<POS>();
	static{
		partsOfSpeech.add(POS.ADVERB);
		partsOfSpeech.add(POS.ADJECTIVE);
		partsOfSpeech.add(POS.NOUN);
		partsOfSpeech.add(POS.VERB);
	}
	
	public static void close(){
		if (initialized){
			dictionary.close();
		}
	}
	
	public static List<ISynset> getRelatedSynsets(ISynset synset, Pointer relation){
		List<ISynset> returnList = new ArrayList<ISynset>();
		
		//-----------------------------------------
		// Get the related synset IDs under the
		// given pointer / relation
		List<ISynsetID> relatedSynsetIds = synset.getRelatedSynsets(relation);
		
		//-----------------------------------------
		// Get the related synsets from their IDs
		for (ISynsetID relatedSynsetId : relatedSynsetIds){
			returnList.add(dictionary.getSynset(relatedSynsetId));
		}
		
		return returnList;
	}
	
	public static List<IWord> getRelatedWordObjects(IWord wordObject, Pointer relation){
		List<IWord> returnList = new ArrayList<IWord>();
		
		//-----------------------------------------
		// Get the related word object IDs under the
		// given pointer / relation
		List<IWordID> relatedWordObjectIds = wordObject.getRelatedWords(relation);
		
		//-----------------------------------------
		// Get the related word objects from their IDs
		for (IWordID relatedWordObjectId : relatedWordObjectIds){
			returnList.add(dictionary.getWord(relatedWordObjectId));
		}
		
		return returnList;
	}
	
	public static List<ISynset> getSynsets(POS pos){
		List<ISynset> returnList = new ArrayList<ISynset>();
		
		Iterator<ISynset> iterator = dictionary.getSynsetIterator(pos);
		while (iterator.hasNext()){
			ISynset synset = iterator.next();
			returnList.add(synset);
		}
		
		return returnList;
	}
	
	public static List<IWord> getWordObjects(ISynset synset){
		//-----------------------------------------
		// Note to self: word numbers in the synset start from 1, not 0!!
		return synset.getWords();
	}
	
	public static List<IWord> getWordObjects(POS pos){
		List<IWord> returnList = new ArrayList<IWord>();

		Iterator<IIndexWord> iterator = dictionary.getIndexWordIterator(pos);
		while (iterator.hasNext()){
			IIndexWord indexWord = iterator.next();
			List<IWordID> wordIds = indexWord.getWordIDs();
			
			//-----------------------------------------
			// Go thru each word ID in the dictionary
			for (IWordID wordId : wordIds){
				IWord wordObject = dictionary.getWord(wordId);
				returnList.add(wordObject);
			}
		}
		
		return returnList;
	}
	
	public static List<IWord> getWordObjects(String wordString){
		Set<IWord> cache = new HashSet<IWord>();
		
		for (POS pos : partsOfSpeech){
			cache.addAll(getWordObjects(wordString, pos));
		}
		
		return new ArrayList<IWord>(cache);
	}
	
	public static List<IWord> getWordObjects(String wordString, POS pos){
		Set<IWord> cache = new HashSet<IWord>();
		
		IIndexWord indexWord = dictionary.getIndexWord(wordString, pos);
		if (indexWord != null){
			List<IWordID> wordIds = indexWord.getWordIDs();
			for (IWordID wordId : wordIds){
				IWord wordObject = dictionary.getWord(wordId);
				cache.add(wordObject);
			}
		}
		
		return new ArrayList<IWord>(cache);
	}
	
	public static String getWordString(IWord wordObject){
		return wordObject.getLemma().replace("_", " ");
	}
	
	public static void initialize() throws IOException{
		//-----------------------------------------
		// Find out whether we are running things from eclipse
		// or from jar
		boolean runningFromEclipse = false;
		URL resource = WordNetDictionary.class.getResource("WordNetDictionary.class");
		String resourceAsString = resource.toString();

		if (resourceAsString.startsWith("file")){
			runningFromEclipse = true;
		}
		else if (resourceAsString.startsWith("jar")){
			runningFromEclipse = false;
		}
		else{
			logger.warn("Unknown method of running java -- it's neither eclipse nor jar");
		}
		
		if (runningFromEclipse){
			initialize("etc/mpii/biotm/wrapper/wordnet/"); // Default location
		}
		else{
			initializeForHadoop();
		}
	}
	
	public static void initialize(Properties settings) throws IOException{
		String wordNetFolder = settings.getProperty("wordnet_directory");
		initialize(wordNetFolder);
	}
	
	public static void initialize(File wordNetDictFolder) throws IOException{
		if (!initialized){
			dictionary = new Dictionary(wordNetDictFolder);
			dictionary.open();

			initialized = true;
		}
	}
	
	public static void initialize(String wordNetDictFolder) throws IOException{
		File file = new File(wordNetDictFolder);
		initialize(file);
	}
	
	public static void initializeForHadoop() throws IOException{
		if (!initialized){
			dictionary = new Dictionary();
			dictionary.open();

			initialized = true;
		}
	}
	
	public static boolean isAdverb(String wordString){
		return isPos(wordString, POS.ADVERB);
	}
	
	public static boolean isAdjective(String wordString){
		return isPos(wordString, POS.ADJECTIVE);
	}
	
	public static boolean isNoun(String wordString){
		return isPos(wordString, POS.NOUN);
	}
	
	public static boolean isPos(String wordString, POS pos){
		List<IWord> wordObjects = getWordObjects(wordString, pos);
		return !wordObjects.isEmpty();
	}
	
	public static boolean isVerb(String wordString){
		return isPos(wordString, POS.VERB);
	}
	
	public static boolean isInDictionary(String wordString){
		List<IWord> wordObjects = getWordObjects(wordString);
		return !wordObjects.isEmpty();
	}
}
